Package: ARDL
Type: Package
Title: ARDL, ECM and Bounds-Test for Cointegration
Description: Creates complex autoregressive distributed lag (ARDL) models and 
    constructs the underlying unrestricted and restricted error correction 
    model (ECM) automatically, just by providing the order. It also performs
    the bounds-test for cointegration as described in Pesaran et al. (2001) 
    <doi:10.1002/jae.616> and provides the multipliers and the cointegrating
    equation. The validity and the accuracy of this package have been verified 
    by successfully replicating the results of Pesaran et al. (2001) in 
    Natsiopoulos and Tzeremes (2022) <doi:10.1002/jae.2919>.
Version: 0.2.5
Authors@R: c(
  person("Kleanthis", "Natsiopoulos", 
  email = "klnatsio@gmail.com", role = c("aut", "cre"), 
  comment = c(ORCID = "0000-0003-1180-2984")), 
  person("Nickolaos", "Tzeremes", 
  email = "bus9nt@econ.uth.gr", role = "ths",
  comment = c(ORCID = "0000-0002-6938-3404")),
  person("Daniel", "Finnan", 
  email = "dan@custom-made.org.uk", role = "aut",
  comment = c(ORCID = "0009-0009-7468-5288")))
BugReports: https://github.com/Natsiopoulos/ARDL/issues
License: GPL-3
URL: https://github.com/Natsiopoulos/ARDL
Encoding: UTF-8
LazyData: true
Depends: R (>= 3.5.0)
Suggests: strucchange, tseries, AICcmodavg, sandwich, testthat (>=
        3.0.0)
Imports: aod, dplyr, dynlm, gridExtra, ggplot2, lmtest, msm, stringr,
        zoo, stats
RoxygenNote: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-09 20:19:35 UTC; KNatsiopoulos001
Author: Kleanthis Natsiopoulos [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1180-2984>),
  Nickolaos Tzeremes [ths] (ORCID:
    <https://orcid.org/0000-0002-6938-3404>),
  Daniel Finnan [aut] (ORCID: <https://orcid.org/0009-0009-7468-5288>)
Maintainer: Kleanthis Natsiopoulos <klnatsio@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-10 05:12:54 UTC
Built: R 4.4.3; ; 2026-05-10 06:41:21 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(ardl,default)
S3method(ardl,uecm)
S3method(coint_eq,default)
S3method(coint_eq,recm)
S3method(delta_method,ardl)
S3method(delta_method,uecm)
S3method(multipliers,ardl)
S3method(multipliers,uecm)
S3method(predict,ardl)
S3method(uecm,ardl)
S3method(uecm,default)
export(ardl)
export(auto_ardl)
export(bounds_f_test)
export(bounds_t_test)
export(coint_eq)
export(multipliers)
export(plot_delay)
export(plot_lr)
export(recm)
export(to_lm)
export(uecm)
importFrom(dplyr,"%>%")
importFrom(zoo,merge.zoo)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# ARDL <img src="man/figures/logo.png" align="right" width="120" />

<!-- badges: start -->

[![downloads](https://cranlogs.r-pkg.org/badges/grand-total/ARDL)](https://CRAN.R-project.org/package=ARDL)
[![CRAN
status](https://www.r-pkg.org/badges/version/ARDL)](https://CRAN.R-project.org/package=ARDL)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![codecov](https://codecov.io/gh/Natsiopoulos/ARDL/branch/master/graph/badge.svg?token=PVZAY57F3D)](https://app.codecov.io/gh/Natsiopoulos/ARDL)
<!-- badges: end -->

## Overview

`ARDL` creates complex autoregressive distributed lag (ARDL) models and
constructs the underlying unrestricted and restricted error correction
model (ECM) automatically, just by providing the order. It also performs
the bounds-test for cointegration as described in [Pesaran et
al. (2001)](https://onlinelibrary.wiley.com/doi/abs/10.1002/jae.616) and
provides the multipliers and the cointegrating equation. The validity
and the accuracy of this package have been verified by successfully
replicating the results of Pesaran et al. (2001) in [Natsiopoulos and
Tzeremes
(2022)](https://onlinelibrary.wiley.com/doi/abs/10.1002/jae.2919).

## Why `ARDL`?

- Estimate complex ARDL models just providing the ARDL order
- Estimate the conditional ECM just providing the underlying ARDL model
  or the order
- Estimate the long-run, short-run, delay, and interim multipliers
- Apply the bound test for no cointegration (*Pesaran et al., 2001*)
  - Both the *F-test* and the *t-test* are available
  - The *p-value* is also available along with the *critical value
    bounds* for specific level of statistical significance
  - Exact *p-values* and *critical value bounds* are available, along
    with the asymptotic ones

## Installation

``` r
# You can install the released version of ARDL from CRAN:
install.packages("ARDL")

# Or the latest development version from GitHub:
install.packages("devtools")
devtools::install_github("Natsiopoulos/ARDL")
```

## Usage

This is a basic example which shows how to use the main functions of the
`ARDL` package.

Assume that we want to model the `LRM` (logarithm of real money, M2) as
a function of `LRY`, `IBO` and `IDE` (see `?denmark`). The problem is
that applying an OLS regression on non-stationary data would result into
a spurious regression. The estimated parameters would be consistent only
if the series were cointegrated.

``` r
library(ARDL)
data(denmark)
```

First, we find the best ARDL specification. We search up to order 5.

``` r
models <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark, max_order = 5)

# The top 20 models according to the AIC
models$top_orders
#>    LRM LRY IBO IDE       AIC
#> 1    3   1   3   2 -251.0259
#> 2    3   1   3   3 -250.1144
#> 3    2   2   0   0 -249.6266
#> 4    3   2   3   2 -249.1087
#> 5    3   2   3   3 -248.1858
#> 6    2   2   0   1 -247.7786
#> 7    2   1   0   0 -247.5643
#> 8    2   2   1   1 -246.6885
#> 9    3   3   3   3 -246.3061
#> 10   2   2   1   2 -246.2709
#> 11   2   1   1   1 -245.8736
#> 12   2   2   2   2 -245.7722
#> 13   1   1   0   0 -245.6620
#> 14   2   1   2   2 -245.1712
#> 15   3   1   2   2 -245.0996
#> 16   1   0   0   0 -244.4317
#> 17   1   1   0   1 -243.7702
#> 18   5   5   5   5 -243.3120
#> 19   4   1   3   2 -243.0728
#> 20   4   1   3   3 -242.4378

# The best model was found to be the ARDL(3,1,3,2)
ardl_3132 <- models$best_model
ardl_3132$order
#> LRM LRY IBO IDE 
#>   3   1   3   2
summary(ardl_3132)
#> 
#> Time series regression with "zooreg" data:
#> Start = 1974 Q4, End = 1987 Q3
#> 
#> Call:
#> dynlm::dynlm(formula = full_formula, data = data, start = start, 
#>     end = end)
#> 
#> Residuals:
#>       Min        1Q    Median        3Q       Max 
#> -0.029939 -0.008856 -0.002562  0.008190  0.072577 
#> 
#> Coefficients:
#>             Estimate Std. Error t value Pr(>|t|)    
#> (Intercept)   2.6202     0.5678   4.615 4.19e-05 ***
#> L(LRM, 1)     0.3192     0.1367   2.336 0.024735 *  
#> L(LRM, 2)     0.5326     0.1324   4.024 0.000255 ***
#> L(LRM, 3)    -0.2687     0.1021  -2.631 0.012143 *  
#> LRY           0.6728     0.1312   5.129 8.32e-06 ***
#> L(LRY, 1)    -0.2574     0.1472  -1.749 0.088146 .  
#> IBO          -1.0785     0.3217  -3.353 0.001790 ** 
#> L(IBO, 1)    -0.1062     0.5858  -0.181 0.857081    
#> L(IBO, 2)     0.2877     0.5691   0.505 0.616067    
#> L(IBO, 3)    -0.9947     0.3925  -2.534 0.015401 *  
#> IDE           0.1255     0.5545   0.226 0.822161    
#> L(IDE, 1)    -0.3280     0.7213  -0.455 0.651847    
#> L(IDE, 2)     1.4079     0.5520   2.550 0.014803 *  
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> Residual standard error: 0.0191 on 39 degrees of freedom
#> Multiple R-squared:  0.988,  Adjusted R-squared:  0.9843 
#> F-statistic: 266.8 on 12 and 39 DF,  p-value: < 2.2e-16
```

Then we can estimate the UECM (Unrestricted Error Correction Model) of
the underlying ARDL(3,1,3,2).

``` r
uecm_3132 <- uecm(ardl_3132)
summary(uecm_3132)
#> 
#> Time series regression with "zooreg" data:
#> Start = 1974 Q4, End = 1987 Q3
#> 
#> Call:
#> dynlm::dynlm(formula = full_formula, data = data, start = start, 
#>     end = end)
#> 
#> Residuals:
#>       Min        1Q    Median        3Q       Max 
#> -0.029939 -0.008856 -0.002562  0.008190  0.072577 
#> 
#> Coefficients:
#>              Estimate Std. Error t value Pr(>|t|)    
#> (Intercept)   2.62019    0.56777   4.615 4.19e-05 ***
#> L(LRM, 1)    -0.41685    0.09166  -4.548 5.15e-05 ***
#> L(LRY, 1)     0.41538    0.11761   3.532  0.00108 ** 
#> L(IBO, 1)    -1.89172    0.39111  -4.837 2.09e-05 ***
#> L(IDE, 1)     1.20534    0.44690   2.697  0.01028 *  
#> d(L(LRM, 1)) -0.26394    0.10192  -2.590  0.01343 *  
#> d(L(LRM, 2))  0.26867    0.10213   2.631  0.01214 *  
#> d(LRY)        0.67280    0.13116   5.129 8.32e-06 ***
#> d(IBO)       -1.07852    0.32170  -3.353  0.00179 ** 
#> d(L(IBO, 1))  0.70701    0.46874   1.508  0.13953    
#> d(L(IBO, 2))  0.99468    0.39251   2.534  0.01540 *  
#> d(IDE)        0.12546    0.55445   0.226  0.82216    
#> d(L(IDE, 1)) -1.40786    0.55204  -2.550  0.01480 *  
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> Residual standard error: 0.0191 on 39 degrees of freedom
#> Multiple R-squared:  0.7458, Adjusted R-squared:  0.6676 
#> F-statistic: 9.537 on 12 and 39 DF,  p-value: 3.001e-08
```

And also the RECM (Restricted Error Correction Model) of the underlying
ARDL(3,1,3,2), allowing the constant to join the long-run relationship
(case 2), instead of the short-run (case 3).

``` r
recm_3132 <- recm(uecm_3132, case = 2)
summary(recm_3132)
#> 
#> Time series regression with "zooreg" data:
#> Start = 1974 Q4, End = 1987 Q3
#> 
#> Call:
#> dynlm::dynlm(formula = full_formula, data = data, start = start, 
#>     end = end)
#> 
#> Residuals:
#>       Min        1Q    Median        3Q       Max 
#> -0.029939 -0.008856 -0.002562  0.008190  0.072577 
#> 
#> Coefficients:
#>              Estimate Std. Error t value Pr(>|t|)    
#> d(L(LRM, 1)) -0.26394    0.09008  -2.930 0.005405 ** 
#> d(L(LRM, 2))  0.26867    0.09127   2.944 0.005214 ** 
#> d(LRY)        0.67280    0.11591   5.805 7.03e-07 ***
#> d(IBO)       -1.07852    0.30025  -3.592 0.000837 ***
#> d(L(IBO, 1))  0.70701    0.44359   1.594 0.118300    
#> d(L(IBO, 2))  0.99468    0.36491   2.726 0.009242 ** 
#> d(IDE)        0.12546    0.48290   0.260 0.796248    
#> d(L(IDE, 1)) -1.40786    0.48867  -2.881 0.006160 ** 
#> ect          -0.41685    0.07849  -5.311 3.63e-06 ***
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> Residual standard error: 0.01819 on 43 degrees of freedom
#>   (0 observations deleted due to missingness)
#> Multiple R-squared:  0.7613, Adjusted R-squared:  0.7113 
#> F-statistic: 15.24 on 9 and 43 DF,  p-value: 9.545e-11
```

Let’s test if there is a long-run levels relationship (cointegration)
using the **bounds test** from *Pesaran et al. (2001)*.

``` r
# The bounds F-test (under the case 2) rejects the NULL hypothesis (let's say, assuming alpha = 0.01) with p-value = 0.004418.
bounds_f_test(ardl_3132, case = 2)
#> 
#>  Bounds F-test (Wald) for no cointegration
#> 
#> data:  d(LRM) ~ L(LRM, 1) + L(LRY, 1) + L(IBO, 1) + L(IDE, 1) + d(L(LRM,     1)) + d(L(LRM, 2)) + d(LRY) + d(IBO) + d(L(IBO, 1)) + d(L(IBO,     2)) + d(IDE) + d(L(IDE, 1))
#> F = 5.1168, p-value = 0.004418
#> alternative hypothesis: Possible cointegration
#> null values:
#>    k    T 
#>    3 1000

# The bounds t-test (under the case 3) rejects the NULL hypothesis (let's say, assuming alpha = 0.01) with p-value = 0.005538.
# We also provide the critical value bounds for alpha = 0.01.
tbounds <- bounds_t_test(uecm_3132, case = 3, alpha = 0.01)
tbounds
#> 
#>  Bounds t-test for no cointegration
#> 
#> data:  d(LRM) ~ L(LRM, 1) + L(LRY, 1) + L(IBO, 1) + L(IDE, 1) + d(L(LRM,     1)) + d(L(LRM, 2)) + d(LRY) + d(IBO) + d(L(IBO, 1)) + d(L(IBO,     2)) + d(IDE) + d(L(IDE, 1))
#> t = -4.5479, Lower-bound I(0) = -3.4430, Upper-bound I(1) = -4.3799,
#> p-value = 0.005538
#> alternative hypothesis: Possible cointegration
#> null values:
#>    k    T 
#>    3 1000

# Here is a more clear view of the main results.
tbounds$tab
#>   statistic Lower-bound I(0) Upper-bound I(1) alpha     p.value
#> t -4.547939        -3.442978        -4.379886  0.01 0.005538316
```

Here we have the short-run and the long-run multipliers (with standard
errors, t-statistics and p-values).

``` r
multipliers(ardl_3132, type = "sr")
#>          Term   Estimate Std. Error    t value     Pr(>|t|)
#> 1 (Intercept)  2.6201916  0.5677679  4.6148990 4.186867e-05
#> 2         LRY  0.6727993  0.1311638  5.1294603 8.317401e-06
#> 3         IBO -1.0785180  0.3217011 -3.3525465 1.790030e-03
#> 4         IDE  0.1254643  0.5544522  0.2262852 8.221614e-01
multipliers(ardl_3132)
#>          Term   Estimate Std. Error   t value     Pr(>|t|)
#> 1 (Intercept)  6.2856579  0.7719160  8.142930 6.107445e-10
#> 2         LRY  0.9964676  0.1239310  8.040503 8.358472e-10
#> 3         IBO -4.5381160  0.5202961 -8.722180 1.058619e-10
#> 4         IDE  2.8915201  0.9950853  2.905801 6.009239e-03
```

We can also estimate and visualize the delay multipliers along with
their standard errors.

``` r
mult15 <- multipliers(ardl_3132, type = 15, se = TRUE)
plot_delay(mult15, interval = 0.95)
```

<img src="man/figures/README-delay-multipliers-1.png" width="100%" />

Now let’s graphically check the estimated long-run relationship
(cointegrating equation) against the dependent variable `LRM`.

``` r
ce <- coint_eq(ardl_3132, case = 2)
```

``` r
plot_lr(ardl_3132, coint_eq = ce, show.legend = TRUE)
```

<img src="man/figures/README-lr-plot-1.png" width="100%" />

Forecasting using an `ardl` model is simple with the `predict()`
function, which dispatches to `predict.ardl()`, a method created
specifically for these models in this package.

``` r
#load zoo to create the zoo dataset
library(zoo)
#> Warning: package 'zoo' was built under R version 4.4.3
#> 
#> Attaching package: 'zoo'
#> The following objects are masked from 'package:base':
#> 
#>     as.Date, as.Date.numeric
# Create new observations
new_values <- zoo(data.frame(
        LRY=c(6.073861, 6.069912, 6.075906, 6.081180, 6.098292),
        IBO=c(0.1182830, 0.1098737, 0.1098398, 0.1193263, 0.1127251),
        IDE=c(0.07293421, 0.06980752, 0.07280321, 0.07112329, 0.07202889)
    ), as.yearqtr(c("1987 4", "1988 1", "1988 2", "1988 3", "1988 4"), format = "%Y %q"))
# Forecast using the new data
predicted_values <- predict(ardl_3132, new_values)

plot(denmark$LRM, xlim = c(start(denmark), end(predicted_values)))
lines(predicted_values, col = "red")
```

<img src="man/figures/README-Predict-ardl-1.png" width="100%" />

## Ease of use

Let’s see what it takes to build the above ARDL(3,1,3,2) model.

**Using the `ARDL` package (literally one line of code):**

``` r
ardl_model <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
```

**Without the `ARDL` package:**</br> *(Using the `dynlm` package,
because striving with the `lm` function would require extra data
transformation to behave like time-series)*

``` r
library(dynlm)
#> Warning: package 'dynlm' was built under R version 4.4.3

dynlm_ardl_model <- dynlm(LRM ~ L(LRM, 1) + L(LRM, 2) + L(LRM, 3) + LRY + L(LRY, 1) +
                           IBO + L(IBO, 1) + L(IBO, 2) + L(IBO, 3) +
                           IDE + L(IDE, 1) + L(IDE, 2), data = denmark)
```

``` r
identical(ardl_model$coefficients, dynlm_ardl_model$coefficients)
#> [1] TRUE
```

An ARDL model has a relatively simple structure, although the difference
in typing effort is noticeable.

Not to mention the complex transformation for an ECM. The extra typing
is the least of your problems trying to do this. First you would need to
figure out the exact structure of the model!

**Using the `ARDL` package (literally one line of code):**

``` r
uecm_model <- uecm(ardl_model)
```

**Without the `ARDL` package:**

``` r
dynlm_uecm_model <- dynlm(d(LRM) ~ L(LRM, 1) + L(LRY, 1) + L(IBO, 1) +
                                   L(IDE, 1) + d(L(LRM, 1)) + d(L(LRM, 2)) +
                                   d(LRY) + d(IBO) + d(L(IBO, 1)) + d(L(IBO, 2)) +
                                   d(IDE) + d(L(IDE, 1)), data = denmark)
```

``` r
identical(uecm_model$coefficients, dynlm_uecm_model$coefficients)
#> [1] TRUE
```

## References

Natsiopoulos, Kleanthis, & Tzeremes, Nickolaos G. (2022). ARDL bounds
test for cointegration: Replicating the Pesaran et al. (2001) results
for the UK earnings equation using R. *Journal of Applied Econometrics*,
37(5), 1079-1090. <https://doi.org/10.1002/jae.2919>

Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds testing
approaches to the analysis of level relationships. *Journal of Applied
Econometrics*, 16(3), 289-326

#' Unrestricted ECM regression
#'
#' \code{uecm} is a generic function used to construct Unrestricted Error
#' Correction Models (UECM). The function invokes two different
#' \code{\link[utils]{methods}}. The default method works exactly like
#' \code{\link{ardl}}. The other method requires an object of
#' \code{\link[base]{class}} 'ardl'. Both methods create the conditional UECM,
#' which is the UECM of the underlying ARDL.
#'
#' @inherit ardl details
#'
#' @return \code{uecm} returns an object of \code{\link[base]{class}}
#'   \code{c("uecm", "dynlm", "lm")}. In addition, attributes 'order', 'data',
#'   'parsed_formula' and 'full_formula' are provided.
#'
#' @section Mathematical Formula: The formula of an Unrestricted ECM conditional
#'   to an \eqn{ARDL(p,q_{1},\dots,q_{k})}{ARDL(p,q1,...,qk)} is: \deqn{\Delta
#'   y_{t} = c_{0} + c_{1}t + \pi_{y}y_{t-1} + \sum_{j=1}^{k}\pi_{j}x_{j,t-1} +
#'   \sum_{i=1}^{p-1}\psi_{y,i}\Delta y_{t-i} +
#'   \sum_{j=1}^{k}\sum_{l=1}^{q_{j}-1} \psi_{j,l}\Delta x_{j,t-l} +
#'   \sum_{j=1}^{k}\omega_{j}\Delta x_{j,t} + \epsilon_{t}}
#'   \deqn{\psi_{j,l} = 0 \;\; \forall \;\; q_{j} \leq 1, \;\;\;\;\; \psi_{y,i}
#'   = 0 \;\; if \;\; p = 1}
#'   In addition, \eqn{x_{j,t-1}} and \eqn{\Delta x_{j,t}}
#'   cancel out becoming \eqn{x_{j,t} \;\; \forall \;\; q_{j} = 0}
#'
#' @seealso \code{\link{ardl}} \code{\link{recm}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords models ts
#' @export
#' @examples
#' data(denmark)
#'
#' ## Estimate the UECM, conditional to it's underlying ARDL(3,1,3,2) -----
#'
#' # Indirectly
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' uecm_3132 <- uecm(ardl_3132)
#'
#' # Directly
#' uecm_3132_ <- uecm(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' identical(uecm_3132, uecm_3132_)
#' summary(uecm_3132)
#'
#' ## Post-estimation testing ---------------------------------------------
#'
#' library(lmtest) # for bgtest(), bptest(), and resettest()
#' library(tseries) # for jarque.bera.test()
#' library(strucchange) # for efp(), and sctest()
#'
#' # Breusch-Godfrey test for higher-order serial correlation
#' bgtest(uecm_3132, order = 4)
#'
#' # Breusch-Pagan test against heteroskedasticity
#' bptest(uecm_3132)
#'
#' # Ramsey's RESET test for functional form
#' \dontrun{
#'     # This produces an error.
#'     # resettest() cannot use data of class 'zoo' such as the 'denmark' data
#'     # used to build the original model
#'     resettest(uecm_3132, type = c("regressor"))
#' }
#'
#' uecm_3132_lm <- to_lm(uecm_3132, data_class = "ts")
#' resettest(uecm_3132_lm, power = 2)
#'
#' # Jarque-Bera test for normality
#' jarque.bera.test(residuals(uecm_3132))
#'
#' # CUSUM test for structural change detection
#' \dontrun{
#'     # This produces an error.
#'     # efp() does not understand special functions such as "d()" and "L()"
#'     efp(uecm_3132$full_formula, data = uecm_3132$model)
#' }
#'
#' uecm_3132_lm_names <- to_lm(uecm_3132, fix_names = TRUE)
#' fluctuation <- efp(uecm_3132_lm_names$full_formula,
#'                    data = uecm_3132_lm_names$model)
#' sctest(fluctuation)
#' plot(fluctuation)
#'

uecm <- function(...) {
    UseMethod("uecm")
}

#' @rdname uecm
#'
#' @param object An object of \code{\link[base]{class}} 'ardl'.
#'
#' @export
#'

uecm.ardl <- function(object, ...) {

    parsed_formula <- object$parsed_formula
    order <- object$order
    data <- object$data

    start <- start(object)
    end <- end(object)

    uecm_formula <- build_uecm_formula(parsed_formula = parsed_formula, order = order)
    full_formula <- stats::formula(uecm_formula$full)

    uecm_model <- dynlm::dynlm(full_formula, data = data, start = start, end = end)
    # for model compatibility in the global env
    attr(uecm_model$terms, ".Environment") <- .GlobalEnv
    attr(attr(uecm_model$model, "terms"), ".Environment") <- .GlobalEnv
    attr(full_formula, ".Environment") <- .GlobalEnv

    uecm_model$order <- order
    uecm_model$data <- data
    uecm_model$parsed_formula <- parsed_formula
    uecm_model$full_formula <- full_formula

    uecm_model <- structure(
        uecm_model,
        class = c("uecm", class(uecm_model))
    )


    return(uecm_model)
}

#' @rdname uecm
#'
#' @param order A specification of the order of the underlying ARDL model (e.g.,
#'   for the UECM of an ARDL(1,0,2) model it should be \code{order = c(1,0,2)}).
#'   A numeric vector of the same length as the total number of variables
#'   (excluding the fixed ones, see 'Details'). It should only contain positive
#'   integers or 0. An integer could be provided if all variables are of the
#'   same order.
#' @inheritParams ardl
#'
#' @export
#'

uecm.default <- function(formula, data, order, start = NULL, end = NULL, ...) {

    if (!any(c("ts", "zoo", "zooreg") %in% class(data))) {
        data <- stats::ts(data, start = 1, end = nrow(data), frequency = 1)
    }
    parsed_formula <- parse_formula(formula = formula, colnames_data = colnames(data))
    order <- parse_order(orders = order, order_name = "order",
                         var_names = parsed_formula$z_part$var, kz = parsed_formula$kz)

    uecm_formula <- build_uecm_formula(parsed_formula = parsed_formula, order = order)
    full_formula <- stats::formula(uecm_formula$full)

    uecm_model <- dynlm::dynlm(full_formula, data = data, start = start, end = end, ...)
    # for model compatibility in the global env
    attr(uecm_model$terms, ".Environment") <- .GlobalEnv
    attr(attr(uecm_model$model, "terms"), ".Environment") <- .GlobalEnv
    attr(full_formula, ".Environment") <- .GlobalEnv

    uecm_model$order <- order
    uecm_model$data <- data
    uecm_model$parsed_formula <- parsed_formula
    uecm_model$full_formula <- full_formula

    uecm_model <- structure(
        uecm_model,
        class = c("uecm", class(uecm_model))
    )

    return(uecm_model)
}

test_that("correct results using ardl or direct inputs", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    uecm_model <- uecm(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))

    expect_equal(uecm_model, uecm(ardl_model))

    # check if non time-series data work
    ardl_model_df <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                          data = data.frame(PSS2001), start = 9,
                          order=c(6,1,5,4,5))
    arduecm_model_df <- uecm(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                             data = data.frame(PSS2001), start = 9,
                             order=c(6,1,5,4,5))

    expect_equal(uecm_model$coefficients, uecm(ardl_model_df)$coefficients)
    expect_equal(uecm_model$coefficients, arduecm_model_df$coefficients)
})

test_that("correct results case 2 & 3", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    expected_mult <- data.frame(c("(Intercept)", "Prod", "UR", "Wedge", "Union"),
                                c(2.70108265, 1.06332501, -0.10460915, -0.94285059, 1.48067052),
                                c(0.2414848597, 0.0499903071, 0.0336700352, 0.2649817179, 0.3106107337),
                                c(11.18530847, 21.27062375, -3.10689148, -3.55817222, 4.76696509))
    colnames(expected_mult) <- c("Term", "Estimate", "Std. Error", "t value")

    expect_equal(multipliers(ardl_model)[,-5], expected_mult)

    ardl_model <- ardl(Union ~ Prod + UR + Wedge | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(5,0,5,1))

    expect_equal(round(multipliers(ardl_model)[,"Pr(>|t|)"],6), c(0.013860, 0.768666, 0.919328, 0.440212))
})

test_that("correct results case 1", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    expected_mult <- data.frame(c("Prod", "UR", "Wedge", "Union"),
                                c(0.4047793, -0.6813798, -10.7233397, -2.6065488),
                                c(2.1361117, 1.8758181, 23.4065331, 3.3974199),
                                c(0.189493512, -0.363244080, -0.458134473, -0.767214197))
    colnames(expected_mult) <- c("Term", "Estimate", "Std. Error", "t value")

    expect_equal(multipliers(ardl_model)[,-5], expected_mult)
    expect_equal(round(multipliers(ardl_model)[,"Pr(>|t|)"],6), c(0.850205, 0.717418, 0.648146, 0.4453))
})

test_that("equivalence long-run and sum of interim", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    mult_lr <- multipliers(ardl_model, type = "lr")
    mult_inter200 <- multipliers(ardl_model, type = 200)

    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='(Intercept)'],5), round(sum(mult_inter200$'(Intercept)'$Delay),5))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='Prod'],4), round(sum(mult_inter200$Prod$Delay),4))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='UR'],5), round(sum(mult_inter200$UR$Delay),5))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='Wedge'],5), round(sum(mult_inter200$Wedge$Delay),5))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='Union'],5), round(sum(mult_inter200$Union$Delay),5))

    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union + trend(w, scale=FALSE)| D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    mult_lr <- multipliers(ardl_model, type = "lr")
    mult_inter200 <- multipliers(ardl_model, type = 200)

    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='(Intercept)'],5), round(mult_inter200$'(Intercept)'$Interim[201],5))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='trend(w, scale = FALSE)'],5), round(mult_inter200$'trend(w, scale = FALSE)'$Interim[201],5))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='Prod'],4), round(mult_inter200$'Prod'$Interim[201],4))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='UR'],5), round(mult_inter200$'UR'$Interim[201],5))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='Wedge'],5), round(mult_inter200$'Wedge'$Interim[201],5))
    expect_equal(round(mult_lr$Estimate[mult_lr$Term=='Union'],4), round(mult_inter200$'Union'$Interim[201],4))
})

test_that("correct s.e. for delay multipliers", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    mult <- multipliers(ardl_model, type = 3, se = TRUE)
    expected_se_c <- c(0.1554, 0.0686, 0.0658, 0.0654)
    expected_se_ur <- c(0.0083, 0.0108, 0.0126, 0.0136)

    expect_equal(round(mult$`(Intercept)`[,"Std. Error Delay"], 4), expected_se_c)
    expect_equal(round(mult$UR[,"Std. Error Delay"], 4), expected_se_ur)

    # no intercept
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1| D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    mult <- multipliers(ardl_model, type = 3, se = TRUE)
    expected_se_prod <- c(0.1021, 0.0602, 0.0337, 0.0392)
    expected_se_ur <- c(0.0091, 0.0111, 0.0131, 0.0137)

    expect_equal(round(mult$Prod[,"Std. Error Delay"], 4), expected_se_prod)
    expect_equal(round(mult$UR[,"Std. Error Delay"], 4), expected_se_ur)

    # intercept and trend
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union + trend(w)| D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    mult <- multipliers(ardl_model, type = 3, se = TRUE)
    expected_se_c <- c(0.2006, 0.0774, 0.0638, 0.0581)
    expected_se_trend <- c(0.0014, 0.0005, 0.0003, 0.0001)
    expected_se_ur <- c(0.0083, 0.0109, 0.0127, 0.0137)

    expect_equal(round(mult$`(Intercept)`[,"Std. Error Delay"], 4), expected_se_c)
    expect_equal(round(mult$`trend(w)`[,"Std. Error Delay"], 4), expected_se_trend)
    expect_equal(round(mult$UR[,"Std. Error Delay"], 4), expected_se_ur)

    # include 0 order
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,0,5,4,5))
    mult <- multipliers(ardl_model, type = 3, se = TRUE)
    expected_se_c <- c(0.1426, 0.0700, 0.0677, 0.0652)
    expected_se_ur <- c(0.0083, 0.0107, 0.0125, 0.0128)

    expect_equal(round(mult$`(Intercept)`[,"Std. Error Delay"], 4), expected_se_c)
    expect_equal(round(mult$UR[,"Std. Error Delay"], 4), expected_se_ur)
})

test_that("type='sr' equals type=0", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))

    expect_equal(multipliers(ardl_model, type = "sr"), multipliers(ardl_model, type = 0))
})

test_that("equal results with ardl or uecm input", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    ardl_model_c <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                         data = PSS2001, start = c(1972, 01),
                         order=c(6,1,5,4,5))
    ardl_model_ct <- ardl(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                          data = PSS2001, start = c(1972, 01),
                          order=c(6,1,5,4,5))

    expect_equal(multipliers(ardl_model), multipliers(uecm(ardl_model)))
    expect_equal(multipliers(ardl_model, type = "sr"), multipliers(uecm(ardl_model), type = 0))
    expect_equal(multipliers(ardl_model, type = 2), multipliers(uecm(ardl_model), type = 2))

    expect_equal(multipliers(ardl_model_c), multipliers(uecm(ardl_model_c)))
    expect_equal(multipliers(ardl_model_c, type = "sr"), multipliers(uecm(ardl_model_c), type = 0))
    expect_equal(multipliers(ardl_model_c, type = 2), multipliers(uecm(ardl_model_c), type = 2))

    expect_equal(multipliers(ardl_model_ct), multipliers(uecm(ardl_model_ct)))
    expect_equal(multipliers(ardl_model_ct, type = "sr"), multipliers(uecm(ardl_model_ct), type = 0))
    expect_equal(multipliers(ardl_model_ct, type = 2), multipliers(uecm(ardl_model_ct), type = 2))
})

test_that("correct results when input is uecm, type=0 and there is a 0 order", {
#Ensure that the results are correct when the input is uecm and the order includes a 0
uecm_model_0 <- uecm(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                     data = PSS2001, start = c(1972, 01),
                     order=c(6,1,0,4,0))
ardl_model_0 <- ardl(uecm_model_0)

expect_equal(multipliers(uecm_model_0, type = "sr"), multipliers(ardl_model_0, type = "sr"))
})

test_that("wrong type causes expected error", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))

    expect_error(multipliers(ardl_model, type = "aa"),
                 "'type' should be one of 'lr', 'sr' or a number between 0 and 200")
    expect_error(multipliers(uecm(ardl_model), type = "aa"),
                 "'type' should be one of 'lr', 'sr' or a number between 0 and 200")
})

test_that("correct results using ardl or uecm inputs", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    uecm_model <- uecm(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    ardl_model_c <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                         data = PSS2001, start = c(1972, 01),
                         order=c(6,1,5,4,5))
    uecm_model_c <- uecm(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                         data = PSS2001, start = c(1972, 01),
                         order=c(6,1,5,4,5))
    ardl_model_ct <- ardl(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                          data = PSS2001, start = c(1972, 01),
                          order=c(6,1,5,4,5))
    uecm_model_ct <- uecm(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                          data = PSS2001, start = c(1972, 01),
                          order=c(6,1,5,4,5))

    expect_equal(recm(ardl_model, case=1), recm(uecm_model, case=1))

    expect_equal(recm(ardl_model_c, case=2), recm(ardl_model_c, case=2))
    expect_equal(recm(ardl_model_c, case=3), recm(ardl_model_c, case=3))

    expect_equal(recm(uecm_model_ct, case=4), recm(uecm_model_ct, case=4))
    expect_equal(recm(uecm_model_ct, case=5), recm(uecm_model_ct, case=5))

    # check that ect is correct
    ect_uecm <- uecm_model$coefficients[1]
    names(ect_uecm) <- "ect"
    expect_equal(recm(ardl_model, case=1)$coefficients["ect"], ect_uecm)

    ect_uecm <- uecm_model_c$coefficients[2]
    names(ect_uecm) <- "ect"
    expect_equal(recm(ardl_model_c, case=2)$coefficients["ect"], ect_uecm)
    expect_equal(recm(ardl_model_c, case=3)$coefficients["ect"], ect_uecm)

    ect_uecm <- uecm_model_ct$coefficients[3]
    names(ect_uecm) <- "ect"
    expect_equal(recm(ardl_model_ct, case=4)$coefficients["ect"], ect_uecm)
    expect_equal(recm(ardl_model_ct, case=5)$coefficients["ect"], ect_uecm)
})

test_that("correct results when order includes 0", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,0,5,4,5))
    uecm_model_c <- uecm(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                         data = PSS2001, start = c(1972, 01),
                         order=c(6,0,5,4,5))
    ardl_model_ct <- ardl(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                          data = PSS2001, start = c(1972, 01),
                          order=c(6,0,5,4,5))
    uecm_model_ct <- uecm(ardl_model_ct)

    expect_equal(recm(ardl_model, case=1)$coefficients["ect"], uecm(ardl_model)$coeff["L(w, 1)"], ignore_attr = TRUE)
    expect_equal(recm(uecm_model_c, case=2)$coefficients["ect"], uecm_model_c$coeff["L(w, 1)"], ignore_attr = TRUE)
    expect_equal(recm(ardl_model_ct, case=4)$coefficients["ect"], uecm(ardl_model_ct)$coeff["L(w, 1)"], ignore_attr = TRUE)
    expect_equal(names(recm(uecm_model_ct, case=5)$coefficients)[-24], names(uecm_model_ct$coefficients)[-c(3:7)])
})

test_that("correct results", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))

    expect_equal(coint_eq(ardl_model, case = 2), coint_eq(uecm(ardl_model), case = 2))
    expect_equal(coint_eq(ardl_model, case = 3), coint_eq(recm(ardl_model, case = 3)))
})

test_that("case 1", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))

    expect_equal(coint_eq(ardl_model, case = 1), coint_eq(uecm(ardl_model), case = 1))
})

test_that("case 4 & case 5 & linear time trend", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    ardl_model_scale <- ardl(w ~ Prod + UR + Wedge + Union + trend(w, scale=FALSE) | D7475 + D7579,
                             data = PSS2001, start = c(1972, 01),
                             order=c(6,1,5,4,5))
    ardl_model_time <- ardl(w ~ Prod + UR + Wedge + Union + time(w) | D7475 + D7579,
                            data = ts(PSS2001), start = 9,
                            order=c(6,1,5,4,5))

    expect_equal(coint_eq(ardl_model, case = 4), coint_eq(uecm(ardl_model), case = 4))
    expect_equal(coint_eq(ardl_model, case = 5), coint_eq(uecm(ardl_model), case = 5))
    expect_equal(coint_eq(ardl_model_scale, case = 4), coint_eq(uecm(ardl_model_scale), case = 4))
    expect_equal(coint_eq(ardl_model_time, case = 4), coint_eq(uecm(ardl_model_time), case = 4))
})

test_that("verify prediction results all cases", {
    # Using data.frame as the test case, time series objects tested separately
    newdata <- data.frame("Prod" = c(1,2,3), "UR"= c(4,5,6), "Wedge"=c(7,8,9), "Union" = c(11,22,33),
                          "D7475"= c(0,1,0), "D7579"=c(1,0,0))
    # Case 1: No intercept and no trend
    case_i <- ardl(w ~ Prod + UR + Wedge + Union - 1 | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001)
    # Case 2 & 3: Restricted intercept and no trend / Unrestricted intercept and no trend
    case_ii_iii <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001)
    # Case 4 & 5: Unrestricted intercept and restricted trend / Unrestricted intercept and unrestricted trend
    case_iv_v <- ardl(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001)
    # Alternative trend methods
    case_iv_v_time <- ardl(w ~ Prod + UR + Wedge + Union + time(w) | D7475 + D7579, order =c(3,0,1,4,3), data = ts(PSS2001))
    case_iv_v_trend_scale <- ardl(w ~ Prod + UR + Wedge + Union + trend(w, scale = FALSE) | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001)
    result_case_i <- predict.ardl(case_i, newdata)
    result_case_ii_iii <- predict.ardl(case_ii_iii, newdata)
    result_case_iv_v <- predict.ardl(case_iv_v, newdata)
    result_case_iv_v_time <- predict.ardl(case_iv_v_time, newdata)
    result_case_iv_v_trend_scale <- predict.ardl(case_iv_v_trend_scale, newdata)
    # These are externally calculated
    case_i <- setNames(c(-13.1721023693382, -40.3134240474061, -57.8052029786566), seq(1, 3))
    case_ii_iii <- setNames(c(-7.40819925882942, -24.8395014334613, -29.1887358923117), seq(1, 3))
    case_iv_v <- setNames(c(-6.61455842285156, -23.5381169607463, -27.919945003499), seq(1, 3))
    # Note: Ignores small numeric differences
    expect_equal(case_i, result_case_i, tolerance=1e-3)
    expect_equal(case_ii_iii, result_case_ii_iii, tolerance=1e-3)
    expect_equal(case_iv_v, result_case_iv_v, tolerance=1e-3)
    expect_equal(case_iv_v, result_case_iv_v_time, tolerance=1e-3)
    expect_equal(case_iv_v, result_case_iv_v_trend_scale, tolerance=1e-3)
})

test_that("unsupported newdata object", {
    newdata <- matrix(c(1,2,3,  4,5,6,  7,8,9,  11,22,33,  0,1,0,  1,0,0),
                      nrow = 3,
                      ncol = 6
                      )
    case_i <- ardl(w ~ Prod + UR + Wedge + Union - 1 | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001)
    expect_error(predict.ardl(case_i, newdata))
})

test_that("verify equivalence of different newdata types", {
    library(zoo)
    model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001)
    # dataframe
    newdata_df <- data.frame(
        "Prod" = c(1,2,3), "UR"= c(4,5,6), "Wedge"=c(7,8,9), "Union" = c(11,22,33),
        "D7475"= c(0,1,0), "D7579"=c(1,0,0))
    result_df <- predict.ardl(model, newdata_df)
    # setting frequency and start timestamp as expected for time series objects
    # ts object
    newdata_ts_timestamp <- ts(newdata_df, start = c(1998, 1), frequency = 4)
    result_ts_timestamp <- predict.ardl(model, newdata_ts_timestamp)
    expect_equal(unname(result_df), c(unname(result_ts_timestamp)))
    # zoo object
    dates <- as.yearqtr(c("1998 1", "1998 2", "1998 3"), format = "%Y %q")
    newdata_zoo_timestamp <- zoo(newdata_df, dates)
    result_zoo_timestamp <- predict.ardl(model, newdata_zoo_timestamp)
    expect_equal(unname(result_df), c(coredata(result_zoo_timestamp)))
    # zooreg object
    newdata_zooreg_timestamp <- zooreg(newdata_df, start = c(1998, 1), frequency = 4)
    result_zooreg_timestamp <- predict.ardl(model, newdata_zooreg_timestamp)
    expect_equal(unname(result_df), c(coredata(result_zooreg_timestamp)))
})

test_that("check integrity of newdata and ardl object", {
    # NA values
    newdata <- data.frame("Prod" = c(NA,2,3), "UR"= c(4,5,6), "Wedge"=c(7,8,9), "Union" = c(11,22,33),
                          "D7475"= c(0,1,0), "D7579"=c(1,0,0))
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001)
    expect_error(predict.ardl(ardl_model, newdata))
    # Empty newdata
    expect_error(predict.ardl(ardl_model))
    # Not ARDL model
    newdata <- data.frame("Prod" = c(1,2,3), "UR"= c(4,5,6), "Wedge"=c(7,8,9), "Union" = c(11,22,33),
                          "D7475"= c(0,1,0), "D7579"=c(1,0,0))
    uecm_model <- uecm(ardl_model)
    expect_error(predict.ardl(uecm_model, newdata))
    # Missing cols
    newdata <- data.frame("UR"= c(4,5,6), "Wedge"=c(7,8,9), "Union" = c(11,22,33),
                          "D7475"= c(0,1,0), "D7579"=c(1,0,0))
    expect_error(predict.ardl(ardl_model, newdata))
    # Give newdata in another order
    newdata <- data.frame("UR"= c(4,5,6), "Prod" = c(1,2,3), "Wedge"=c(7,8,9), "Union" = c(11,22,33),
                          "D7475"= c(0,1,0), "D7579"=c(1,0,0))
    expect_error(predict.ardl(ardl_model, newdata))
})

test_that("check time series objects", {
    PSS2001_ts <- ts(PSS2001, start = c(1970, 1), frequency = 4)
    PSS2001_zoo <- as.zoo(window(PSS2001, start = as.yearqtr("1970 1", format = "%Y %q"), end = as.yearqtr("1997 4", format = "%Y %q")))
    PSS2001_zooreg <- zooreg(PSS2001, start = c(1970, 1), frequency = 4)
    
    ardl_ts <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001_ts)
    ardl_zoo <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001_zoo)
    ardl_zooreg <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001_zooreg)
    newdata <- data.frame("Prod" = c(1,2,3), "UR"= c(4,5,6), "Wedge"=c(7,8,9), "Union" = c(11,22,33),
                          "D7475"= c(0,1,0), "D7579"=c(1,0,0))
    # Check frequency is the same
    # PSS2001 has frequency = 4
    newdata_ts <- ts(newdata, start = c(1998, 1), frequency = 3)
    expect_error(predict.ardl(ardl_ts, newdata_ts))
    newdata_zoo <- zoo(newdata, frequency = 1)
    expect_error(predict.ardl(ardl_zoo, newdata_zoo))
    newdata_zooreg <- zooreg(newdata, start = c(1998, 1), frequency = 5)
    expect_error(predict.ardl(ardl_zooreg, newdata_zooreg))
    # Check time series are consecutive
    newdata_ts <- ts(newdata, start = c(1997, 3), frequency = 4)
    expect_error(predict.ardl(ardl_ts, newdata_ts))
    dates <- as.yearqtr(c("1998 2", "1998 3", "1998 4"), format = "%Y %q")
    newdata_zoo <- zoo(newdata, dates)
    expect_error(predict.ardl(ardl_zoo, newdata_zoo))
    newdata_zooreg <- zooreg(newdata, start = c(1995, 2), frequency = 4)
    expect_error(predict.ardl(ardl_zooreg, newdata_zooreg))
    # As a double check we'll carry out basic check on monthly frequency
    PSS2001_monthly <- ts(coredata(PSS2001), start = c(1970, 1), frequency = 12)
    ardl_monthly <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579, order =c(3,0,1,4,3), data = PSS2001_monthly)
    # Incorrect frequency
    newdata_ts <- ts(newdata, start = c(1979, 5), frequency = 3)
    expect_error(predict.ardl(ardl_monthly, newdata_ts))
    # Not consecutive
    newdata_ts <- ts(newdata, start = c(1980, 1), frequency = 4)
    expect_error(predict.ardl(ardl_monthly, newdata_ts))
})

test_that("From ardl to lm", {
    ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
    ardl_3132_lm <- to_lm(ardl_3132)

    expect_equal(summary(ardl_3132)$coefficients,
                 summary(ardl_3132_lm)$coefficients, ignore_attr = TRUE)
    expect_equal(ardl_3132$residuals, ardl_3132_lm$residuals)
    expect_equal(ardl_3132$effects, ardl_3132_lm$effects)
    expect_equal(ardl_3132$rank, ardl_3132_lm$rank)
    expect_equal(ardl_3132$fitted.values, ardl_3132_lm$fitted.values)
    expect_equal(ardl_3132$assign, ardl_3132_lm$assign)
    expect_equal(ardl_3132$qr, ardl_3132_lm$qr, ignore_attr= TRUE)
    expect_equal(ardl_3132$df.residual, ardl_3132_lm$df.residual)
    expect_equal(ardl_3132$xlevels, ardl_3132_lm$xlevels)
    expect_equal(ardl_3132$model, ardl_3132_lm$model, ignore_attr= TRUE)

    # Check data_class = "ts"
    ardl_3132_lm_ts <- to_lm(ardl_3132, data_class = "ts")
    expect_equal(ardl_3132_lm$coefficients, ardl_3132_lm_ts$coefficients)
})

test_that("From uecm to lm", {
    uecm_3132 <- uecm(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
    uecm_3132_lm <- to_lm(uecm_3132)

    expect_equal(summary(uecm_3132)$coefficients,
                 summary(uecm_3132_lm)$coefficients, ignore_attr = TRUE)
    expect_equal(uecm_3132$residuals, uecm_3132_lm$residuals)
    expect_equal(uecm_3132$effects, uecm_3132_lm$effects)
    expect_equal(uecm_3132$rank, uecm_3132_lm$rank)
    expect_equal(uecm_3132$fitted.values, uecm_3132_lm$fitted.values)
    expect_equal(uecm_3132$assign, uecm_3132_lm$assign)
    expect_equal(uecm_3132$qr, uecm_3132_lm$qr, ignore_attr= TRUE)
    expect_equal(uecm_3132$df.residual, uecm_3132_lm$df.residual)
    expect_equal(uecm_3132$xlevels, uecm_3132_lm$xlevels)
    expect_equal(uecm_3132$model, uecm_3132_lm$model, ignore_attr= TRUE)

    # fix_names = TRUE
    uecm_3132_lm_names <- to_lm(uecm_3132, fix_names = TRUE)
    names_FALSE <- c("d.LRM", "L(LRM, 1)", "L(LRY, 1)", "L(IBO, 1)", "L(IDE, 1)",
                     "d(L(LRM, 1))", "d(L(LRM, 2))", "d(LRY)", "d(IBO)",
                     "d(L(IBO, 1))", "d(L(IBO, 2))", "d(IDE)", "d(L(IDE, 1))")
    names_TRUE <- c("d.LRM", "L.LRM.1", "L.LRY.1", "L.IBO.1", "L.IDE.1",
                    "d.L.LRM.1", "d.L.LRM.2", "d.LRY", "d.IBO", "d.L.IBO.1",
                    "d.L.IBO.2", "d.IDE", "d.L.IDE.1")
    expect_equal(names(uecm_3132_lm$model), names_FALSE)
    expect_equal(names(uecm_3132_lm_names$model), names_TRUE)
})

test_that("Exclude intercept when -1", {
    ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE -1, data = denmark, order = c(3,1,3,2))
    ardl_3132_lm <- to_lm(ardl_3132)

    expect_equal(summary(ardl_3132)$coefficients,
                 summary(ardl_3132_lm)$coefficients, ignore_attr = TRUE)
    expect_equal(ardl_3132$residuals, ardl_3132_lm$residuals)
    expect_equal(ardl_3132$effects, ardl_3132_lm$effects)
    expect_equal(ardl_3132$rank, ardl_3132_lm$rank)
    expect_equal(ardl_3132$fitted.values, ardl_3132_lm$fitted.values)
    expect_equal(ardl_3132$assign, ardl_3132_lm$assign)
    expect_equal(ardl_3132$qr, ardl_3132_lm$qr, ignore_attr= TRUE)
    expect_equal(ardl_3132$df.residual, ardl_3132_lm$df.residual)
    expect_equal(ardl_3132$xlevels, ardl_3132_lm$xlevels)
    expect_equal(ardl_3132$model, ardl_3132_lm$model, ignore_attr= TRUE)

    # check trend
    ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE + trend(LRM), data = denmark, order = c(3,1,3,2))
    ardl_3132_lm <- to_lm(ardl_3132)

    expect_equal(summary(ardl_3132)$coefficients,
                 summary(ardl_3132_lm)$coefficients, ignore_attr = TRUE)

    # check trend and no intercept
    ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE + trend(LRM,scale=FALSE) -1, data = denmark, order = c(3,1,3,2))
    ardl_3132_lm <- to_lm(ardl_3132)

    expect_equal(summary(ardl_3132)$coefficients,
                 summary(ardl_3132_lm)$coefficients, ignore_attr = TRUE)
})

test_that("From recm to lm", {
    # Case 2
    recm_3132 <- recm(uecm(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2)), case = 2)
    recm_3132_lm <- to_lm(recm_3132)

    expect_equal(summary(recm_3132)$coefficients,
                 summary(recm_3132_lm)$coefficients, ignore_attr = TRUE)
    expect_equal(recm_3132$residuals, recm_3132_lm$residuals)
    expect_equal(recm_3132$effects, recm_3132_lm$effects)
    expect_equal(recm_3132$rank, recm_3132_lm$rank)
    expect_equal(recm_3132$fitted.values, recm_3132_lm$fitted.values)
    expect_equal(recm_3132$assign, recm_3132_lm$assign)
    expect_equal(recm_3132$qr, recm_3132_lm$qr, ignore_attr= TRUE)
    expect_equal(recm_3132$df.residual, recm_3132_lm$df.residual)
    expect_equal(recm_3132$xlevels, recm_3132_lm$xlevels)
    expect_equal(recm_3132$model, recm_3132_lm$model, ignore_attr= TRUE)

    # Case 3
    recm_3132 <- recm(uecm(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2)), case = 3)
    recm_3132_lm <- to_lm(recm_3132)

    expect_equal(summary(recm_3132)$coefficients,
                 summary(recm_3132_lm)$coefficients, ignore_attr = TRUE)
    expect_equal(recm_3132$residuals, recm_3132_lm$residuals)
    expect_equal(recm_3132$effects, recm_3132_lm$effects)
    expect_equal(recm_3132$rank, recm_3132_lm$rank)
    expect_equal(recm_3132$fitted.values, recm_3132_lm$fitted.values)
    expect_equal(recm_3132$assign, recm_3132_lm$assign)
    expect_equal(recm_3132$qr, recm_3132_lm$qr, ignore_attr= TRUE)
    expect_equal(recm_3132$df.residual, recm_3132_lm$df.residual)
    expect_equal(recm_3132$xlevels, recm_3132_lm$xlevels)
    expect_equal(recm_3132$model, recm_3132_lm$model, ignore_attr= TRUE)

    # Case 1
    recm_3132 <- recm(uecm(LRM ~ LRY + IBO + IDE -1, data = denmark, order = c(3,1,3,2)), case = 1)
    recm_3132_lm <- to_lm(recm_3132)

    expect_equal(summary(recm_3132)$coefficients,
                 summary(recm_3132_lm)$coefficients, ignore_attr = TRUE)

    # Case 4
    recm_3132 <- recm(uecm(LRM ~ LRY + IBO + IDE + trend(LRM), data = denmark, order = c(3,1,3,2)), case = 4)
    recm_3132_lm <- to_lm(recm_3132)

    expect_equal(summary(recm_3132)$coefficients,
                 summary(recm_3132_lm)$coefficients, ignore_attr = TRUE)

    # Case 5
    recm_3132 <- recm(uecm(LRM ~ LRY + IBO + IDE + trend(LRM, scale=FALSE), data = denmark, order = c(3,1,3,2)), case = 5)
    recm_3132_lm <- to_lm(recm_3132)

    expect_equal(summary(recm_3132)$coefficients,
                 summary(recm_3132_lm)$coefficients, ignore_attr = TRUE)
})

test_that("Check dynlm models with lagged dependent variable", {
    dynlm_3132 <- dynlm::dynlm(L(LRM, 2) ~ -1 +L(LRM, 3) + LRY + L(LRY, 3) + IBO +
                            L(IBO, 4) + L(IBO, 5) + L(IBO, 6) + IDE + L(IDE, 3) +
                            L(IDE, 4), data=denmark)
    dynlm_3132_lm <- to_lm(dynlm_3132)

    expect_equal(summary(dynlm_3132)$coefficients,
                 summary(dynlm_3132_lm)$coefficients, ignore_attr = TRUE)
})

test_that("ardl model coefficients", {
    ardl61545 <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                      order = c(6,1,5,4,5), data = PSS2001, start = c(1972, 01))
    ardl_coeffs <- c(0.61914,
                     0.35330,0.08955,-0.19545,0.39035,-0.06366,0.19669,
                     0.31530,-0.07157,
                     0.00340,-0.01186,-0.01249,0.02545,-0.00160,-0.02688,
                     -0.29713,0.03274,-0.04517,-0.09436,0.18780,
                     -0.96873,-1.60663,2.89356,-0.07990,-1.89362,1.99472,
                     0.02931,0.01685)
    names(ardl_coeffs) <- c("(Intercept)",
                            "L(w, 1)", "L(w, 2)", "L(w, 3)", "L(w, 4)", "L(w, 5)", "L(w, 6)",
                            "Prod", "L(Prod, 1)",
                            "UR", "L(UR, 1)", "L(UR, 2)", "L(UR, 3)", "L(UR, 4)", "L(UR, 5)",
                            "Wedge", "L(Wedge, 1)", "L(Wedge, 2)", "L(Wedge, 3)", "L(Wedge, 4)",
                            "Union", "L(Union, 1)", "L(Union, 2)", "L(Union, 3)", "L(Union, 4)", "L(Union, 5)",
                            "D7475", "D7579")

    expect_equal(round(ardl61545$coefficients,5), ardl_coeffs)

    # check if non time-series data work
    ardl61545_df <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                         order = c(6,1,5,4,5), data = data.frame(PSS2001), start = 9)

    expect_equal(ardl61545$coefficients, ardl61545_df$coefficients)
})

test_that("ardl order", {
    ardl61545 <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                      order = c(6,1,5,4,5), data = PSS2001, start = c(1972, 01))
    expect_equal(ardl61545$order, c(6,1,5,4,5), ignore_attr = TRUE)
})

test_that("ardl parsed_formula", {
    ardl61545 <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                      order = c(6,1,5,4,5), data = PSS2001, start = c(1972, 01))
    expect_equal(ardl61545$parsed_formula$y_part$var, "w")
    expect_equal(ardl61545$parsed_formula$x_part$var, c("Prod", "UR", "Wedge", "Union"))
    expect_equal(ardl61545$parsed_formula$z_part$var, c("w", "Prod", "UR", "Wedge", "Union"))
    expect_equal(ardl61545$parsed_formula$fixed_part$var, c("D7475", "D7579"))
    expect_equal(ardl61545$parsed_formula$kz, 5)
    expect_equal(ardl61545$parsed_formula$kx, 4)
    expect_equal(ardl61545$parsed_formula$kw, 1)
    expect_equal(ardl61545$parsed_formula$kfixed, 2)
})

test_that("check equivalence of methods", {
    ardl61545 <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                      order = c(6,1,5,4,5), data = PSS2001, start = c(1972, 01))
    uecm61545 <- uecm(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                      order = c(6,1,5,4,5), data = PSS2001, start = c(1972, 01))

    expect_equal(ardl61545, ardl(uecm61545))
})

#' Restricted ECM regression
#'
#' Creates the Restricted Error Correction Model (RECM). This is the conditional
#' RECM, which is the RECM of the underlying ARDL.
#'
#' Note that the statistical significance of 'ect' in a RECM should not be
#' tested using the corresponding t-statistic (or the p-value) because it
#' doesn't follow a standard t-distribution. Instead, the
#' \code{\link{bounds_t_test}} should be used.
#'
#' @param object An object of \code{\link[base]{class}} 'ardl' or 'uecm'.
#' @param case An integer from 1-5 or a character string specifying whether the
#'   'intercept' and/or the 'trend' have to participate in the short-run or the
#'   long-run relationship (cointegrating equation) (see section 'Cases' below).
#'
#' @return \code{recm} returns an object of \code{\link[base]{class}}
#'   \code{c("recm", "dynlm", "lm")}. In addition, attributes 'order', 'data',
#'   'parsed_formula' and 'full_formula' are provided.
#'
#' @section Mathematical Formula: The formula of a Restricted ECM conditional to
#'   an \eqn{ARDL(p,q_{1},\dots,q_{k})}{ARDL(p,q1,...,qk)} is: \deqn{\Delta
#'   y_{t} = c_{0} + c_{1}t + \sum_{i=1}^{p-1}\psi_{y,i}\Delta y_{t-i} +
#'   \sum_{j=1}^{k}\sum_{l=1}^{q_{j}-1} \psi_{j,l}\Delta x_{j,t-l} +
#'   \sum_{j=1}^{k}\omega_{j}\Delta x_{j,t} + \pi_{y}ECT_{t} + \epsilon_{t}}
#'   \deqn{\psi_{j,l} = 0 \;\; \forall \;\; q_{j} = 1, \psi_{j,l} = \omega_{j} =
#'    0 \;\; \forall \;\; q_{j} = 0}
#' \describe{
#'   \item{Under Case 1:}{\itemize{
#'      \item \eqn{c_{0}=c_{1}=0}
#'      \item \eqn{ECT = y_{t-1} - (\sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
#'   \item{Under Case 2:}{\itemize{
#'      \item \eqn{c_{0}=c_{1}=0}
#'      \item \eqn{ECT = y_{t-1} - (\mu + \sum_{j=1}^{k}\theta_{j} x_{j,t-1})}}}
#'   \item{Under Case 3:}{\itemize{
#'      \item \eqn{c_{1}=0}
#'      \item \eqn{ECT = y_{t-1} - (\sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
#'   \item{Under Case 4:}{\itemize{
#'      \item \eqn{c_{1}=0}
#'      \item \eqn{ECT = y_{t-1} - (\delta(t-1)+ \sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
#'   \item{Under Case 5:}{\itemize{
#'      \item \eqn{ECT = y_{t-1} - (\sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
#' }
#' In all cases, \eqn{x_{j,t-1}} in \eqn{ECT} is replaced by \eqn{x_{j,t} \;\;\;\;\; \forall \;\; q_{j} = 0}
#'
#' @section Cases: According to \cite{Pesaran et al. (2001)}, we distinguish the
#' long-run relationship (cointegrating equation) (and thus the bounds-test and
#' the Restricted ECMs) between 5 different cases. These differ in terms of
#' whether the 'intercept' and/or the 'trend' are restricted to participate in
#' the long-run relationship or they are unrestricted and so they participate in
#' the short-run relationship.
#'
#' \describe{
#'   \item{Case 1:}{\itemize{
#'     \item No \emph{intercept} and no \emph{trend}.
#'     \item \code{case} inputs: 1 or "n" where "n" stands for none.}}
#'   \item{Case 2:}{\itemize{
#'     \item Restricted \emph{intercept} and no \emph{trend}.
#'     \item \code{case} inputs: 2 or "rc" where "rc" stands for restricted
#'      constant.}}
#'   \item{Case 3:}{\itemize{
#'     \item Unrestricted \emph{intercept} and no \emph{trend}.
#'     \item \code{case} inputs: 3 or "uc" where "uc" stands for unrestricted
#'     constant.}}
#'   \item{Case 4:}{\itemize{
#'     \item Unrestricted \emph{intercept} and restricted \emph{trend}.
#'     \item \code{case} inputs: 4 or "ucrt" where "ucrt" stands for
#'     unrestricted constant and restricted trend.}}
#'   \item{Case 5:}{\itemize{
#'     \item Unrestricted \emph{intercept} and unrestricted \emph{trend}.
#'     \item \code{case} inputs: 5 or "ucut" where "ucut" stands for
#'      unrestricted constant and unrestricted trend.}}
#' }
#'
#' Note that you can't restrict (or leave unrestricted) a parameter that doesn't
#' exist in the input model. For example, you can't compute \code{recm(object,
#' case=3)} if the object is an ARDL (or UECM) model with no intercept. The same
#' way, you can't compute \code{bounds_f_test(object, case=5)} if the object is
#' an ARDL (or UECM) model with no linear trend.
#'
#' @section References: Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds
#'   testing approaches to the analysis of level relationships. \emph{Journal of
#'   Applied Econometrics}, 16(3), 289-326
#'
#' @seealso \code{\link{ardl}} \code{\link{uecm}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords models ts
#' @export
#' @examples
#' data(denmark)
#'
#' ## Estimate the RECM, conditional to it's underlying ARDL(3,1,3,2) -----
#'
#' # Indirectly from an ARDL
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' recm_3132 <- recm(ardl_3132, case = 2)
#'
#' # Indirectly from an UECM
#' uecm_3132 <- uecm(ardl_3132)
#' recm_3132_ <- recm(uecm_3132, case = 2)
#' identical(recm_3132, recm_3132_)
#' summary(recm_3132)
#'
#' ## Error Correction Term (ect) & Speed of Adjustment -------------------
#'
#' # The coefficient of the ect,
#' # shows the Speed of Adjustment towards equilibrium.
#' # Note that this can be also be obtained from an UECM,
#' # through the coefficient of the term L(y, 1) (where y is the dependent variable).
#' tail(recm_3132$coefficients, 1)
#' uecm_3132$coefficients[2]
#'
#' ## Post-estimation testing ---------------------------------------------
#'
#' # See examples in the help file of the uecm() function
#'

recm <- function(object, case) {
    # no visible binding for global variable NOTE solution
    y <- NULL; rm(y)

    class(object)[4] <- "recm_indicator"
    coint_eq_list <- coint_eq(object = object, case = case)
    design_matrix <- coint_eq_list$design_matrix
    coint_eq <- coint_eq_list$coint_eq
    data <- coint_eq_list$data
    parsed_formula <- coint_eq_list$parsed_formula
    case <- coint_eq_list$case
    order <- coint_eq_list$order
    recm_formula <- build_recm_formula(parsed_formula = parsed_formula, order = order, case = case)
    if (0 %in% order) {
        design_matrix <- stats::ts(design_matrix, start = stats::start(data), frequency = stats::frequency(data))
        x_0_order <- colnames(design_matrix)[-c(1:3)][order[-1] == 0]
        x_q_order <- colnames(design_matrix)[-1][!(colnames(design_matrix)[-1] %in% x_0_order)]
        coint_eq <- rowSums(zoo::cbind.zoo(stats::lag(design_matrix[,x_q_order], -1),  design_matrix[,x_0_order]))
        coint_eq <- stats::ts(coint_eq, start = stats::start(data), frequency = stats::frequency(data))
        data_full <- zoo::cbind.zoo(data, stats::lag(design_matrix[,"y"], -1) - coint_eq)
    } else {
        # error term, u
        u <- design_matrix %>%
            dplyr::mutate(u = y - coint_eq) %>%
            dplyr::select(u) %>%
            unlist() %>%
            stats::ts(., start = stats::start(data), frequency = stats::frequency(data))
        # create a new data table with the lagged error term u, naming the error correction term (ect)
        data_full <- zoo::cbind.zoo(data, stats::lag(u, -1))
    }
    colnames(data_full) <- c(colnames(data), "ect")
    data <- data_full; rm(data_full)
    full_formula <- stats::formula(recm_formula$full)
    start <- start(object)
    end  <- end(object)

    recm_model <- dynlm::dynlm(full_formula, data = data, start = start, end = end)
    # for model compatibility in the global env
    attr(recm_model$terms, ".Environment") <- .GlobalEnv
    attr(attr(recm_model$model, "terms"), ".Environment") <- .GlobalEnv
    attr(full_formula, ".Environment") <- .GlobalEnv
    attr(recm_model, "class") <- c(class(recm_model), "recm")
    #recm_model <- structure(recm_model, class = c("recm", class(recm_model)))
    recm_model$order <- order
    recm_model$data <- data
    recm_model$parsed_formula <- parsed_formula
    recm_model$full_formula <- full_formula

    return(recm_model)
}

#' ARDL formula specification builder
#'
#' It creates the ARDL specification according to the given "formula" and their
#' corresponding "orders".
#'
#' @param parsed_formula A list containing the formula parts as returned from
#'   \code{\link{parse_formula}}.
#' @param order A numeric vector with the ARDL order as returned from
#'   \code{\link[=parse_order]{parse_order(restriction = FALSE)}}.
#'
#' @return \code{build_ardl_formula} returns a list containing the full formula
#'   and the independent and dependent parts of the formula separated. The full
#'   formula is ready to be used as input in the \code{dynlm} function.
#'
#' @seealso \code{\link{build_uecm_formula}}, \code{\link{build_recm_formula}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

build_ardl_formula <- function(parsed_formula, order) {

    ardl_formula <- list()
    ardl_formula$dep_part <- paste0(parsed_formula$y_part$var, " ~ ")

    ardl_formula$indep_part$w$var <- c()
    if (length(parsed_formula$w_part$var) > 0) {
        ardl_formula$indep_part$w <- paste0(parsed_formula$w_part$var, " + ", collapse = "")
    }

    ardl_formula$indep_part$y <- c()
    ardl_formula$indep_part$y <- paste0("L(", parsed_formula$y_part$var, ", ", 1:order[1], ")", " + ", collapse = "")

    ardl_formula$indep_part$x <- c()
    for(i in 1:parsed_formula$kx) {
        if (order[i + 1] != 0) {
            temp <- paste0(parsed_formula$x_part$var[i], " + ", paste0("L(", parsed_formula$x_part$var[i], ", ", 1:order[i + 1], ")", " + ", collapse = ""), collapse = "")
        } else {
            temp <- paste0(parsed_formula$x_part$var[i], " + ")
        }
        ardl_formula$indep_part$x <- paste(ardl_formula$indep_part$x, temp, collapse = "")
    }
    rm(temp)

    if (parsed_formula$kfixed > 0) {
        ardl_formula$indep_part$fixed <- paste0(parsed_formula$fixed_part$var, " + ", collapse = "")
    } else {
        ardl_formula$indep_part$fixed <- ""
    }

    ardl_formula$full <- paste0(ardl_formula$dep_part, ardl_formula$indep_part$w,
                                ardl_formula$indep_part$y, ardl_formula$indep_part$x,
                                ardl_formula$indep_part$fixed, collapse = "") %>%
        stringr::str_trim(side = "both")
    ardl_formula$full <- ifelse(stringr::str_sub(ardl_formula$full,
                                        start = stringr::str_count(ardl_formula$full),
                                        end = stringr::str_count(ardl_formula$full)) == "+",
                                stringr::str_sub(ardl_formula$full, start = 1, end = stringr::str_count(ardl_formula$full) -1 ),
                                ardl_formula$full) %>%
        stringr::str_trim(side = "both")

    return_list <- list(dep_part = ardl_formula$dep_part, indep_part = ardl_formula$indep_part,
                        full = ardl_formula$full)

    return(return_list)
}

#' UECM formula specification builder
#'
#' It creates the UECM (Unrestricted Error Correction Model) specification
#' according to the given "formula" and the corresponding "order" of the
#' underlying ARDL.
#'
#' @param order A numeric vector with the underlying ARDL order as returned from
#'   \code{\link[=parse_order]{parse_order(restriction = FALSE)}}.
#' @inheritParams build_ardl_formula
#'
#' @return \code{build_uecm_formula} returns a list containing the full formula
#'   and the independent and dependent parts of the formula separated. The full
#'   formula is ready to be used as input in the \code{dynlm} function.
#'
#' @seealso \code{\link{build_ardl_formula}}, \code{\link{build_recm_formula}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

build_uecm_formula <- function(parsed_formula, order) {

    uecm_formula <- list()
    uecm_formula$dep_part <- paste0("d(", parsed_formula$y_part$var, ")", " ~ ")

    uecm_formula$indep_part$w <- c()
    if (length(parsed_formula$w_part$var) > 0) {
        uecm_formula$indep_part$w <- paste0(parsed_formula$w_part$var, " + ", collapse = "")
    }

    uecm_formula$indep_part$levels$y <- c()
    uecm_formula$indep_part$levels$y <- paste0("L(", parsed_formula$y_part$var, ", ", 1, ")", " + ", collapse = "")

    uecm_formula$indep_part$levels$x <- c()
    uecm_formula$indep_part$diff$x <- c()
    for(i in 1:parsed_formula$kx){
        if (order[i + 1] == 0) {
            temp <- paste0(parsed_formula$x_part$var[i], " + ")
            temp1 <- ""
        } else if (order[i + 1] == 1) {
            temp <- paste0("L(", parsed_formula$x_part$var[i], ", ", 1, ")", " + ", collapse = "")
            temp1 <- paste0("d(", parsed_formula$x_part$var[i], ") + ", collapse = "")
        } else {
            temp <- paste0("L(", parsed_formula$x_part$var[i], ", ", 1, ")", " + ", collapse = "")
            temp1 <- paste0("d(", parsed_formula$x_part$var[i], ") + ",
                        paste0("d(L(", parsed_formula$x_part$var[i], ", ", 1:(order[i + 1] - 1), "))", " + ", collapse = ""),
                        collapse = "")
        }
        uecm_formula$indep_part$levels$x <- paste(uecm_formula$indep_part$levels$x, temp, collapse = "")
        uecm_formula$indep_part$diff$x <- paste(uecm_formula$indep_part$diff$x, temp1, collapse = "")
    }
    rm(temp); rm(temp1)

    uecm_formula$indep_part$diff$y <- c()
    if (order[1] == 1) {
        uecm_formula$indep_part$diff$y <- ""
    } else { # AR(0) models are discarded anyway
        uecm_formula$indep_part$diff$y <- paste0("d(L(", parsed_formula$y_part$var, ", ", 1:(order[1] - 1), "))", " + ", collapse = "")
    }

    if (parsed_formula$kfixed > 0) {
        uecm_formula$indep_part$fixed <- paste0(parsed_formula$fixed_part$var, " + ", collapse = "")
    } else {
        uecm_formula$indep_part$fixed <- ""
    }

    uecm_formula$full <- paste0(uecm_formula$dep_part, uecm_formula$indep_part$w,
                                uecm_formula$indep_part$levels$y, uecm_formula$indep_part$levels$x,
                                uecm_formula$indep_part$diff$y, uecm_formula$indep_part$diff$x,
                                uecm_formula$indep_part$fixed, collapse = "") %>%
        stringr::str_trim(side = "both")
    uecm_formula$full <- ifelse(stringr::str_sub(uecm_formula$full,
                                        start = stringr::str_count(uecm_formula$full),
                                        end = stringr::str_count(uecm_formula$full)) == "+",
                                stringr::str_sub(uecm_formula$full, start = 1, end = stringr::str_count(uecm_formula$full) -1 ),
                                uecm_formula$full) %>%
        stringr::str_trim(side = "both")

    return_list <- list(dep_part = uecm_formula$dep_part, indep_part = uecm_formula$indep_part,
                        full = uecm_formula$full)

    return(return_list)
}

#' RECM formula specification builder
#'
#' It creates the RECM (Restricted Error Correction Model) specification
#' according to the given "formula" and the corresponding "order" of the
#' underlying ARDL.
#'
#' @param case An integer from 1-5 or a character string specifying whether the
#'   'intercept' or the 'trend' have to participate in the 
#'   long-run/cointegrating relationship/equation (see 'Details').
#' @inheritParams build_uecm_formula
#'
#' @return \code{build_recm_formula} returns a list containing the full formula
#'   and the independent and dependent parts of the formula separated. The full
#'   formula is ready to be used as input in the \code{dynlm} function providing
#'   also the 'ect' (error correction term) to the data.
#'
#' @seealso \code{\link{build_ardl_formula}}, \code{\link{build_uecm_formula}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

build_recm_formula <- function(parsed_formula, order, case) {

    recm_formula <- list()
    recm_formula$dep_part <- paste0("d(", parsed_formula$y_part$var, ")", " ~ ")
    recm_formula$indep_part$w <- c()
    recm_formula$indep_part$w <- paste0(parsed_formula$w_part$var, " + ", collapse = "")
    if (case %in% c(1, 2)) {
        recm_formula$indep_part$w <- "- 1 + "
    } else if (case %in% c(3, 4)) {
        recm_formula$indep_part$w <- ""
    } else if (case == 5) {
        recm_formula$indep_part$w <- paste0(parsed_formula$w_part$var, " + ", collapse = "")
    }

    recm_formula$indep_part$diff$x <- c()
    for(i in 1:parsed_formula$kx){
        if (order[i + 1] == 0) {
            temp <- ""
        } else if (order[i + 1] == 1) {
            temp <- paste0("d(", parsed_formula$x_part$var[i], ") + ", collapse = "")
        } else {
            temp <- paste0("d(", parsed_formula$x_part$var[i], ") + ",
                        paste0("d(L(", parsed_formula$x_part$var[i], ", ", 1:(order[i + 1] - 1), "))", " + ", collapse = ""),
                        collapse = "")
        }
        recm_formula$indep_part$diff$x <- paste(recm_formula$indep_part$diff$x, temp, collapse = "")
    }
    rm(temp)

    recm_formula$indep_part$diff$y <- c()
    if (order[1] == 1) {
        recm_formula$indep_part$diff$y <- ""
    } else { # AR(0) models are discarded anyway
        recm_formula$indep_part$diff$y <- paste0("d(L(", parsed_formula$y_part$var, ", ", 1:(order[1] - 1), "))", " + ", collapse = "")
    }

    if (parsed_formula$kfixed > 0) {
        recm_formula$indep_part$fixed <- paste0(parsed_formula$fixed_part$var, " + ", collapse = "")
    } else {
        recm_formula$indep_part$fixed <- ""
    }

    recm_formula$indep_part$ect <- "ect"

    recm_formula$full <- paste0(recm_formula$dep_part, recm_formula$indep_part$w,
                                recm_formula$indep_part$diff$y, recm_formula$indep_part$diff$x,
                                recm_formula$indep_part$fixed, recm_formula$indep_part$ect,
                                collapse = "") %>%
        stringr::str_trim(side = "both")
    recm_formula$full <- ifelse(stringr::str_sub(recm_formula$full,
                                        start = stringr::str_count(recm_formula$full),
                                        end = stringr::str_count(recm_formula$full)) == "+",
                                stringr::str_sub(recm_formula$full, start = 1, end = stringr::str_count(recm_formula$full) -1 ),
                                recm_formula$full) %>%
        stringr::str_trim(side = "both")

    return_list <- list(dep_part = recm_formula$dep_part, indep_part = recm_formula$indep_part,
                        full = recm_formula$full)

    return(return_list)
}

#' Convert dynlm model (ardl, uecm, recm) to lm model
#'
#' Takes a \code{\link[dynlm]{dynlm}} model of \code{\link[base]{class}} 'ardl',
#' 'uecm' or 'recm' and converts it into an \code{\link[stats]{lm}} model. This
#' can help using the model as a regular \code{\link[stats]{lm}} model with
#' functions that are not compatible with \code{\link[dynlm]{dynlm}} models such
#' as the \code{\link[stats]{predict}} function to forecast.
#'
#' @param object An object of \code{\link[base]{class}} 'ardl', 'uecm' or 'recm'.
#' @param fix_names A logical, indicating whether the variable names should be
#' rewritten without special functions and character in the names such as "d()"
#' or "L()". When \code{fix_names = TRUE}, the characters "(", and "," are
#' replaces with ".", and ")" and spaces are deleted. The name of the dependent
#' variable is always transformed, regardless of the value of this parameter.
#' Default is FALSE.
#' @param data_class If "ts", it converts the data class to
#' \code{\link[stats]{ts}} (see examples for its usage). The default is
#' \code{\link[base]{NULL}}, which uses the same data provided in the original
#' object.
#' @param ... Currently unused argument.
#'
#' @return \code{to_lm} returns an object of \code{\link[base]{class}}
#' \code{"lm"}.
#'
#' @seealso \code{\link{ardl}}, \code{\link{uecm}}, \code{\link{recm}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords models ts
#' @export
#' @examples
#' ## Convert ARDL into lm ------------------------------------------------
#'
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' ardl_3132_lm <- to_lm(ardl_3132)
#' summary(ardl_3132)$coefficients
#' summary(ardl_3132_lm)$coefficients
#'
#' ## Convert UECM into lm ------------------------------------------------
#'
#' uecm_3132 <- uecm(ardl_3132)
#' uecm_3132_lm <- to_lm(uecm_3132)
#' summary(uecm_3132)$coefficients
#' summary(uecm_3132_lm)$coefficients
#'
#' ## Convert RECM into lm ------------------------------------------------
#'
#' recm_3132 <- recm(ardl_3132, case = 2)
#' recm_3132_lm <- to_lm(recm_3132)
#' summary(recm_3132)$coefficients
#' summary(recm_3132_lm)$coefficients
#'
#' ## Use the lm model to forecast ----------------------------------------
#'
#' # Forecast using the in-sample data
#' insample_data <- ardl_3132$model
#' head(insample_data)
#' predicted_values <- predict(ardl_3132_lm, newdata = insample_data)
#'
#' # The predicted values are expected to be the same as the fitted values
#' ardl_3132$fitted.values
#' predicted_values
#'
#' # Convert to ts class for the plot
#' predicted_values <- ts(predicted_values, start = c(1974,4), frequency=4)
#' plot(denmark$LRM, lwd=4) #The input dependent variable
#' lines(ardl_3132$fitted.values, lwd=4, col="blue") #The fitted values
#' lines(predicted_values, lty=2, lwd=2, col="red") #The predicted values
#'
#' ## Convert to lm for post-estimation testing ---------------------------
#'
#' # Ramsey's RESET test for functional form
#' library(lmtest) # for resettest()
#' library(strucchange) # for efp(), and sctest()
#'
#' \dontrun{
#'     # This produces an error.
#'     # resettest() cannot use data of class 'zoo' such as the 'denmark' data
#'     # used to build the original model
#'     resettest(uecm_3132, type = c("regressor"))
#' }
#'
#' uecm_3132_lm <- to_lm(uecm_3132, data_class = "ts")
#' resettest(uecm_3132_lm, power = 2)
#'
#' # CUSUM test for structural change detection
#' \dontrun{
#'     # This produces an error.
#'     # efp() does not understand special functions such as "d()" and "L()"
#'     efp(uecm_3132$full_formula, data = uecm_3132$model)
#' }
#'
#' uecm_3132_lm_names <- to_lm(uecm_3132, fix_names = TRUE)
#' fluctuation <- efp(uecm_3132_lm_names$full_formula,
#'                    data = uecm_3132_lm_names$model)
#' sctest(fluctuation)
#' plot(fluctuation)
#'

to_lm <- function(object, fix_names = FALSE, data_class = NULL, ...) {
    objmodel <- object$model
    if (!is.null(data_class)) {
        if (data_class == "ts") {
            objmodel <- stats::ts(objmodel, start = stats::start(objmodel[,1]), frequency = stats::frequency(objmodel[,1]))
        }
    }
    dep_var <- colnames(objmodel)[1]
    fix_names_fun <- function(text) {
        text <- gsub(" ", "", text) %>%
            gsub("(", ".", ., fixed = TRUE) %>%
            gsub(")", "", ., fixed = TRUE) %>%
            gsub(",", ".", ., fixed = TRUE)
        return(text)
    }
    y <- fix_names_fun(dep_var)

    if (fix_names) {
        colnames(objmodel) <- sapply(colnames(objmodel), fix_names_fun)
        if (attr(object$terms,"intercept") == 0) {
            formula <- formula(paste0(y, " ~ . -1"))
        } else {
            formula <- formula(paste0(y, " ~ ."))
        }

        full_formula <- as.character(object$full_formula)
        lm_model <- stats::lm(formula, data = objmodel)
        lm_model$full_formula <- stats::formula(paste0(y, "~", fix_names_fun(full_formula[3])))

        return(lm_model)
    } else {
        colnames(objmodel)[1] <- y
        if (attr(object$terms,"intercept") == 0) {
            formula <- formula(paste0(y, " ~ . -1"))
        } else {
            formula <- formula(paste0(y, " ~ ."))
        }

        return(stats::lm(formula, data = objmodel))
    }
}

#' Bounds Wald-test for no cointegration
#'
#' \code{bounds_f_test} performs the Wald bounds-test for no cointegration
#' \cite{Pesaran et al. (2001)}. It is a Wald test on the parameters of a UECM
#' (Unrestricted Error Correction Model) expressed either as a Chisq-statistic
#' or as an F-statistic.
#'
#' @param alpha A numeric value between 0 and 1 indicating the significance
#'   level of the critical value bounds. If \code{NULL} (default), no critical
#'   value bounds for a specific level of significance are provide, only the
#'   p-value. See section 'alpha, bounds and p-value' below for details.
#' @param pvalue A logical indicating whether you want the p-value to be
#'   provided. The default is \code{TRUE}. See section 'alpha, bounds and
#'   p-value' below for details.
#' @param exact A logical indicating whether you want asymptotic (T = 1000) or
#'   exact sample size critical value bounds and p-value. The default is
#'   \code{FALSE} for asymptotic. See section 'alpha, bounds and p-value' below
#'   for details.
#' @param R An integer indicating how many iterations will be used if
#'   \code{exact = TRUE}. Default is 40000.
#' @param test A character vector indicating whether you want the Wald test to
#'   be expressed as 'F' or as 'Chisq' statistic. Default is "F".
#' @param vcov_matrix The estimated covariance matrix of the parameter estimates
#'   of the uecm model. Used to estimate the test statistic. The default is 
#'   \code{vcov(object)} (when \code{vcov_matrix = NULL}), but other estimations
#'   of the covariance matrix of the regression's estimated coefficients can 
#'   also be used (e.g., using \code{\link[sandwich]{vcovHC}} or 
#'   \code{\link[sandwich]{vcovHAC}}). Only applicable if the input object is of
#'   class "uecm".
#' @inheritParams recm
#'
#' @return A list with class "htest" containing the following components:
#'   \item{\code{method}}{a character string indicating what type of test was
#'      performed.}
#'   \item{\code{alternative}}{a character string describing the alternative
#'      hypothesis.}
#'   \item{\code{statistic}}{the value of the test statistic.}
#'   \item{\code{null.value}}{the value of the population parameters \code{k}
#'      (the number of independent variables) and \code{T} (the number of
#'      observations) specified by the null hypothesis.}
#'   \item{\code{data.name}}{a character string giving the name(s) of the data.}
#'   \item{\code{parameters}}{numeric vector containing the critical value
#'      bounds.}
#'   \item{\code{p.value}}{the p-value of the test.}
#'   \item{\code{PSS2001parameters}}{numeric vector containing the critical
#'      value bounds as presented by \cite{Pesaran et al. (2001)}. See section
#'      'alpha, bounds and p-value' below for details.}
#'   \item{\code{tab}}{data.frame containing the statistic, the critical value
#'      bounds, the alpha level of significance and the p-value.}
#'
#' @section Hypothesis testing:
#'   \deqn{\Delta y_{t} = c_{0} + c_{1}t +
#'   \pi_{y}y_{t-1} + \sum_{j=1}^{k}\pi_{j}x_{j,t-1} +
#'   \sum_{i=1}^{p-1}\psi_{y,i}\Delta y_{t-i} +
#'   \sum_{j=1}^{k}\sum_{l=1}^{q_{j}-1} \psi_{j,l}\Delta x_{j,t-l} +
#'   \sum_{j=1}^{k}\omega_{j}\Delta x_{j,t} + \epsilon_{t}}
#'
#' \describe{
#'   \item{Cases 1, 3, 5:}{}
#' }
#'   \deqn{\mathbf{H_{0}:} \pi_{y} = \pi_{1} = \dots = \pi_{k} = 0}
#'   \deqn{\mathbf{H_{1}:} \pi_{y} \neq \pi_{1} \neq \dots \neq \pi_{k} \neq 0}
#'
#' \describe{
#'   \item{Case 2:}{}
#' }
#'   \deqn{\mathbf{H_{0}:} \pi_{y} = \pi_{1} = \dots = \pi_{k} = c_{0} = 0}
#'   \deqn{\mathbf{H_{1}:} \pi_{y} \neq \pi_{1} \neq \dots \neq \pi_{k} \neq c_{0} \neq 0}
#'
#' \describe{
#'   \item{Case 4:}{}
#' }
#'   \deqn{\mathbf{H_{0}:} \pi_{y} = \pi_{1} = \dots = \pi_{k} = c_{1} = 0}
#'   \deqn{\mathbf{H_{1}:} \pi_{y} \neq \pi_{1} \neq \dots \neq \pi_{k} \neq c_{1} \neq 0}
#'
#' @section alpha, bounds and p-value: In this section it is explained how the
#'   critical value bounds and p-values are obtained.
#'   \itemize{
#'      \item If \code{exact = FALSE}, then the asymptotic (T = 1000) critical
#'          value bounds and p-value are provided.
#'      \item Only the asymptotic critical value bounds and p-values, and only
#'          for k <= 10 are precalculated, everything else has to be computed.
#'      \item Precalculated critical value bounds and p-values were simulated
#'          using \code{set.seed(2020)} and \code{R = 70000}.
#'      \item Precalculated critical value bounds exist only for \code{alpha}
#'          being one of the 0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.15 or 0.2,
#'          everything else has to be computed.
#'      \item If \code{alpha} is one of the 0.1, 0.05, 0.025 or 0.01 (and
#'          \code{exact = FALSE} and k <= 10),\cr
#'          \code{PSS2001parameters} shows the critical value bounds presented
#'          in \cite{Pesaran et al. (2001)}
#'          (less precise).
#'   }
#'
#' @inheritSection recm Cases
#' @inheritSection recm References
#' @seealso \code{\link{bounds_t_test}} \code{\link{ardl}} \code{\link{uecm}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords htest ts
#' @export
#' @examples
#' data(denmark)
#'
#' ## How to use cases under different models (regarding deterministic terms)
#'
#' ## Construct an ARDL(3,1,3,2) model with different deterministic terms -
#'
#' # Without constant
#' ardl_3132_n <- ardl(LRM ~ LRY + IBO + IDE -1, data = denmark, order = c(3,1,3,2))
#'
#' # With constant
#' ardl_3132_c <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#'
#' # With constant and trend
#' ardl_3132_ct <- ardl(LRM ~ LRY + IBO + IDE + trend(LRM), data = denmark, order = c(3,1,3,2))
#'
#' ## F-bounds test for no level relationship (no cointegration) ----------
#'
#' # For the model without a constant
#' bounds_f_test(ardl_3132_n, case = 1)
#' # or
#' bounds_f_test(ardl_3132_n, case = "n")
#'
#' # For the model with a constant
#' # Including the constant term in the long-run relationship (restricted constant)
#' bounds_f_test(ardl_3132_c, case = 2)
#' # or
#' bounds_f_test(ardl_3132_c, case = "rc")
#'
#' # Including the constant term in the short-run relationship (unrestricted constant)
#' bounds_f_test(ardl_3132_c, case = "uc")
#' # or
#' bounds_f_test(ardl_3132_c, case = 3)
#'
#' # For the model with constant and trend
#' # Including the constant term in the short-run and the trend in the long-run relationship
#' # (unrestricted constant and restricted trend)
#' bounds_f_test(ardl_3132_ct, case = "ucrt")
#' # or
#' bounds_f_test(ardl_3132_ct, case = 4)
#'
#' # For the model with constant and trend
#' # Including the constant term and the trend in the short-run relationship
#' # (unrestricted constant and unrestricted trend)
#' bounds_f_test(ardl_3132_ct, case = "ucut")
#' # or
#' bounds_f_test(ardl_3132_ct, case = 5)
#'
#' ## Note that you can't restrict a deterministic term that doesn't exist
#'
#' # For example, the following tests will produce an error:
#' \dontrun{
#' bounds_f_test(ardl_3132_c, case = 1)
#' bounds_f_test(ardl_3132_ct, case = 3)
#' bounds_f_test(ardl_3132_c, case = 4)
#' }
#'
#' ## Asymptotic p-value and critical value bounds (assuming T = 1000) ----
#'
#' # Include critical value bounds for a certain level of significance
#'
#' # F-statistic is larger than the I(1) bound (for a=0.05) as expected (p-value < 0.05)
#' bft <- bounds_f_test(ardl_3132_c, case = 2, alpha = 0.05)
#' bft
#' bft$tab
#'
#' # Traditional but less precise critical value bounds, as presented in Pesaran et al. (2001)
#' bft$PSS2001parameters
#'
#' # F-statistic is slightly larger than the I(1) bound (for a=0.005)
#' # as p-value is slightly smaller than 0.005
#' bounds_f_test(ardl_3132_c, case = 2, alpha = 0.005)
#'
#' ## Exact sample size p-value and critical value bounds -----------------
#'
#' # Setting a seed is suggested to allow the replication of results
#' # 'R' can be increased for more accurate resutls
#'
#' # F-statistic is smaller than the I(1) bound (for a=0.01) as expected (p-value > 0.01)
#' # Note that the exact sample p-value (0.01285) is very different than the
#' # asymptotic (0.004418)
#' # It can take more than 30 seconds
#' \dontrun{
#' set.seed(2020)
#' bounds_f_test(ardl_3132_c, case = 2, alpha = 0.01, exact = TRUE)
#' }
#'
#' ## "F" and "Chisq" statistics ------------------------------------------
#'
#' # The p-value is the same, the test-statistic and critical value bounds are different
#' # but analogous
#' bounds_f_test(ardl_3132_c, case = 2, alpha = 0.01)
#' bounds_f_test(ardl_3132_c, case = 2, alpha = 0.01, test = "Chisq")

bounds_f_test <- function(object, case, alpha = NULL, pvalue = TRUE, exact = FALSE,
    R = 40000, test = c("F", "Chisq"), vcov_matrix = NULL) {
    # no visible binding for global variable NOTE solution
    k <- I0 <- fI0 <- fI1 <- NULL; rm(k, I0, fI0, fI1)

    if (isTRUE(all.equal(c("ardl", "dynlm", "lm"), class(object)))) {
        object <- uecm(object)
        if (!is.null(vcov_matrix)) {
            stop("'vcov_matrix' is only applicable if the input object is of class 'uecm'", call. = FALSE)
        }
    }

    if (is.null(vcov_matrix)) {
        vcov_matrix <- stats::vcov(object)
    }
    alpha_long_seq <- c(seq(0, 0.1, by = 0.001), seq(0.11, 1, by = 0.01))
    alpha_short_seq <- c(0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.15, 0.2)
    if (is.null(alpha) & (pvalue == FALSE)) {
        stop("Specify an 'alpha' level, set 'pvalue' = TRUE or both", call. = FALSE)
    }
    if ((!is.null(alpha) & (length(alpha) != 1)) | (!is.null(alpha) & !any(dplyr::near(alpha, alpha_long_seq)))) {
        stop("alpha must be either 'NULL' or one of the numbers produced by the following code:\n",
            "c(seq(0.001, 0.1, by = 0.001), seq(0.11, 0.99, by = 0.01))", call. = FALSE)
    }
    case <- parse_case(parsed_formula = object$parsed_formula, case = case)
    test <- match.arg(test)
    kx <- object$parsed_formula$kx
    T <- length(object$residuals)
    dname <- deparse(object$full_formula)
    if (exact == FALSE) {
        nullvalue <- c(kx, 1000)
    } else {
        nullvalue <- c(kx, T)
    }
    names(nullvalue) <- c("k", "T")

    indep_vars <- object$parsed_formula$x_part$var
    restricted_coef <- sapply(1:length(indep_vars), function(i) {
        ifelse(object$order[i + 1] == 0, indep_vars[i], paste0("L(", indep_vars[i], ", 1)"))
    })
    restricted_names <- c(paste0("L(", object$parsed_formula$y_part$var, ", 1)"), restricted_coef)
    restricted_coef <- which(names(stats::coef(object)) %in% restricted_names)

    if (case == 2) {
        restricted_names <- c(restricted_names, names(stats::coef(object))[1])
        restricted_coef <- which(names(stats::coef(object)) %in% restricted_names)
    } else if (case == 4) {
        restricted_names <- c(restricted_names, names(stats::coef(object))[2])
        restricted_coef <- which(names(stats::coef(object)) %in% restricted_names)
    }
    chi2_statistic <- aod::wald.test(b = stats::coef(object), Sigma = vcov_matrix, Terms = restricted_coef)$result
    if (test == "F") {
        statistic <- chi2_statistic$chi2[1] / chi2_statistic$chi2[2]
    } else {
        statistic <- chi2_statistic$chi2[1]
    }
    names(statistic) <- test
    if ((exact == FALSE) & (kx <= 10) & any(c(is.null(alpha), alpha %in% alpha_short_seq))) {
        caselatin <- ifelse(case == 1, "i",
                     ifelse(case == 2, "ii",
                     ifelse(case == 3, "iii",
                     ifelse(case == 4, "iv", "v"))))
        if (!is.null(alpha)) {
            if (alpha %in% c(0.1, 0.05, 0.025, 0.01)) {
                bounds_pss2001 <- eval(parse(text = paste0("crit_val_bounds_pss2001$f$", caselatin))) %>%
                    dplyr::filter(alpha %in% !!alpha, k == kx) %>%
                    dplyr::select(I0, I1)
                if (test != "F") {
                    if (case %in% c(1, 3, 5)) {
                        bounds_pss2001 <- bounds_pss2001*(kx+1)
                    } else {
                        bounds_pss2001 <- bounds_pss2001*(kx+2)
                    }
                }
            }
            if (alpha %in% alpha_short_seq) {
                bounds <- eval(parse(text = paste0("crit_val_bounds$I0$", caselatin))) %>%
                    dplyr::filter(alpha == !!alpha, k == kx) %>%
                    dplyr::select(fI0) %>%
                    dplyr::rename(I0 = fI0)
                bounds <- cbind(bounds,
                    eval(parse(text = paste0("crit_val_bounds$I1$", caselatin))) %>%
                        dplyr::filter(alpha == !!alpha, k == kx) %>%
                        dplyr::select(fI1) %>%
                        dplyr::rename(I1 = fI1))
                if (test != "F") {
                    if (case %in% c(1, 3, 5)) {
                        bounds <- bounds*(kx+1)
                    } else {
                        bounds <- bounds*(kx+2)
                    }
                }
            }
        }
        if (pvalue == TRUE) {
            critvalbounds <- eval(parse(text = paste0("crit_val_bounds$I1$", caselatin))) %>%
                             dplyr::mutate(I1 = if (test == "F") {
                                    fI1
                                } else if (case %in% c(1, 3, 5)){
                                    fI1*(kx+1)
                                } else {
                                    fI1*(kx+2)
                                })
            if (statistic %in% critvalbounds$I1) {
                p_value <- critvalbounds %>%
                    dplyr::filter(k == kx) %>%
                    dplyr::filter(I1 == statistic) %>%
                    dplyr::select(alpha) %>%
                    unlist()
            } else { # linear interpolation
                if (max(dplyr::filter(critvalbounds, k == kx)$I1) < statistic) {
                    p_value <- 0.000001
                } else if (min(dplyr::filter(critvalbounds, k == kx)$I1) > statistic) {
                    p_value <- 0.999999
                } else {
                    p_value <- critvalbounds %>%
                        dplyr::filter(k == kx) %>%
                        dplyr::filter(I1 %in% c(max(I1[which(I1 < statistic)]),
                            min(I1[which(I1 > statistic)]))) %>%
                        dplyr::mutate(p_value = min(alpha) + (max(alpha)-min(alpha)) * ((max(I1) - statistic) /
                            (max(I1)-min(I1)))) %>%
                        dplyr::select(p_value) %>% unlist() %>% .[1]
                }
            }
        }
    } else {
        T_asy_or_exact <- ifelse(exact == FALSE, 1000, T)
        wb <- f_bounds_sim(case = case, k = kx, alpha = alpha_long_seq,
                              T = T_asy_or_exact, R = R)
        if (!is.null(alpha)) {
            if (test == "F") {
                bounds <- wb$f_bounds[which(dplyr::near(alpha, alpha_long_seq)),]
            } else {
                bounds <- wb$chisq_bounds[which(dplyr::near(alpha, alpha_long_seq)),]
            }
        }
        if (pvalue == TRUE) {
            len <- length(alpha_long_seq)
            if (test == "F") {
                I1 <- wb$f_bounds$I1[1:len]
            } else {
                I1 <- wb$chisq_bounds$I1[1:len]
            }
            if (statistic %in% I1) {
                p_value <- alpha_long_seq[statistic == I1]
            } else { # linear interpolation
                if (max(I1) < statistic) {
                    p_value <- 0.000001
                } else if (min(I1) > statistic) {
                    p_value <- 0.999999
                } else {
                    interp_cond <- I1 %in% c(max(I1[which(I1 < statistic)]),
                                              min(I1[which(I1 > statistic)]))
                    interp_I1 <- I1[interp_cond]
                    interp_alpha <- alpha_long_seq[interp_cond]
                    p_value <- min(interp_alpha) + (max(interp_alpha)-min(interp_alpha)) *
                               ((max(interp_I1) - statistic) / (max(interp_I1)-min(interp_I1)))
                }
            }
        }
    }

    method <- paste0("Bounds ", test, "-test (Wald) for no cointegration")
    alternative <- "Possible cointegration"
    rval <- list(method = method, alternative = alternative, statistic = statistic,
                 null.value = nullvalue, data.name = dname)
    tab <- data.frame(statistic = statistic)
    if (!is.null(alpha)) {
        parameters <- c(bounds$I0, bounds$I1)
        names(parameters) <- c("Lower-bound I(0)", "Upper-bound I(1)")
        tab <- cbind(tab, "Lower-bound I(0)" = parameters[1], "Upper-bound I(1)" = parameters[2],
                                  alpha = alpha)
        rval$parameters <- parameters
        if ((alpha %in% c(0.1, 0.05, 0.025, 0.01)) & exact == FALSE) {
            PSS2001parameters <- c(bounds_pss2001$I0, bounds_pss2001$I1)
            names(PSS2001parameters) <- c("Lower-bound I(0)", "Upper-bound I(1)")
            rval$PSS2001parameters <- PSS2001parameters
        }
    }
    if (pvalue == TRUE) {
        tab <- cbind(tab, p.value = p_value)
        rval$p.value <- p_value
    }
    rval$tab <- tab
    rval <- structure(rval, class = c("htest", class(rval)))
    return(rval)
}

#' Bounds t-test for no cointegration
#'
#' \code{bounds_t_test} performs the t-bounds test for no cointegration
#' \cite{Pesaran et al. (2001)}. It is a t-test on the parameters of a UECM
#' (Unrestricted Error Correction Model).
#'
#' @param case An integer (1, 3 or 5) or a character string specifying whether
#'   the 'intercept' and/or the 'trend' have to participate in the
#'   short-run relationship (see section 'Cases' below). Note that the t-bounds
#'   test can't be applied for cases 2 and 4.
#' @inheritParams bounds_f_test
#' @inherit bounds_f_test return
#'
#' @section Hypothesis testing: \deqn{\Delta y_{t} = c_{0} + c_{1}t +
#'   \pi_{y}y_{t-1} + \sum_{j=1}^{k}\pi_{j}x_{j,t-1} +
#'   \sum_{i=1}^{p-1}\psi_{y,i}\Delta y_{t-i} +
#'   \sum_{j=1}^{k}\sum_{l=1}^{q_{j}-1} \psi_{j,l}\Delta x_{j,t-l} +
#'   \sum_{j=1}^{k}\omega_{j}\Delta x_{j,t} + \epsilon_{t}}
#'   \deqn{\mathbf{H_{0}:} \pi_{y} = 0}
#'   \deqn{\mathbf{H_{1}:} \pi_{y} \neq 0}
#'
#' @inheritSection bounds_f_test alpha, bounds and p-value
#' @inheritSection bounds_f_test Cases
#' @inheritSection bounds_f_test References
#' @seealso \code{\link{bounds_f_test}} \code{\link{ardl}} \code{\link{uecm}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords htest ts
#' @export
#' @examples
#' data(denmark)
#'
#' ## How to use cases under different models (regarding deterministic terms)
#'
#' ## Construct an ARDL(3,1,3,2) model with different deterministic terms -
#'
#' # Without constant
#' ardl_3132_n <- ardl(LRM ~ LRY + IBO + IDE -1, data = denmark, order = c(3,1,3,2))
#'
#' # With constant
#' ardl_3132_c <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#'
#' # With constant and trend
#' ardl_3132_ct <- ardl(LRM ~ LRY + IBO + IDE + trend(LRM), data = denmark, order = c(3,1,3,2))
#'
#' ## t-bounds test for no level relationship (no cointegration) ----------
#'
#' # For the model without a constant
#' bounds_t_test(ardl_3132_n, case = 1)
#' # or
#' bounds_t_test(ardl_3132_n, case = "n")
#'
#' # For the model with a constant
#' # Including the constant term in the short-run relationship (unrestricted constant)
#' bounds_t_test(ardl_3132_c, case = "uc")
#' # or
#' bounds_t_test(ardl_3132_c, case = 3)
#'
#' # For the model with constant and trend
#' # Including the constant term and the trend in the short-run relationship
#' # (unrestricted constant and unrestricted trend)
#' bounds_t_test(ardl_3132_ct, case = "ucut")
#' # or
#' bounds_t_test(ardl_3132_ct, case = 5)
#'
#' ## Note that you can't use bounds t-test for cases 2 and 4, or use a wrong model
#'
#' # For example, the following tests will produce an error:
#' \dontrun{
#' bounds_t_test(ardl_3132_n, case = 2)
#' bounds_t_test(ardl_3132_c, case = 4)
#' bounds_t_test(ardl_3132_ct, case = 3)
#' }
#'
#' ## Asymptotic p-value and critical value bounds (assuming T = 1000) ----
#'
#' # Include critical value bounds for a certain level of significance
#'
#' # t-statistic is larger than the I(1) bound (for a=0.05) as expected (p-value < 0.05)
#' btt <- bounds_t_test(ardl_3132_c, case = 3, alpha = 0.05)
#' btt
#' btt$tab
#'
#' # Traditional but less precise critical value bounds, as presented in Pesaran et al. (2001)
#' btt$PSS2001parameters
#'
#' # t-statistic doesn't exceed the I(1) bound (for a=0.005) as p-value is greater than 0.005
#' bounds_t_test(ardl_3132_c, case = 3, alpha = 0.005)
#'
#' ## Exact sample size p-value and critical value bounds -----------------
#'
#' # Setting a seed is suggested to allow the replication of results
#' # 'R' can be increased for more accurate resutls
#'
#' # t-statistic is smaller than the I(1) bound (for a=0.01) as expected (p-value > 0.01)
#' # Note that the exact sample p-value (0.009874) is very different than the
#' # asymptotic (0.005538)
#' # It can take more than 90 seconds
#' \dontrun{
#' set.seed(2020)
#' bounds_t_test(ardl_3132_c, case = 3, alpha = 0.01, exact = TRUE)
#' }

bounds_t_test <- function(object, case, alpha = NULL, pvalue = TRUE,
                               exact = FALSE, R = 40000, vcov_matrix = NULL) {
    # no visible binding for global variable NOTE solution
    k <- I0 <- tI0 <- tI1 <- NULL; rm(k, I0, tI0, tI1)

    if (isTRUE(all.equal(c("ardl", "dynlm", "lm"), class(object)))) {
        object <- uecm(object)
        if (!is.null(vcov_matrix)) {
            stop("'vcov_matrix' is only applicable if the input object is of class 'uecm'", call. = FALSE)
        }
    }

    if (is.null(vcov_matrix)) {
        vcov_matrix <- stats::vcov(object)
    }
    alpha_long_seq <- c(seq(0, 0.1, by = 0.001), seq(0.11, 1, by = 0.01))
    alpha_short_seq <- c(0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.15, 0.2)
    if (is.null(alpha) & (pvalue == FALSE)) {
        stop("Specify an 'alpha' level, set 'pvalue' = TRUE or both", call. = FALSE)
    }
    if ((!is.null(alpha) & (length(alpha) != 1)) | (!is.null(alpha) & !any(dplyr::near(alpha, alpha_long_seq)))) {
        stop("alpha must be either 'NULL' or one of the numbers produced by the following code:\n",
            "c(seq(0.001, 0.1, by = 0.001), seq(0.11, 0.99, by = 0.01))", call. = FALSE)
    }
    case <- parse_case(parsed_formula = object$parsed_formula, case = case)
    if (!(case %in% c(1, 3, 5))) {
        stop("The t-bounds test applies only when 'case' is either 1, 3 or 5", call. = FALSE)
    }
    kx <- object$parsed_formula$kx
    T <- length(object$residuals)
    dname <- deparse(object$full_formula)
    if (exact == FALSE) {
        nullvalue <- c(kx, 1000)
    } else {
        nullvalue <- c(kx, T)
    }
    names(nullvalue) <- c("k", "T")

    indep_vars <- object$parsed_formula$x_part$var
    restricted_coef <- sapply(1:length(indep_vars), function(i) {
        ifelse(object$order[i + 1] == 0, indep_vars[i], paste0("L(", indep_vars[i], ", 1)"))
    })
    restricted_names <- c(paste0("L(", object$parsed_formula$y_part$var, ", 1)"), restricted_coef)
    restricted_coef <- which(names(stats::coef(object)) %in% restricted_names)

    statistic <- lmtest::coeftest(object, vcov = vcov_matrix)[restricted_coef[1], "t value"]
    names(statistic) <- "t"

    if ((exact == FALSE) & (kx <= 10) & any(c(is.null(alpha), alpha %in% alpha_short_seq))) {
        caselatin <- ifelse(case == 1, "i",
                     ifelse(case == 3, "iii", "v"))
        if (!is.null(alpha)) {
            if (alpha %in% c(0.1, 0.05, 0.025, 0.01)) {
                bounds_pss2001 <- eval(parse(text = paste0("crit_val_bounds_pss2001$t$", caselatin))) %>%
                    dplyr::filter(alpha %in% !!alpha, k == kx) %>%
                    dplyr::select(I0, I1)
            }
            if (alpha %in% alpha_short_seq) {
                bounds <- eval(parse(text = paste0("crit_val_bounds$I0$", caselatin))) %>%
                    dplyr::filter(alpha == !!alpha, k == kx) %>%
                    dplyr::select(tI0) %>%
                    dplyr::rename(I0 = tI0)
                bounds <- cbind(bounds,
                    eval(parse(text = paste0("crit_val_bounds$I1$", caselatin))) %>%
                        dplyr::filter(alpha == !!alpha, k == kx) %>%
                        dplyr::select(tI1) %>%
                        dplyr::rename(I1 = tI1))
            }
        }
        if (pvalue == TRUE) {
            critvalbounds <- eval(parse(text = paste0("crit_val_bounds$I1$", caselatin)))
            if (statistic %in% critvalbounds$tI1) {
                p_value <- critvalbounds %>%
                    dplyr::filter(k == kx) %>%
                    dplyr::filter(tI1 == statistic) %>%
                    dplyr::select(alpha) %>%
                    unlist()
            } else {
                if (max(dplyr::filter(critvalbounds, k == kx)$tI1) < statistic) {
                    p_value <- 0.999999
                } else if (min(dplyr::filter(critvalbounds, k == kx)$tI1) > statistic) {
                    p_value <- 0.000001
                } else {
                    p_value <- critvalbounds %>%
                        dplyr::filter(k == kx) %>%
                        dplyr::filter(tI1 %in% c(max(tI1[which(tI1 < statistic)]),
                                                 min(tI1[which(tI1 > statistic)]))) %>%
                        dplyr::mutate(p_value = max(alpha) - (max(alpha)-min(alpha)) * ((max(tI1) - statistic) /
                            (max(tI1)-min(tI1)))) %>%
                        dplyr::select(p_value) %>% unlist() %>% .[1]
                }
            }
        }
    } else {
        T_asy_or_exact <- ifelse(exact == FALSE, 1000, T)
        tb <- t_bounds_sim(case = case, k = kx, alpha = alpha_long_seq,
                              T = T_asy_or_exact, R = R)
        if (!is.null(alpha)) {
            bounds <- tb[which(dplyr::near(alpha, alpha_long_seq)),]
        }
        if (pvalue == TRUE) {
            len <- length(alpha_long_seq)
            I1 <- tb$I1[1:len]
            if (statistic %in% I1) {
                p_value <- alpha_long_seq[statistic == I1]
            } else { # linear interpolation
                if (max(I1) < statistic) {
                    p_value <- 0.999999
                } else if (min(I1) > statistic) {
                    p_value <- 0.000001
                } else {
                    interp_cond <- I1 %in% c(max(I1[which(I1 < statistic)]),
                                              min(I1[which(I1 > statistic)]))
                    interp_I1 <- I1[interp_cond]
                    interp_alpha <- alpha_long_seq[interp_cond]
                    p_value <- max(interp_alpha) - (max(interp_alpha)-min(interp_alpha)) *
                               ((max(interp_I1) - statistic) / (max(interp_I1)-min(interp_I1)))
                }
            }
        }
    }
    method <- "Bounds t-test for no cointegration"
    alternative <- "Possible cointegration"
    rval <- list(method = method, alternative = alternative, statistic = statistic,
                 null.value = nullvalue, data.name = dname)
    tab <- data.frame(statistic = statistic)
    if (!is.null(alpha)) {
        parameters <- c(bounds$I0, bounds$I1)
        names(parameters) <- c("Lower-bound I(0)", "Upper-bound I(1)")
        tab <- cbind(tab, "Lower-bound I(0)" = parameters[1], "Upper-bound I(1)" = parameters[2],
                                  alpha = alpha)
        rval$parameters <- parameters
        if ((alpha %in% c(0.1, 0.05, 0.025, 0.01)) & exact == FALSE) {
            PSS2001parameters <- c(bounds_pss2001$I0, bounds_pss2001$I1)
            names(PSS2001parameters) <- c("Lower-bound I(0)", "Upper-bound I(1)")
            rval$PSS2001parameters <- PSS2001parameters
        }
    }
    if (pvalue == TRUE) {
        tab <- cbind(tab, p.value = p_value)
        rval$p.value <- p_value
    }
    rval$tab <- tab
    rval <- structure(rval, class = c("htest", class(rval)))
    return(rval)
}

test_that("case 1", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union -1 | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(5,4,5,1,6))
    bounds_test <- bounds_f_test(ardl_model, case = 1)
    expected_tab <- data.frame(statistic=4.5677603, p.value=0.0082357202)
    row.names(expected_tab) <- "F"

    expect_equal(bounds_test$tab, expected_tab)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # bounds_t_test
    bounds_test <- bounds_t_test(ardl_model, case = 1)
    expected_tab <- data.frame(statistic=1.1355325, p.value=0.99210825)
    row.names(expected_tab) <- "t"
    expect_equal(bounds_test$tab, expected_tab)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))
})

test_that("case 4", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    bounds_test <- bounds_f_test(ardl_model, case = 4)
    expected_tab <- data.frame(statistic=4.6919878, p.value=0.014067839)
    row.names(expected_tab) <- "F"

    expect_equal(bounds_test$tab, expected_tab)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))
})

test_that("test = 'Chisq'", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    bounds_test <- bounds_f_test(ardl_model, case = 3, test = "Chisq")
    expected_tab <- data.frame(statistic=27.750442, p.value=0.0044907254)
    row.names(expected_tab) <- "Chisq"

    expect_equal(bounds_test$tab, expected_tab)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))
})

test_that("uecm input", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    bounds_test <- bounds_f_test(uecm(ardl_model), case = 3)
    expected_tab <- data.frame(statistic=5.5500885, p.value=0.0044907254)
    row.names(expected_tab) <- "F"

    expect_equal(bounds_test$tab, expected_tab)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # bounds_t_test
    bounds_test <- bounds_t_test(uecm(ardl_model), case = 3)
    expected_tab <- data.frame(statistic=-3.914273, p.value=0.06086908)
    row.names(expected_tab) <- "t"
})

test_that("exact = FALSE & kx <= 10 & alpha %in% c(0.1, 0.05, 0.025, 0.01)", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    bounds_test <- bounds_f_test(ardl_model, case = 3, alpha = 0.05)
    expected_tab <- data.frame(statistic=5.5500885,
                               "Lower-bound I(0)" = 2.8942417, "Upper-bound I(1)" = 4.0273194,
                               alpha = 0.05, p.value=0.0044907254)
    row.names(expected_tab) <- "F"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # & PSS2001 critical values
    expected_critvals <- c("Lower-bound I(0)" = 2.86, "Upper-bound I(1)" = 4.01)
    expect_equal(bounds_test$PSS2001parameters, expected_critvals)

    # bounds_t_test
    bounds_test <- bounds_t_test(ardl_model, case = 3, alpha = 0.05)
    expected_tab <- data.frame(statistic=-3.9142734,
                               "Lower-bound I(0)" = -2.8649811, "Upper-bound I(1)" = -4.0002457,
                               alpha = 0.05, p.value=0.060869085)
    row.names(expected_tab) <- "t"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # & PSS2001 critical values
    expected_critvals <- c("Lower-bound I(0)" = -2.86, "Upper-bound I(1)" = -3.99)
    expect_equal(bounds_test$PSS2001parameters, expected_critvals)

    # & test = "Chisq" & case %in% c(1, 3, 5)
    bounds_test <- bounds_f_test(ardl_model, case = 3, alpha = 0.1, test = "Chisq")

    expected_tab <- data.frame(statistic=27.750442,
                               "Lower-bound I(0)" = 12.310873, "Upper-bound I(1)" = 17.6785096,
                               alpha = 0.1, p.value=0.0044907254)
    row.names(expected_tab) <- "Chisq"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # & PSS2001 critical values
    expected_critvals <- c("Lower-bound I(0)" = 12.25, "Upper-bound I(1)" = 17.60)
    expect_equal(bounds_test$PSS2001parameters, expected_critvals)

    # & test = "Chisq" & case %in% c(2, 4)
    bounds_test <- bounds_f_test(ardl_model, case = 2, alpha = 0.01, test = "Chisq")

    expected_tab <- data.frame(statistic=66.391527,
                               "Lower-bound I(0)" = 20.0125226, "Upper-bound I(1)" = 26.16848,
                               alpha = 0.01, p.value=1e-06)
    row.names(expected_tab) <- "Chisq"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # & PSS2001 critical values
    expected_critvals <- c("Lower-bound I(0)" = 19.74, "Upper-bound I(1)" = 26.22)
    expect_equal(bounds_test$PSS2001parameters, expected_critvals)
})

test_that("exact = FALSE & kx <= 10 & alpha %in% c(0.005 0.075 0.150 0.200) & test = Chisq", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union + trend(w) | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    bounds_test <- bounds_f_test(ardl_model, case = 5, alpha = 0.15, test="Chisq")
    expected_tab <- data.frame(statistic=19.567343,
                               "Lower-bound I(0)" = 13.83472, "Upper-bound I(1)" = 18.6088492,
                               alpha = 0.15, p.value=0.118128143)
    row.names(expected_tab) <- "Chisq"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # bounds_t_test (not test = Chisq, only t-test)
    bounds_test <- bounds_t_test(ardl_model, case = 5, alpha = 0.15)
    expected_tab <- data.frame(statistic=-2.63792694,
                               "Lower-bound I(0)" = -2.9359784, "Upper-bound I(1)" = -3.816559,
                               alpha = 0.15, p.value=0.61754341)
    row.names(expected_tab) <- "t"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))

    # case %in% c(2, 4)
    bounds_test <- bounds_f_test(ardl_model, case = 4, alpha = 0.15, test="Chisq")
    expected_tab <- data.frame(statistic=28.151927,
                               "Lower-bound I(0)" = 14.7611036, "Upper-bound I(1)" = 19.5069829,
                               alpha = 0.15, p.value=0.0140678391)
    row.names(expected_tab) <- "Chisq"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=1000))
})

test_that("NOT exact = FALSE & kx <= 10 & alpha %in% c(0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.150, 0.200)", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    set.seed(2022)
    bounds_test <- bounds_f_test(ardl_model, case = 3, alpha = 0.06, exact = TRUE, R= 10000)
    expected_tab <- data.frame(statistic=5.5500885,
                               "Lower-bound I(0)" = 2.8971731, "Upper-bound I(1)" = 4.0672186,
                               alpha = 0.06, p.value=0.007782627)
    row.names(expected_tab) <- "F"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=104))

    # bounds_t_test
    set.seed(2022)
    bounds_test <- bounds_t_test(ardl_model, case = 3, alpha = 0.06, exact = TRUE, R= 10000)
    expected_tab <- data.frame(statistic=-3.9142734,
                               "Lower-bound I(0)" = -2.8126763, "Upper-bound I(1)" = -3.9143047,
                               alpha = 0.06, p.value=0.060005925)
    row.names(expected_tab) <- "t"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=104))

    # test = Chisq
    set.seed(2022)
    bounds_test <- bounds_f_test(ardl_model, case = 3, alpha = 0.06, exact = TRUE,
                                 R= 10000, test = "Chisq")
    expected_tab <- data.frame(statistic=27.750442,
                               "Lower-bound I(0)" = 14.4858655, "Upper-bound I(1)" = 20.336093,
                               alpha = 0.06, p.value=0.007782627)
    row.names(expected_tab) <- "Chisq"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=104))

    # Clearly rejects H0 (p.value=1e-06)
    set.seed(2022)
    bounds_test <- bounds_f_test(ardl_model, case = 2, alpha = 0.06, exact = TRUE, R= 10000)
    expected_tab <- data.frame(statistic=11.0652544,
                               "Lower-bound I(0)" = 2.59362986, "Upper-bound I(1)" = 3.57422109,
                               alpha = 0.06, p.value=1e-06)
    row.names(expected_tab) <- "F"

    expect_equal(bounds_test$tab, expected_tab, ignore_attr = TRUE)
    expect_equal(bounds_test$null.value, c(k=4, T=104))
})

test_that("exact = TRUE", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    set.seed(2022)
    bounds_test <- bounds_f_test(ardl_model, case = 3, exact = TRUE, R = 10000)
    expected_tab <- data.frame(statistic=5.5500885, p.value=0.007782627)
    row.names(expected_tab) <- "F"

    expect_equal(bounds_test$tab, expected_tab)
    expect_equal(bounds_test$null.value, c(k=4, T=104))

    # bounds_t_test
    set.seed(2022)
    bounds_test <- bounds_t_test(ardl_model, case = 3, exact = TRUE, R = 10000)
    expected_tab <- data.frame(statistic=-3.9142734, p.value=0.060005925)
    row.names(expected_tab) <- "t"

    expect_equal(bounds_test$tab, expected_tab)
    expect_equal(bounds_test$null.value, c(k=4, T=104))
})

test_that("is.null(alpha) & (pvalue == FALSE) causes expected error", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))

    expect_error(bounds_f_test(ardl_model, case = 3, pvalue = FALSE),
                 "Specify an 'alpha' level, set 'pvalue' = TRUE or both")
    # bounds_t_test
    expect_error(bounds_t_test(ardl_model, case = 3, pvalue = FALSE),
                 "Specify an 'alpha' level, set 'pvalue' = TRUE or both")
})

test_that("alpha is acceptable value", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))

    expect_error(bounds_f_test(ardl_model, case = 3, alpha = 0.00000005),
                 paste0("alpha must be either 'NULL' or one of the numbers produced by the following code:\n",
                        "c\\(seq\\(0.001, 0.1, by = 0.001\\), seq\\(0.11, 0.99, by = 0.01\\)\\)"))
    # bounds_t_test
    expect_error(bounds_t_test(ardl_model, case = 3, alpha = 0.00000005),
                 paste0("alpha must be either 'NULL' or one of the numbers produced by the following code:\n",
                        "c\\(seq\\(0.001, 0.1, by = 0.001\\), seq\\(0.11, 0.99, by = 0.01\\)\\)"))
})

test_that("case %in% c(2, 4) causes expected error", {
    ardl_model <- ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                       data = PSS2001, start = c(1972, 01),
                       order=c(6,1,5,4,5))
    expect_error(bounds_t_test(ardl_model, case = 2),
                 "The t-bounds test applies only when 'case' is either 1, 3 or 5")
})

#' Formula parser
#'
#' It parses the formula and separates the dependent, independent and fixed
#' variables and also the constant and linear trends (if present).
#'
#' The notation we follow (e.g., using y, x, z, w etc.) is according to
#' \cite{Pesaran et al. (2001)}.
#'
#' The \code{formula} should contain only variables that exist in the data
#' provided through \code{data} plus some additional functions supported by
#' \code{\link[dynlm]{dynlm}} (i.e., \code{trend()}).
#'
#' You can also specify fixed variables that are not supposed to be lagged (e.g.
#' dummies etc.) simply by placing them after \code{|}. For example, \code{y ~
#' x1 + x2 | z1 + z2} where \code{z1} and \code{z2} are the fixed variables and
#' should not be considered in \code{order}. Note that the \code{|} notion
#' should not be confused with the same notion in \code{dynlm} where it
#' introduces instrumental variables.
#'
#' @param formula A "formula" describing the linear model. Details for model
#'   specification are given under 'Details'.
#' @param colnames_data A character vector containing the colnames of the data
#'   used in the formula (usually via \code{colnames(data)}).
#'
#' @return A list containing other lists with the names of the dependent,
#'   independent and fixed variables, the constant and linear trends and the
#'   number of variables in each category.
#'
#' @section References: Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds
#'   testing approaches to the analysis of level relationships. \emph{Journal of
#'   Applied Econometrics}, 16(3), 289-326
#'
#' @seealso \code{\link{parse_order}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

parse_formula <- function(formula, colnames_data) {

    # turn formula (from class = formula) into character
    formula <- Reduce(paste, deparse(formula))

    # split at fixed_part at |
    formula <- stringr::str_split(formula, pattern = "[|]")[[1]]
    w_part <- list(var = "")
    if (length(formula) == 1) {
        fixed_part <- list(var = "")
    } else if (length(formula) > 2) {
        stop("Formula can not contain more than one '|' symbols.", call. = FALSE)
    } else {
        # split fixed_part at + and trim white spaces
        fixed_part <- stringr::str_split(formula[2], pattern = "[+]")[[1]] %>%
            stringr::str_trim(side = "both") %>%
            list(var = .)
        if (fixed_part$var[1] == "") stop("The '|' symbol must be followed by at least one variable name.", call. = FALSE)
        # check if linear/non-constant intercept exists in fixed_part
        temp <- stringr::str_sub(stringr::str_replace_all(fixed_part$var, " ", ""), 1, 6) %in% "trend(" |
            stringr::str_sub(stringr::str_replace_all(fixed_part$var, " ", ""), 1, 5) %in% "time("
        if (sum(temp) != 0) {
            w_part$var <- fixed_part$var[temp]
            if (sum(temp) == length(fixed_part$var)) {
                fixed_part$var <- ""
            } else {
                fixed_part$var <- fixed_part$var[!(temp)]
            }
        }
        rm(temp)
    }

    formula <- stringr::str_split(formula[1], pattern = "[~]")[[1]] # split at ~
    if (length(formula) == 1) {
        stop("The formula must contain the '~' symbol.", call. = FALSE)
    } else if (length(formula) > 2){
        stop("The formula must contain exactly one '~' symbol.", call. = FALSE)
    }

    # split y_part at + (if any) and trim white spaces
    y_part <- stringr::str_split(formula[1], pattern = "[+]")[[1]] %>%
        stringr::str_trim(side = "both") %>%
        list(var = .)
    if ((length(y_part$var) > 1) | y_part$var[1] == "") {
        stop("The formula must contain exactly one dependent variable.", call. = FALSE)
    }

    # identify if there is an interpect exclusion
    formula = stringr::str_split(formula[2], pattern = "[-]")[[1]] %>%
        stringr::str_trim(side = "both")
    if (length(formula) == 2) {
        if (stringr::str_sub(formula[2], start = 1, end = 1) == "1") {
            if (w_part$var[1] != "") { # constant intercept exclusion
                w_part$var <- c("- 1", w_part$var)
            } else {
                w_part$var <- "- 1"
            }
        } else {
            stop("The formula can only accept the term '-1' for the exclusion of constant intercept.", call. = FALSE)
        }
    } else if (length(formula) > 2) {
        stop("The formula can only accept one '-1' constant intercept exclusion term.", call. = FALSE)
    }

    # split x_part at + and trim white spaces
    if (length(formula) != 1) {
        formula = paste0(formula[1], stringr::str_sub(formula[2], start = 2, end = stringr::str_count(formula[2])))
    }
    x_part <- stringr::str_split(formula, pattern = "[+]")[[1]] %>%
        stringr::str_trim(side = "both") %>%
        list(var = .)
    if (x_part$var[1] == "") {
        warnings("The formula contains only an AR (Autoregressive) part.")
    }
    # check if linear/non-constant intercept exists in x_part
    temp <- stringr::str_sub(stringr::str_replace_all(x_part$var, " ", ""), 1, 6) %in% "trend(" |
        stringr::str_sub(stringr::str_replace_all(x_part$var, " ", ""), 1, 5) %in% "time("
    if (sum(temp) != 0) {
        if (w_part$var[1] != "") {
            w_part$var <- c(w_part$var, x_part$var[temp])
        } else {
            w_part$var <- x_part$var[temp]
        }
        if (sum(temp) == length(x_part$var)) {
            x_part$var <- ""
        } else {
            x_part$var <- x_part$var[!(temp)]
        }
    }
    rm(temp)

    # create the union set of variables
    if (y_part$var[1] == "") y_part$var <- c()
    if (x_part$var[1] == "") x_part$var <- c()
    if (fixed_part$var[1] == "") fixed_part$var <- c()
    if (w_part$var[1] == "") w_part$var <- c()
    z_part <- list(var = c(y_part$var, x_part$var))
    kz <- length(z_part$var)
    kx <- length(x_part$var)
    temp <- stringr::str_sub(stringr::str_replace_all(w_part$var, " ", ""), 1, 5)
    kw <- if ((length(w_part$var) == 0) | (any(c("trend", "time(") %in% temp) & ("-1" %in% temp))) {
        1
    } else if (any(c("trend", "time(") %in% temp) & !("-1" %in% temp)) {
        2
    } else if (!any(c("trend", "time(") %in% temp) & ("-1" %in% temp)) {
        0
    }
    rm(temp)
    kfixed <- length(fixed_part$var)

    # check for correct function inputs
    # check if variables exist
    if (sum(!(c(z_part$var, fixed_part$var) %in% colnames_data)) >= 1) {
        stop(c("Variable(s) ",
                   paste0("'", c(z_part$var, fixed_part$var)[!(c(z_part$var, fixed_part$var) %in% colnames_data)], sep = "', "),
                   "not found in 'data'"), call. = FALSE)
    }

    return_list <- list(y_part = y_part, x_part = x_part, z_part = z_part,
                        w_part = w_part, fixed_part = fixed_part,
                        kz = kz, kx = kx, kw = kw, kfixed = kfixed)
    return(return_list)

}

#' Order parser
#'
#' It parses the order and checks the integrity of the order input.
#'
#' @param orders A numeric vector of the same length as the total number of
#'   variables (excluding the fixed ones). If the input is \code{order} or
#'   \code{max_order} it should only contain positive integers or 0. If the
#'   input is \code{fixed_order} it should also contain the value '-1'
#'   indicating that a specific order should not be fixed. An integer could be
#'   provided if all variables are of the same order (or all '-1' in the case of
#'   \code{fixed_order}).
#' @param order_name The name of the function argument that is passed into
#'   \code{order}.
#' @param var_names The names of the variables corresponding to the orders.
#' @param kz An integer. The number of dependent and independent variables.
#' @param restriction When the input in \code{orders} is either \code{order} or
#'   \code{max_order} it should be \code{FALSE} (default). When the input is
#'   \code{fixed_order} it should be '-1' indicating that the input in
#'   \code{orders} is a restriction for the 'order' of the model (either upper
#'   bound or fixed order).
#'
#' @return A numeric vector of the same length as the total number of variables
#'   (excluding the fixed ones).
#'
#' @seealso \code{\link{parse_formula}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

parse_order <- function(orders, order_name, var_names, kz, restriction = FALSE) {
    if (restriction == FALSE) {
        restriction = 0
    }

    # check if orders length is correct
    if (!(length(orders) %in% c(1, kz))) {
        stop("The length of '", order_name, "' is not correct", call. = FALSE)
    }
    # check for AR(0) models
    if (orders[1] == 0) {
        stop("AR(0) models are not allowed. The order of the dependent variable should be a positive integer", call. = FALSE)
    }
    # check for negatives (or other restriction e.g. -1) and decimals in orders
    if ((any(orders < restriction)) | (sum(orders %% 1) != 0)) {
        if (restriction == 0) {
            stop("'", order_name, "' can only contain positive integers or 0", call. = FALSE)
        } else if (restriction == (-1)) {
            stop("'", order_name, "' can only contain positive integers, 0 or '-1'", call. = FALSE)
        }
    }

    if (length(orders) == 1) {
        orders <- rep(orders[1], kz)
    }
    names(orders) <- var_names

    return(orders)
}

#' Case parser
#'
#' It parses the 'case' and checks the integrity of the 'case' input and the
#' compatibility with the formula.
#'
#' @inherit recm details
#'
#' @param parsed_formula A list containing the formula parts as returned from
#'   \code{\link{parse_formula}}.
#' @inheritParams recm
#'
#' @return An integer from 1-5 representing the case.
#'
#' @inheritSection recm References
#' @seealso \code{\link{parse_formula}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

parse_case <- function(parsed_formula, case) {

    if ((length(case) != 1) | !(case[1]) %in% c(1:5, 'n', 'rc', 'uc', 'ucrt', 'ucut')) {
        stop("'case' must be a number between 1 and 5 or one of the 'n', 'rc', 'uc', 'ucrt' or 'ucut'", call. = FALSE)
    } else if (case == "n") {
        case = 1
    } else if (case == "rc") {
        case = 2
    } else if (case == "uc") {
        case = 3
    } else if (case == "ucrt") {
        case = 4
    } else if (case == "ucut") {
        case = 5
    }

    w_check <- stringr::str_sub(stringr::str_replace_all(parsed_formula$w_part$var, " ", ""), 1, 5)
    if ((case == 1) & (any(c("trend", "time(") %in% w_check ) | !("-1" %in% w_check))) {
        stop("Trying to impose case 1 (no constant, no linear trend) but the underlying ARDL model includes at least one of them", call. = FALSE)
    } else if ((case %in% c(2,3)) & any(c("-1", "trend", "time(") %in% w_check)) {
        stop("Trying to impose case ", ifelse(case == 2,
            paste0(case, " (restricted constant, no linear trend)"),
            paste0(case, " (unrestricted constant, no linear trend)")), " but the underlying ARDL model either doesn't include a constant or includes a trend", call. = FALSE)
    } else if ((case %in% c(4,5)) & (!any(c("-1", "trend", "time(") %in% w_check))) {
        stop("Trying to impose case ", ifelse(case == 4,
            paste0(case, " (unrestricted constant, restricted linear trend)"),
            paste0(case, " (unrestricted constant, unrestricted linear trend)")), " but the underlying ARDL model doesn't include one or both of them", call. = FALSE)
    }

    return(case)
}

#' Predict method for ARDL model fits
#'
#' Makes predictions based on fitted ARDL models, extending the
#' generic function \code{\link[stats]{predict}} for use with
#' \code{\link{ardl}}. It is designed to take advantage of lags of the dependent
#' variable (AR components) from the fitted model and new values of the
#' independent variable(s) to calculate predictions.
#'
#' @param object A fitted model of \code{\link[base]{class}} \code{\link{ardl}}.
#' @param newdata A time series object (\code{\link[stats]{ts}},
#' \code{\link[zoo]{zoo}}, \code{\link[zoo]{zooreg}}) or a
#' \code{\link[base]{data.frame}} containing the new values of the independent
#' variables \eqn{x} required to make predictions.
#' @param ... further arguments passed to or from other methods.
#'
#' @details
#' \code{predict.ardl} works recursively to calculate predictions based on
#' \eqn{y_{t-1}, ..., y_{t-p}} values contained within a fitted model, and new
#' values of \eqn{x_{T+1}, ..., x_{T+h}}, where \eqn{T} denotes the last observation
#' of \eqn{t} from the original data and \eqn{h} is the number of new
#' observations that will be used to calculate predictions. These are supplied
#' through the \code{newdata} argument. It is important to note that for a given
#' prediction \eqn{\hat{y}_{T+1}}, the independent variables provided through
#' \code{newdata}, are expected to follow consecutively from the last observation
#' \eqn{x_{T}} used to create the model, i.e. \eqn{x_{T+1}, ..., x_{T+h}},
#' with no gaps. If a time series object is passed to the function, the frequency
#' and whether the observations follow consecutively will be checked.
#'
#' @section Calculation example:
#'
#' We want to predict \eqn{\hat{y}_{T+1}} given an \eqn{ARDL(2,2)} model:
#'
#' \deqn{
#'      \hat{y}_{T+1} = Z \boldsymbol{\beta}
#' }
#' where \eqn{Z} is our design matrix and \eqn{\beta} is a vector of our model
#' coefficients. We need 2 lags for our dependent and independent variables:
#' \eqn{y_T}, \eqn{y_{T-1}} and \eqn{X_T}, \eqn{X_{T-1}}.
#' Plus, \eqn{X_{T+1}} (from \code{newdata}), which in our design matrix is
#' equivalent to time \eqn{t} in our original model:
#' \deqn{
#' Z =
#'     \left[
#'         \begin{array}{c|cccccc}
#'         & \mathbf{c} & \mathbf{y_{t-1}} & \mathbf{y_{t-2}} & \mathbf{x_t} &
#'         \mathbf{x_{t-1}} & \mathbf{x_{t-2}} \\
#'         \hline
#'         T\!+\!1 & 1 & y_{T} & y_{T-1} & x_{T+1} & x_{T} & x_{T-1}
#'         \end{array}
#'         \right]
#' }
#' We proceed to the next iteration, predicting \eqn{\hat{y}_{T+2}}, using the
#' previously calculated value of \eqn{\hat{y}_{T+1}}:
#' \deqn{
#' Z =
#' \left[
#'     \begin{array}{c|cccccc}
#'     & \mathbf{c} & \mathbf{y_{t-1}} & \mathbf{y_{t-2}} & \mathbf{x_t} & \mathbf{x_{t-1}} & \mathbf{x_{t-2}} \\
#'     \hline
#'     T\!+\!2 & 1 & y_{T+1} & y_{T} & x_{T+2} & x_{T+1} & x_{T}
#'     \end{array}
#'     \right]
#' }
#'
#' Iteration continues until we've calculated all the necessary values of
#' \eqn{\hat{y}_{T+h}} for \code{newdata}.
#'
#' @return \code{predict.ardl} returns values for \eqn{\hat{y}_{T+1}, ...,
#' \hat{y}_{T+h}} equal in length to the number of rows in \code{newdata},
#' as a \code{\link[base]{vector}}.
#'
#' @seealso \code{\link{ardl}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @author Daniel Finnan, \email{dan@@custom-made.org.uk}
#' @keywords models ts
#' @examples
#' data(denmark)
#' ## Estimate an ARDL(2,1,2,1,1) model, using a subset of the denmark data
#' ardl_21211 <- ardl(LRM ~ LRY + LPY + IBO + IDE, data = denmark,
#'                    order = c(2,1,2,1,1),
#'                    start = "1974 Q1", end = "1986 Q2")
#' ## Make predictions based on the remainder of the denmark data not used in
#' ## training the model
#' predict_data <- window(denmark, start = "1986 Q3", end = "1987 Q3")
#' ## Drop the dependent variable, since we're predicting it
#' predict_data <- predict_data[, c("LRY", "LPY", "IBO", "IDE")]
#' predictions <- predict(ardl_21211, predict_data)
#'
#' ## Estimate a ARDL(4,4,4,4) model with a linear trend and dummies
#' ## Create dummies
#' d_74Q1_75Q3 <- ifelse(time(denmark) >= "1974 Q1" & time(denmark) <= "1975 Q3", 1, 0)
#' # Add them to the data
#' denmark_dums <- cbind(denmark, d_74Q1_75Q3)
#' ## Estimate the model
#' ardl_4444 <- ardl(LRM ~ LRY + LPY + IBO + trend(LRM) | d_74Q1_75Q3,
#'                   data = denmark_dums,
#'                   order = c(4,4,4,4),
#'                   start = "1974 Q1", end = "1986 Q2")
#' ## Take the remaining data not used in training the model
#' predict_data <- window(denmark_dums, start = "1986 Q3", end = "1987 Q3")
#' ## Drop the dependent variable
#' predict_data <- predict_data[, c("LRY", "LPY", "IBO", "d_74Q1_75Q3")]
#' ## Compute the predictions
#' predictions <- predict(ardl_4444, predict_data)
#'
#' @method predict ardl
#' @export
predict.ardl <- function(object, newdata, ...) {
    newdata_class <- class(newdata)
    if (missing(newdata)) {
        stop("Need to provide some 'newdata' to predict.", call. = FALSE)
    }
    if (!any(c("ts", "zoo", "zooreg", "data.frame") %in% newdata_class)) {
        stop("'newdata' must be a 'ts', 'zoo', 'zooreg' or 'data.frame' object.", call. = FALSE)
    }
    if (!identical(class(object), c("ardl","dynlm","lm"))) {
        stop("'object' needs to be a fitted model using 'ardl' method.", call. = FALSE)
    }
    if (any(is.na(newdata))) {
        stop("'newdata' contains some NA values.", call. = FALSE)
    }
    if ((any(c("ts", "zoo", "zooreg") %in% newdata_class) && (stats::frequency(object$data) != stats::frequency(newdata)))) {
        stop("'newdata' needs to have the same frequency as the original data.", call. = FALSE)
    }
    # Consecutive series zoo/zooreg
    # Note: we need to subset data in ardl model, since it stores complete dataset, despite start/end parameter
    useddata <- object$data[zoo::index(object),]
    if (any(c("zoo", "zooreg") %in% newdata_class) && !((stats::end(useddata) + stats::deltat(useddata)) == stats::start(newdata))) {
        stop("Observations in 'newdata' need to follow original data consecutively.", call. = FALSE)
    }
    # Consecutive series ts. Note: ts can't use delta date function, for simplisicity we coerce to zoo
    if (stats::is.ts(newdata) && !((stats::end(zoo::as.zoo(useddata)) + stats::deltat(zoo::as.zoo(useddata))) == stats::start(zoo::as.zoo(newdata))) ) {
        stop("Observations in 'newdata' need to follow original data consecutively.", call. = FALSE)
    }
    # Check all necessary columns are in new data
    orig_cols <- c(object$parsed_formula$x_part$var, object$parsed_formula$fixed_part$var)
    newdata_cols <- colnames(newdata)
    if (identical(orig_cols, newdata_cols) == FALSE) {
        stop("'newdata' columns don't match original data.", call. = FALSE)
    }
    # Handling simple data.frames
    if (!any(c("ts", "zoo", "zooreg") %in% newdata_class)) {
        newdata <- stats::ts(newdata, start = 1, end = nrow(newdata),
                             frequency = 1)
    }
    kx <- object$parsed_formula$kx
    kw <- object$parsed_formula$kw
    kfixed <- object$parsed_formula$kfixed
    coeffs <- matrix(object$coefficients)
    #Because when no intercept, object$parsed_formula$w_part$var was returning "-1" instead of NULL
    if (object$parsed_formula$kw == 0) {
        w_part <- NULL
    } else {
        w_part <- object$parsed_formula$w_part$var
    }
    model_matrix <- object$model[,c(object$parsed_formula$y_part$var,
                                    w_part,
                                    object$parsed_formula$x_part$var,
                                    object$parsed_formula$fixed_part$var)]
    model_matrix <- t(matrix(unlist(model_matrix), nrow = ncol(model_matrix),
                             byrow = TRUE, dimnames = dimnames(t(model_matrix))))
    n_ahead <- nrow(newdata)
    if (kw == 2) {
        olddata_w <- zoo::coredata(object$model[,2])
        newdata_w <- cumsum(c(olddata_w[length(olddata_w)],
                              rep(diff(olddata_w, 1)[1], n_ahead)))[-1]
        olddata_w <- object$model[,2]
        newdata_w <- stats::ts(newdata_w, start = stats::end(olddata_w) + 1/stats::frequency(olddata_w),
                        frequency = stats::frequency(olddata_w))
    }
    predictor <- c()
    model_matrix_y <- model_matrix[,1]
    kw_correction <- if(object$parsed_formula$kw == 2){c(-1,-2)} else {-1}
    model_matrix_x <- model_matrix[,kw_correction]
    model_matrix_x <- model_matrix_x[,1:kx]
    for (j in 1:n_ahead) {
        Y <- c()
        design_matrix <- c()
        for (i in 0:c(object$order[1]-1)) {
            lags_y <- dplyr::lag(model_matrix_y, n = i)
            Y <- c(Y, lags_y[length(lags_y)])
        }
        for (k in 1:kx) {
            Xk <- c()
            for (i in 0:object$order[k+1]) {
                model_matrix_xk <- c(model_matrix_x[,k], newdata[1:j,k])
                lags_x <- dplyr::lag(model_matrix_xk, n = i)
                Xk <- c(Xk, lags_x[length(lags_x)])
            }
            design_matrix <- c(design_matrix, Xk)
        }
        if (kfixed != 0) {
            design_matrix <- c(design_matrix,
                               unlist(newdata[j,(kx+1):ncol(newdata)]))#Add newdata for dummies
        }
        design_matrix <- c(Y, design_matrix)
        if (kw == 2) {
            design_matrix <- c(1, newdata_w[j], design_matrix)#Add unit vector
        } else if (kw == 1) {
            design_matrix <- c(1, design_matrix)#Add unit vector and trend
        }
        predictor <- c(predictor, t(coeffs) %*% matrix(design_matrix))
        model_matrix_y <- c(model_matrix_y, predictor[length(predictor)])
    }
    # Return data.frame if newdata isn't a time series object, otherwise return
    # time series object with timestamps
    if (!any(c("ts", "zoo", "zooreg") %in% newdata_class)) {
        # Return as a named num like predict.lm
        return(stats::setNames(predictor, seq(1, n_ahead)))
    } else {
        return(cbind(newdata, predictor) %>% .[,dim(.)[2]])
    }
}

AIC_pss <- function(model){
    # maximized log-likelihood value of the model
    LLp <- stats::logLik(model)
    # number of freely estimated coefficients
    sp <- length(model$coefficients)
    c(LLp - sp)
}
BIC_pss <- function(model){
    # maximized log-likelihood value of the model
    LLp <- stats::logLik(model)
    # number of freely estimated coefficients
    sp <- length(model$coefficients)
    # number of observations
    TT <- log(stats::nobs(model))
    c(LLp - (sp/2)*TT)
}

#' Create plots for the delay multipliers
#'
#' Creates plots for the delay multipliers and their uncertainty intervals based
#' on their estimated standard errors. This is a basic
#' \code{\link[ggplot2]{ggplot}} with a few customizable parameters.
#'
#' @param multipliers A list returned from \code{\link{multipliers}},
#' in which \code{type} is a positive integer to return delay multipliers.
#' @param facets_ncol If a positive integer, it indicates the number of the
#' columns in the facet. If FALSE, each plot is created separately. The default
#' is 2.
#' @param interval If FALSE (default), no uncertainty intervals are drawn. If a
#' positive integer, the intervals are this number times the standard error. If
#' a number between 0 and 1 (e.g. 0.95), the equivalent confidence interval is
#' drawn (e.g. 95\% CI). In case of the confidence intervals, they are based on
#' the Gaussian distribution.
#' @param interval_color The color of the uncertainty intervals. Default is
#' "blue".
#' @param show.legend A logical indicating whether the interval legend is shown.
#' Default is FALSE.
#' @param xlab,ylab Names displayed at the x and y axes respectively. Default is
#' "Period" and "Delay" respectively.
#' @param ... Currently unused argument.
#'
#' @return \code{plot_delay} returns a number of \code{\link[ggplot2]{ggplot}}
#' objects.
#'
#' @seealso \code{\link{multipliers}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords iplots
#' @export
#' @examples
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' delay_mult <- multipliers(ardl_3132, type = 12, se = TRUE)
#'
#' ## Simply plot the delay multipliers -----------------------------------
#'
#' plot_delay(delay_mult)
#'
#' ## Rearrange them ------------------------------------------------------
#'
#' plot_delay(delay_mult, facets_ncol = 1)
#'
#' ## Add 1 standard deviation uncertainty intervals ----------------------
#'
#' plot_delay(delay_mult, interval = 1)
#'
#' ## Add 95% confidence intervals, change color and add legend -----------
#'
#' plot_delay(delay_mult, interval = 0.95, interval_color = "darkgrey",
#'            show.legend = TRUE)

plot_delay <- function(multipliers, facets_ncol = 2,
                       interval = FALSE, interval_color = "blue", show.legend = FALSE,
                       xlab = "Period", ylab = "Delay", ...) {
    plot_data <- function(mult, .data) {
        plot <- ggplot2::ggplot(mult, ggplot2::aes(x = .data[["Period"]], y = .data[["Delay"]])) +
            ggplot2::geom_line() +
            ggplot2::labs(x = xlab, y = ylab) +
            ggplot2::theme_minimal()
        if (!isFALSE(interval)) {
            adjustment <- ifelse((interval %% 1) == 0, interval, stats::qnorm(1-(1-interval)/2))
            plot <- plot +
                ggplot2::geom_ribbon(ggplot2::aes(ymin = .data[["Delay"]] - adjustment*.data[["Std. Error Delay"]],
                                                  ymax = .data[["Delay"]] + adjustment*.data[["Std. Error Delay"]],
                                                  fill = "Std. Error Delay"),
                                     alpha = 0.3, show.legend = show.legend) +
                ggplot2::scale_fill_manual(name = "", values = interval_color,
                                           labels = ifelse((interval %% 1) == 0, paste0(interval, " SE"),
                                                           paste0(100*interval, "%", " CI"))) +
                ggplot2::theme(legend.position = "bottom")
        }
        plot
    }

    plots <- lapply(names(multipliers), function(name) {
        plot_data(multipliers[[name]]) +
            ggplot2::labs(title = name) +
            ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0.5))
    })

    if (!isFALSE(facets_ncol)) {
        gridExtra::grid.arrange(grobs = plots, ncol = facets_ncol)
    } else {
        plots
    }
}

#' Cointegrating equation (long-run level relationship)
#'
#' Creates the cointegrating equation (long-run level relationship) providing an
#' 'ardl', 'uecm' or 'recm' model.
#'
#' @param object An object of \code{\link[base]{class}} 'ardl', 'uecm' or
#'   'recm'.
#' @param case An integer from 1-5 or a character string specifying whether the
#'   'intercept' and/or the 'trend' have to participate in the long-run level
#'   relationship (cointegrating equation) (see section 'Cases' below). If the
#'   input object is of class 'recm', \code{case} is not needed as the model is
#'   already under a certain case.
#' @param ... Currently unused argument.
#'
#' @return \code{coint_eq} returns an numeric vector containing the
#'   cointegrating equation.
#'
#' @inheritSection recm Cases
#' @inheritSection recm References
#' @seealso \code{\link{plot_lr}} \code{\link{ardl}} \code{\link{uecm}} \code{\link{recm}}
#'   \code{\link{bounds_f_test}} \code{\link{bounds_t_test}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords ts
#' @export
#' @examples
#' data(denmark)
#' library(zoo) # for cbind.zoo()
#'
#' ## Estimate the Cointegrating Equation of an ARDL(3,1,3,2) model -------
#'
#' # From an ARDL model (under case 2, restricted constant)
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' ce2_ardl <- coint_eq(ardl_3132, case = 2)
#'
#' # From an UECM (under case 2, restricted constant)
#' uecm_3132 <- uecm(ardl_3132)
#' ce2_uecm <- coint_eq(uecm_3132, case = 2)
#'
#' # From a RECM (under case 2, restricted constant)
#' # Notice that if a RECM has already been estimated under a certain case,
#' # the 'coint_eq()' can't be under different case, so no 'case' argument needed.
#' recm_3132 <- recm(uecm_3132, case = 2)
#' # The RECM is already under case 2, so the 'case' argument is no needed
#' ce2_recm <- coint_eq(recm_3132)
#'
#' identical(ce2_ardl, ce2_uecm, ce2_recm)
#'
#' ## Check for a degenerate level relationship ---------------------------
#'
#' # The bounds F-test under both cases reject the Null Hypothesis of no level relationship.
#' bounds_f_test(ardl_3132, case = 2)
#' bounds_f_test(ardl_3132, case = 3)
#'
#' # The bounds t-test also rejects the NUll Hypothesis of no level relationship.
#' bounds_t_test(ardl_3132, case = 3)
#'
#' # But when the constant enters the long-run equation (case 3)
#' # this becomes a degenerate relationship.
#' ce3_ardl <- coint_eq(ardl_3132, case = 3)
#'
#' plot_lr(ardl_3132, coint_eq = ce2_ardl, show.legend = TRUE)
#'
#' plot_lr(ardl_3132, coint_eq = ce3_ardl, show.legend = TRUE)
#' plot_lr(ardl_3132, coint_eq = ce3_ardl, facets = TRUE, show.legend = TRUE)

coint_eq <- function(object, case) {
    UseMethod("coint_eq")
}

#' @rdname coint_eq
#' @export
#'

coint_eq.recm <- function(object, ...) {
    object$data[,object$parsed_formula$y_part$var]  - stats::lag(object$data[,"ect"], 1)
}

#' @rdname coint_eq
#' @export
#'

coint_eq.default <- function(object, case) {
    if ("recm_indicator" %in% class(object)) {
        recm_indicator <- TRUE
        class(object) <- class(object)[-4]
    } else {
        recm_indicator <- FALSE
    }
    if (isTRUE(all.equal(c("ardl", "dynlm", "lm"), class(object)))) {
        object <- uecm(object)
    } else if (!(isTRUE(all.equal(c("uecm", "dynlm", "lm"), class(object))))) {
        stop(paste0("no applicable for an object of class \"", paste0(class(object), collapse = '" "'), "\""),  call. = FALSE)
    }
    parsed_formula <- object$parsed_formula
    case <- parse_case(parsed_formula = parsed_formula, case = case)
    order <- object$order
    data <- object$data
    lr_multipliers <- multipliers(object = object, type = "lr")
    lr_mult_names <- as.character(lr_multipliers$Term)
    lr_multipliers <- lr_multipliers$Estimate

    if (case == 1) {
        lr_multipliers <- c(0, 0, lr_multipliers)
        trend_var <- rep(0, nrow(data))
    } else if (case == 2) {
        lr_multipliers <- c(lr_multipliers[1], 0, lr_multipliers[2:length(lr_multipliers)])
        trend_var <- rep(0, nrow(data))
    } else if (case == 3) {
        lr_multipliers <- c(0, 0, lr_multipliers[-1])
        trend_var <- rep(0, nrow(data))
    } else if (case == 4) {
        lr_multipliers <- c(0, lr_multipliers[-1])
        is_trend_in_lr <- "trend" %in% stringr::str_sub(stringr::str_replace_all(lr_mult_names[2], " ", ""), 1, 5)
        is_time_in_lr <- "time(" %in% stringr::str_sub(stringr::str_replace_all(lr_mult_names[2], " ", ""), 1, 5)
        if (is_trend_in_lr == TRUE) {
            trimed_trend <- stringr::str_replace_all(lr_mult_names[2], " ", "")
            short_match <- stringr::str_sub(trimed_trend, nchar(trimed_trend) - 7, nchar(trimed_trend))
            long_match <- stringr::str_sub(trimed_trend, nchar(trimed_trend) - 11, nchar(trimed_trend))
            if (any(c("scale=FALSE)", "scale=F)") %in% c(short_match, long_match))) {
                trend_var <- 1:nrow(data)
            } else {
                trend_var <- (1:nrow(data)) / stats::frequency(data)
            }
        } else if (is_time_in_lr == TRUE) {
            trend_var <- stats::time(data) %>% as.numeric()
        }
    } else if (case == 5) {
        lr_multipliers <- c(0, 0, lr_multipliers[-(1:2)])
        trend_var <- rep(0, nrow(data))
    }
    unit_vector <- rep(1, nrow(data))

    # create the design matrix to compute the coint_eq (cointegrating equation in levels)
    design_matrix <- data.frame(unit_vector, trend_var,
        dplyr::as_tibble(data) %>% dplyr::select(parsed_formula$x_part$var))
    # create the coint_eq components (multipliers * variable)
    design_matrix <- lapply(1:ncol(design_matrix), function(i) design_matrix[ ,i] * lr_multipliers[i]) %>%
        as.data.frame()
    # add the dependent variable
    design_matrix <- data.frame(y = data[, parsed_formula$y_part$var], design_matrix)
    # compute the coint_eq
    coint_eq <- design_matrix %>%
        dplyr::mutate(coint_eq = rowSums(.[2:ncol(design_matrix)])) %>%
        dplyr::select(coint_eq) %>%
        .[[1]]
    # to inherit the proper time-series class and properties
    res <- cbind(data, coint_eq)
    res <- res[, "coint_eq"]
    if (recm_indicator) {
        coint_eq <- stats::ts(coint_eq, start = stats::start(data), frequency = stats::frequency(data))
        res <- list(coint_eq = coint_eq, design_matrix = design_matrix, data= data,
                    parsed_formula = parsed_formula, case = case, order = order)
    }
    return(res)
}

#' ARDL model regression
#'
#' A simple way to construct complex ARDL specifications providing just the
#' model order additional to the model formula. It uses
#' \code{\link[dynlm]{dynlm}} under the hood. \code{ardl} is a generic function
#' and the default method constructs an 'ardl' model while the other method
#' takes a model of \code{\link[base]{class}} 'uecm' and converts in into an
#' 'ardl'.
#'
#' The \code{formula} should contain only variables that exist in the data
#' provided through \code{data} plus some additional functions supported by
#' \code{\link[dynlm]{dynlm}} (i.e., \code{trend()}).
#'
#' You can also specify fixed variables that are not supposed to be lagged (e.g.
#' dummies etc.) simply by placing them after \code{|}. For example, \code{y ~
#' x1 + x2 | z1 + z2} where \code{z1} and \code{z2} are the fixed variables and
#' should not be considered in \code{order}. Note that the \code{|} notion
#' should not be confused with the same notion in \code{dynlm} where it
#' introduces instrumental variables.
#'
#' @param formula A "formula" describing the linear model. Details for model
#'   specification are given under 'Details'.
#' @param data A time series object (e.g., "ts", "zoo" or "zooreg") or a data
#'   frame containing the variables in the model. In the case of a data frame,
#'   it is coerced into a \code{\link[stats]{ts}} object with \code{start = 1},
#'   \code{end = nrow(data)} and \code{frequency = 1}. If not found in data, the
#'   variables are NOT taken from any environment.
#' @param order A specification of the order of the ARDL model. A numeric vector
#'   of the same length as the total number of variables (excluding the fixed
#'   ones, see 'Details'). It should only contain positive integers or 0. An
#'   integer could be provided if all variables are of the same order.
#' @param start Start of the time period which should be used for fitting the
#'   model.
#' @param end End of the time period which should be used for fitting the model.
#' @param ... Additional arguments to be passed to the low level regression
#'   fitting functions.
#'
#' @return \code{ardl} returns an object of \code{\link[base]{class}}
#'   \code{c("ardl", "dynlm", "lm")}. In addition, attributes 'order', 'data',
#'   'parsed_formula' and 'full_formula' are provided.
#'
#' @section Mathematical Formula:
#' The general form of an \eqn{ARDL(p,q_{1},\dots,q_{k})}{ARDL(p,q1,...,qk)} is:
#' \deqn{y_{t} = c_{0} + c_{1}t + \sum_{i=1}^{p}b_{y,i}y_{t-i} +
#' \sum_{j=1}^{k}\sum_{l=0}^{q_{j}}b_{j,l}x_{j,t-l} + \epsilon_{t}}
#'
#' @seealso \code{\link{uecm}}, \code{\link{recm}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords models ts
#' @export
#' @examples
#' data(denmark)
#'
#' ## Estimate an ARDL(3,1,3,2) model -------------------------------------
#'
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' summary(ardl_3132)
#'
#' ## Add dummies or other variables that should stay fixed ---------------
#'
#' d_74Q1_75Q3 <- ifelse(time(denmark) >= 1974 & time(denmark) <= 1975.5, 1, 0)
#'
#' # the date can also be setted as below
#' d_74Q1_75Q3_ <- ifelse(time(denmark) >= "1974 Q1" & time(denmark) <= "1975 Q3", 1, 0)
#' identical(d_74Q1_75Q3, d_74Q1_75Q3_)
#' den <- cbind(denmark, d_74Q1_75Q3)
#' ardl_3132_d <- ardl(LRM ~ LRY + IBO + IDE | d_74Q1_75Q3,
#'                     data = den, order = c(3,1,3,2))
#' summary(ardl_3132_d)
#' compare <- data.frame(AIC = c(AIC(ardl_3132), AIC(ardl_3132_d)),
#'                       BIC = c(BIC(ardl_3132), BIC(ardl_3132_d)))
#' rownames(compare) <- c("no dummy", "with dummy")
#' compare
#'
#' ## Estimate an ARDL(3,1,3,2) model with a linear trend -----------------
#'
#' ardl_3132_tr <- ardl(LRM ~ LRY + IBO + IDE + trend(LRM),
#'                      data = denmark, order = c(3,1,3,2))
#'
#' # Alternative time trend specifications:
#' # time(LRM)                 1974 + (0, 1, ..., 55)/4 time(data)
#' # trend(LRM)                (1, 2, ..., 55)/4        (1:n)/freq
#' # trend(LRM, scale = FALSE) (1, 2, ..., 55)          1:n
#'
#' ## Subsample ARDL regression (start after 1975 Q4) ---------------------
#'
#' ardl_3132_sub <- ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                       order = c(3,1,3,2), start = "1975 Q4")
#'
#' # the date can also be setted as below
#' ardl_3132_sub2 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                        order = c(3,1,3,2), start = c(1975,4))
#' identical(ardl_3132_sub, ardl_3132_sub2)
#' summary(ardl_3132_sub)
#'
#' ## Ease of use ---------------------------------------------------------
#'
#' # The model specification of the ardl_3132 model can be created as easy as order=c(3,1,3,2)
#' # or else, it could be done using the dynlm package as:
#' library(dynlm)
#' m <- dynlm(LRM ~ L(LRM, 1) + L(LRM, 2) + L(LRM, 3) + LRY + L(LRY, 1) + IBO + L(IBO, 1) +
#'            L(IBO, 2) + L(IBO, 3) + IDE + L(IDE, 1) + L(IDE, 2), data = denmark)
#' identical(m$coefficients, ardl_3132$coefficients)
#'
#' # The full formula can be extracted from the ARDL model, and this is equal to
#' ardl_3132$full_formula
#' m2 <- dynlm(ardl_3132$full_formula, data = ardl_3132$data)
#' identical(m$coefficients, m2$coefficients)
#'
#' ## Post-estimation testing ---------------------------------------------
#'
#' # See examples in the help file of the uecm() function
#'

ardl <- function(...) {
    UseMethod("ardl")
}

#' @rdname ardl
#'
#' @param object An object of \code{\link[base]{class}} 'uecm'.
#'
#' @export
#'

ardl.uecm <- function(object, ...) {
    parsed_formula <- object$parsed_formula
    order <- object$order
    data <- object$data
    start <- start(object)
    end <- end(object)
    unbuild_formula <- paste0(paste0(parsed_formula$y_part$var, "~"),
                              paste0(parsed_formula$x_part$var, collapse = "+"),
                              if (length(parsed_formula$w_part$var) == 0) { # Case I
                                ""
                              } else if (parsed_formula$w_part$var[1] == "- 1") { # Cases II & III
                                  paste0(parsed_formula$w_part$var)
                              } else { # Cases IV & V
                                  paste0("+", parsed_formula$w_part$var)
                              },
                              ifelse(length(parsed_formula$fixed_part$var) != 0,
                                     paste0("|", paste0(parsed_formula$fixed_part$var,
                                                        collapse = "+")), ""))

    return(ardl(stats::formula(unbuild_formula), data = data, order = order, start = start, end = end))
}

#' @rdname ardl
#'
#' @export
#'

ardl.default <- function(formula, data, order, start = NULL, end = NULL, ...) {
    if (!any(c("ts", "zoo", "zooreg") %in% class(data))) {
        data <- stats::ts(data, start = 1, end = nrow(data), frequency = 1)
    }

    parsed_formula <- parse_formula(formula = formula, colnames_data = colnames(data))
    order <- parse_order(orders = order, order_name = "order",
                         var_names = parsed_formula$z_part$var, kz = parsed_formula$kz)
    ardl_formula <- build_ardl_formula(parsed_formula = parsed_formula, order = order)

    full_formula <- formula(ardl_formula$full)
    ardl_model <- dynlm::dynlm(full_formula, data = data, start = start, end = end, ...)
    # for model compatibility in the global env
    attr(ardl_model$terms, ".Environment") <- .GlobalEnv
    attr(attr(ardl_model$model, "terms"), ".Environment") <- .GlobalEnv
    attr(full_formula, ".Environment") <- .GlobalEnv

    ardl_model$order <- order
    ardl_model$data <- data
    ardl_model$parsed_formula <- parsed_formula
    ardl_model$full_formula <- full_formula

    ardl_model <- structure(
        ardl_model,
        class = c("ardl", class(ardl_model))
    )

    return(ardl_model)
}

test_that("Combination of max_order, fixed_order, starting_order and selection", {
    ardl_model <- auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                            data = PSS2001, start = c(1972, 01),
                            max_order = 6, fixed_order = c(-1,1,-1,-1,-1), starting_order = 5,
                            selection = "AIC_pss", selection_minmax = "max")

    expect_equal(ardl_model$top_orders[,-6],
                 data.frame(w=c(6,6,6,6,6,5,5,5,5,5),
                            Prod=c(1,1,1,1,1,1,1,1,1,1),
                            UR=c(5,5,5,6,6,5,5,5,6,6),
                            Wedge=c(4,5,4,4,5,4,5,4,4,5),
                            Union=c(5,5,6,5,5,5,5,6,5,5)))
    expect_equal(ardl_model$best_model, ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                                             data = PSS2001, start = c(1972, 01),
                                             order=c(6,1,5,4,5)))

    # search_type = "vertical"
    ardl_model_vert <- auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                            data = PSS2001, start = c(1972, 01),
                            max_order = 6, fixed_order = c(-1,1,-1,-1,-1), starting_order = 5,
                            selection = "AIC_pss", selection_minmax = "max",
                            search_type = "vertical")

    expect_equal(ardl_model_vert$top_orders[,-6],
                 data.frame(w=c(6,6,6,6,5,5,5,5),
                            Prod=c(1,1,1,1,1,1,1,1),
                            UR=c(5,5,5,6,5,5,5,6),
                            Wedge=c(4,5,4,5,4,5,4,5),
                            Union=c(5,5,6,5,5,5,6,5)))
    expect_equal(ardl_model_vert$best_model, ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                                                  data = PSS2001, start = c(1972, 01),
                                                  order=c(6,1,5,4,5)))

    # check if non time-series data work
    ardl_model_df <- auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                               data = data.frame(PSS2001), start = 9,
                               max_order = 6, fixed_order = c(-1,1,-1,-1,-1), starting_order = 5,
                               selection = "AIC_pss", selection_minmax = "max")

    expect_equal(ardl_model_df$best_order, c(6,1,5,4,5), ignore_attr = TRUE)


    expect_equal(ardl_model_df$best_model, ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                                                data = data.frame(PSS2001), start = 9,
                                                order=c(6,1,5,4,5)))
})

test_that("missing selection defaults to AIC", {
    expect_equal(auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                           data = PSS2001, start = c(1972, 01),
                           max_order = 6, fixed_order = c(-1,1,-1,-1,-1), starting_order = 5),
                 auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                           data = PSS2001, start = c(1972, 01),
                           max_order = 6, fixed_order = c(-1,1,-1,-1,-1), starting_order = 5,
                           selection = "AIC"))
})

test_that("starting_order is missing", {
    ardl_model_gridF <- auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                                  data = PSS2001, start = c(1972, 01),
                                  max_order = 6, fixed_order = c(-1,1,-1,-1,-1),
                                  grid = FALSE)
    top_orders_gridF <- data.frame(w=c(4,6,4,5,4,5,4,5,5,4,5,4,4,4,3,3,3,3,3,3),
                                   Prod=c(1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1),
                                   UR=c(5,6,5,5,5,5,6,5,6,4,6,4,5,4,4,4,5,4,5,4),
                                   Wedge=c(4,6,5,4,4,5,4,4,4,4,5,4,4,5,1,1,1,2,2,2),
                                   Union=c(5,6,5,5,6,5,5,6,5,5,5,4,4,4,4,5,4,4,4,3))

    #also checks grid = TRUE & selection_minmax = "max"
    ardl_model_gridT <- auto_ardl(w ~ Prod + UR + Wedge | D7475 + D7579,
                                  data = PSS2001, start = c(1972, 01),
                                  max_order = 6, fixed_order = c(-1,1,-1,-1),
                                  selection = "AIC_pss", selection_minmax = "max",
                                  grid = TRUE)
    top_orders_gridT <- data.frame(w=c(4,5,4,5,5,5,4,4,4,4,6,4,4,4,4,6,6,4,6,5),
                                   Prod=c(1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1),
                                   UR=c(2,2,3,0,1,3,1,6,2,0,2,4,6,2,3,0,1,1,3,4),
                                   Wedge=c(6,6,6,6,6,6,6,2,2,6,6,6,1,1,2,6,6,1,6,6))

    expect_equal(ardl_model_gridF$top_orders[,-6], top_orders_gridF)
    expect_equal(ardl_model_gridT$top_orders[,-5], top_orders_gridT, ignore_attr = TRUE)
})

test_that("fixed_order[1] != (-1)", {
    expect_equal(auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                           data = PSS2001, start = c(1972, 01),
                           max_order = 6, fixed_order = c(6,1,-1,-1,-1),
                           starting_order = 5)$top_orders[,-6],
                 data.frame(w=c(6,6,6,6,6),
                            Prod=c(1,1,1,1,1),
                            UR=c(5,5,5,6,6),
                            Wedge=c(4,5,4,4,5),
                            Union=c(5,5,6,5,5)))
})

test_that("avoid search for q<0 when search_type = vertical", {
    ardl_model <- auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                            data = PSS2001, start = c(1972, 01),
                            max_order = 6, fixed_order = c(-1,1,-1,-1,-1),
                            starting_order = c(3,5,1,0,1), search_type = "vertical")

    expect_equal(ardl_model$top_orders[,-6],
                 data.frame(w=c(4,4,6,6,5,5,4,4,6,6,5,5,3,3,3,3,3,4,6,4),
                            Prod=c(1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1),
                            UR=c(2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,2,2,3),
                            Wedge=c(2,2,2,2,2,2,2,1,2,1,2,1,2,2,2,3,1,0,0,0),
                            Union=c(2,3,2,3,2,3,1,1,1,1,1,1,2,3,1,1,1,1,1,1)))
})

test_that("missing max_order causes expected error", {
    expect_error(auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                           data = PSS2001, start = c(1972, 01),
                           fixed_order = c(-1,1,-1,-1,-1), starting_order = 5),
                 "'max_order' is a mandatory argument.")
})

test_that("starting_order[1] < 1 causes expected error", {
    expect_error(auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                           data = PSS2001, start = c(1972, 01),
                           max_order = 6, fixed_order = c(-1,1,-1,-1,-1),
                           starting_order = c(0,5,5,5,5)),
                 "In 'starting_order', the starting order of p \\(first argument\\) can't be less than 1.")
})

test_that("starting_order > max_order causes expected error", {
    expect_error(auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                           data = PSS2001, start = c(1972, 01),
                           max_order = 6, fixed_order = c(-1,1,-1,-1,-1),
                           starting_order = c(5,5,5,7,5)),
                 "'starting_order' can't be greater than 'max_order'.")
})

test_that("fixed_order > max_order causes expected error", {
    expect_error(auto_ardl(w ~ Prod + UR + Wedge + Union | D7475 + D7579,
                           data = PSS2001, start = c(1972, 01),
                           max_order = 6, fixed_order = c(-1,1,-1,7,-1),
                           starting_order = 5),
                 "'fixed_order' can't be greater than 'max_order'.")
})

#' The Danish data on money income prices and interest rates
#'
#' This data set contains the series used by S. Johansen and K. Juselius for
#' estimating a money demand function of Denmark.
#'
#' An object of class "zooreg" "zoo".
#'
#' @format A time-series object with 55 rows and 5 variables. Time period from 1974:Q1
#' until 1987:Q3.
#' \describe{
#'   \item{LRM}{logarithm of real money, M2}
#'   \item{LRY}{logarithm of real income}
#'   \item{LPY}{logarithm of price deflator}
#'   \item{IBO}{bond rate}
#'   \item{IDE}{bank deposit rate}
#' }
#' @source \url{https://onlinelibrary.wiley.com/doi/10.1111/j.1468-0084.1990.mp52002003.x}
#' @references Johansen, S. and Juselius, K. (1990), Maximum Likelihood
#' Estimation and Inference on Cointegration -- with Applications to the Demand
#' for Money, \emph{Oxford Bulletin of Economics and Statistics}, \bold{52, 2},
#' 169--210.
#' @keywords datasets
"denmark"

#' The UK earnings equation data from Pesaran et al. (2001)
#'
#' This data set contains the series used by Pesaran et al. (2001) for estimating the UK earnings equation.
#' The clean format of the data retrieved from the Data Archive of Natsiopoulos and Tzeremes (2022).
#'
#' An object of class "zooreg" "zoo".
#'
#' @format A time-series object with 112 rows and 7 variables. Time period from 1970:Q1
#' until 1997:Q4.
#' \describe{
#'   \item{w}{real wage}
#'   \item{Prod}{labor productivity}
#'   \item{UR}{unemployment rate}
#'   \item{Wedge}{wedge effect}
#'   \item{Union}{union power}
#'   \item{D7475}{income policies 1974:Q1-1975:Q4}
#'   \item{D7579}{income policies 1975:Q1-1979:Q4}
#' }
#' @source \url{http://qed.econ.queensu.ca/jae/datasets/pesaran001/}\cr
#' \url{http://qed.econ.queensu.ca/jae/datasets/natsiopoulos001/}
#' @references M. Hashem Pesaran, Richard J. Smith, and Yongcheol Shin, (2001), "Bounds Testing
#' Approaches to the Analysis of Level Relationships", \emph{Journal of Applied
#' Econometrics}, \bold{16, 3}, 289--326.
#'
#' Kleanthis Natsiopoulos and Nickolaos G. Tzeremes, (2022), "ARDL bounds test for
#' Cointegration: Replicating the Pesaran et al. (2001) Results for the
#' UK Earnings Equation Using R",
#' \emph{Journal of Applied Econometrics}, \bold{37, 5}, 1079--1090.
#' \doi{10.1002/jae.2919}
#' @keywords datasets
"PSS2001"

#' The UK earnings equation data from Natsiopoulos and Tzeremes (2022)
#'
#' This data set contains the series used by Natsiopoulos and Tzeremes (2022) for re-estimating the UK earnings equation.
#' The clean format of the data retrieved from the Data Archive of Natsiopoulos and Tzeremes (2022).
#'
#' An object of class "zooreg" "zoo".
#'
#' @format A time-series object with 196 rows and 9 variables. Time period from 1971:Q1
#' until 2019:Q4.
#' \describe{
#'   \item{time}{time variable}
#'   \item{w}{real wage}
#'   \item{Prod}{labor productivity}
#'   \item{UR}{unemployment rate}
#'   \item{Wedge}{wedge effect}
#'   \item{Union}{union power}
#'   \item{D7475}{income policies 1974:Q1-1975:Q4}
#'   \item{D7579}{income policies 1975:Q1-1979:Q4}
#'   \item{UnionR}{union membership}
#' }
#' @source \url{http://qed.econ.queensu.ca/jae/datasets/natsiopoulos001/}
#' @references Kleanthis Natsiopoulos and Nickolaos G. Tzeremes, (2022), "ARDL bounds test for
#' Cointegration: Replicating the Pesaran et al. (2001) Results for the UK Earnings Equation Using R",
#' \emph{Journal of Applied Econometrics}, \bold{37, 5}, 1079--1090.
#' \doi{10.1002/jae.2919}
#' @keywords datasets
"NT2022"

#' F-test of regression's overall significance
#'
#' \code{f_test_custom} performs an overall significance F-test on a regression.
#' It is used along with \code{\link[stats]{.lm.fit}} to get the F-statistic as
#' this is about 10 times faster than extracting it from a regression using
#' \code{\link[stats]{lm}}.
#'
#' @param dep_var A numeric vector or a matrix with one column representing the
#'   dependent variable.
#' @param indep_vars A matrix representing the independent variables.
#' @param model_res A numeric vector representing the regression's residuals.
#' @param const A logical indicating whether the constant term should be
#'   restricted too.
#'
#' @return \code{f_test_custom} returns a list containing the F-statistic and
#'   the numerator's degrees of freedom.
#' @seealso \code{\link{vcov_custom}} \code{\link{f_bounds_sim}}
#'   \code{\link{t_bounds_sim}}independent
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

f_test_custom <- function(dep_var, indep_vars, model_res, const = TRUE){
    df2 <- nrow(indep_vars) - ncol(indep_vars)
    dep_var_mean <- mean(dep_var)
    mss <- sum(model_res^2)
    if( const == TRUE ) {
        df1 <- ncol(indep_vars) - 1 #number of restricted coefficients
        mss0 <- sum((dep_var - dep_var_mean)^2)
    } else {
        df1 <- ncol(indep_vars) #number of restricted coefficients
        mss0 <- sum((dep_var)^2)
    }
    return(list(f = ((mss0 - mss)/df1) / (mss/df2), df1 = df1))
}

#' Variance-Covariance matrix of a regression
#'
#' \code{vcov_custom} creates the Variance-Covariance matrix of a regression. It
#' is used instead of the \code{\link[stats]{vcov}} because the latter doesn't
#' work with \code{\link[stats]{.lm.fit}}.
#'
#' @param indep_vars A matrix representing the independent variables.
#' @param model_res A numeric vector representing the regression's residuals.
#'
#' @return \code{vcov_custom} returns a Variance-Covariance matrix.
#' @seealso \code{\link{f_test_custom}} \code{\link{f_bounds_sim}}
#'   \code{\link{t_bounds_sim}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

vcov_custom <- function(indep_vars, model_res){
    xx <- t(indep_vars) %*% indep_vars
    return(solve(xx) * sum(model_res ^ 2) / (nrow(indep_vars) - ncol(indep_vars)))
}

#' Critical value bounds stochastic simulation for Wald bounds-test for no
#' cointegration
#'
#' \code{f_bounds_sim} simulates the critical value bounds for the  Wald
#' bounds-test for no cointegration \cite{Pesaran et al. (2001)} expressed both
#' as F-statistics and as Chisq-statistics.
#'
#' @param case An integer from 1-5 specifying whether the 'intercept' and/or the
#'   trend' have to participate in the long-run/cointegrating
#'   relationship/equation (see section 'Cases' in \code{\link{bounds_f_test}}).
#' @param k The number of independent variables.
#' @param alpha A numeric vector between 0 and 1 indicating the significance
#'   level of the critical value bounds. Multiple values can be used.
#' @param T An integer indicating the number of observations.
#' @param R An integer indicating how many iterations will be used. Default is
#'   40000.
#'
#' @return \code{f_bounds_sim} returns a list containing two data frames. One
#'   with the critical value bounds for the F-statistic and one with the
#'   critical value bounds for the Chisq-statistic.
#' @seealso \code{\link{t_bounds_sim}} \code{\link{bounds_f_test}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal ts
#'

f_bounds_sim <- function(case, k, alpha, T, R = 40000) {
	chi2_crit_val0 <- c()
	chi2_crit_val1 <- c()
    f_crit_val0 <- c()
    f_crit_val1 <- c()
	for(i in 1:R) {
        # set the independent covariates
        y1 <- cumsum(stats::rnorm(T))
        dep_var <- as.matrix(diff(y1, 1))
        y1lag1 <- as.matrix(y1[1:(T-1)])

        if (k != 0) {
            x <- matrix(stats::rnorm(T * k), T, k)
            x1 <- apply(x, 2, cumsum)
            xlag1 <- x[1:(T - 1), ]
            x1lag1 <- x1[1:(T - 1), ]
        }

        # set the deterministic linear trend
    	if (case %in% c(1, 2, 3)) {
            # dataset already set
        } else if (case %in% c(4, 5)) {
    	   	if (k == 0) {
    			xlag1 <- c(t = 1:(T - 1))
    			x1lag1 <- c(t = 1:(T - 1))
    		} else {
		 		xlag1 <- cbind(t = c(1:(T - 1)), xlag1)
    			x1lag1 <- cbind(t = c(1:(T - 1)), x1lag1)
    		}
    	}
        xlag1 <- as.matrix(xlag1)
        x1lag1 <- as.matrix(x1lag1)

        if (case == 1) {
    		if (k == 0) {
                indep_vars <- y1lag1
                model0 <- stats::.lm.fit(y = dep_var, x = indep_vars) # I(0) and I(1) are identical
                f_test <- f_test_custom(dep_var, indep_vars, model0$residuals, const = FALSE)
    			f_crit_val0[i] <- f_crit_val1[i] <- f_test$f # F and Chisq are identical for k=0 & no c,t
                chi2_crit_val0[i] <- chi2_crit_val1[i] <- f_test$f / f_test$df1
    		} else {
                indep_vars <- as.matrix(cbind(xlag1, y1lag1))
                model0 <- stats::.lm.fit(y = dep_var, x = indep_vars)
                f_test <- f_test_custom(dep_var, indep_vars, model0$residuals, const = FALSE)
				f_crit_val0[i] <- f_test$f
                chi2_crit_val0[i] <-f_test$f * f_test$df1

                indep_vars <- as.matrix(cbind(x1lag1, y1lag1))
                model1 <- stats::.lm.fit(y = dep_var, x = indep_vars)
                f_test <- f_test_custom(dep_var, indep_vars, model1$residuals, const = FALSE)
				f_crit_val1[i] <- f_test$f
				chi2_crit_val1[i] <- f_test$f * f_test$df1
    		}
    	} else if (case %in% c(2, 5)) {
			if (k == 0) {
    			if (case == 2) {
                    indep_vars <- as.matrix(cbind(1, y1lag1))
                    model0 <- stats::.lm.fit(y = dep_var, x = indep_vars) # I(0) and I(1) are identical
	   				w0 <- w1 <- aod::wald.test(b = stats::coef(model0), Sigma = vcov_custom(indep_vars, model0$residuals), Terms = 1:(k + 2))$result
	   			} else if (case == 5) {
                    indep_vars <- as.matrix(cbind(1, xlag1, y1lag1))
					model0 <- stats::.lm.fit(y = dep_var, x = indep_vars) # I(0) and I(1) are identical
	   				w0 <- w1 <- aod::wald.test(b = stats::coef(model0), Sigma = vcov_custom(indep_vars, model0$residuals), Terms = 3:(k + 3))$result
	   			}
	   		} else {
                indep_vars0 <- as.matrix(cbind(1, xlag1, y1lag1))
    			model0 <- stats::.lm.fit(y = dep_var, x = indep_vars0)
                indep_vars1 <- as.matrix(cbind(1, x1lag1, y1lag1))
    			model1 <- stats::.lm.fit(y = dep_var, x = indep_vars1)
    			if (case == 2) {
	    			w0 <- aod::wald.test(b = stats::coef(model0), Sigma = vcov_custom(indep_vars0, model0$residuals), Terms = 1:(k + 2))$result
    				w1 <- aod::wald.test(b = stats::coef(model1), Sigma = vcov_custom(indep_vars1, model1$residuals), Terms = 1:(k + 2))$result
    			} else if (case == 5) {
	    			w0 <- aod::wald.test(b = stats::coef(model0), Sigma = vcov_custom(indep_vars0, model0$residuals), Terms = 3:(k + 3))$result
    				w1 <- aod::wald.test(b = stats::coef(model1), Sigma = vcov_custom(indep_vars1, model1$residuals), Terms = 3:(k + 3))$result
    			}
    		}
            chi2_crit_val0[i] <- w0[[1]][1]
            chi2_crit_val1[i] <- w1[[1]][1]
            f_crit_val0[i] <- w0[[1]][1] / w0[[1]][2]
            f_crit_val1[i] <- w1[[1]][1] / w1[[1]][2]
		} else if (case %in% c(3, 4)) {
			if (k == 0) {
				if (case == 3) {
                    indep_vars <- as.matrix(cbind(1, y1lag1))
                    model0 <- stats::.lm.fit(y = dep_var, x = indep_vars) # I(0) and I(1) are identical
                } else if (case == 4) {
                    indep_vars <- as.matrix(cbind(1, xlag1, y1lag1))
                    model0 <- stats::.lm.fit(y = dep_var, x = indep_vars) # I(0) and I(1) are identical
                }
    			F0 <- F1 <- f_test_custom(dep_var, indep_vars, model0$residuals, const = TRUE)
			} else {
                indep_vars0 <- as.matrix(cbind(1, xlag1, y1lag1))
    			model0 <- stats::.lm.fit(y = dep_var, x = indep_vars0)
                indep_vars1 <- as.matrix(cbind(1, x1lag1, y1lag1))
    			model1 <- stats::.lm.fit(y = dep_var, x = indep_vars1)
    			F0 <- f_test_custom(dep_var, indep_vars0, model0$residuals, const = TRUE)
    			F1 <- f_test_custom(dep_var, indep_vars1, model1$residuals, const = TRUE)
			}
            chi2_crit_val0[i] <- F0$f * F0$df1
            chi2_crit_val1[i] <- F1$f * F1$df1
            f_crit_val0[i] <- F0$f
            f_crit_val1[i] <- F1$f
		}
	}

	chi2_LB <- stats::quantile(chi2_crit_val0, 1 - alpha)
	chi2_UB <- stats::quantile(chi2_crit_val1, 1 - alpha)
	chi2_bounds <- cbind(chi2_LB, chi2_UB)

    f_LB <- stats::quantile(f_crit_val0, 1 - alpha)
    f_UB <- stats::quantile(f_crit_val1, 1 - alpha)
    f_bounds <- cbind(f_LB, f_UB)

	chi2_mean0 <- mean(chi2_crit_val0)
	chi2_mean1 <- mean(chi2_crit_val1)
	chi2_bounds <- rbind(chi2_bounds, c(chi2_mean0, chi2_mean1))
	chi2_var0 <- stats::var(chi2_crit_val0)
	chi2_var1 <- stats::var(chi2_crit_val1)
	chi2_bounds <- rbind(chi2_bounds, c(chi2_var0, chi2_var1))

    f_mean0 <- mean(f_crit_val0)
    f_mean1 <- mean(f_crit_val1)
    f_bounds <- rbind(f_bounds, c(f_mean0, f_mean1))
    f_var0 <- stats::var(f_crit_val0)
    f_var1 <- stats::var(f_crit_val1)
    f_bounds <- rbind(f_bounds, c(f_var0, f_var1))

	chi2_bounds <- as.data.frame(chi2_bounds)
    f_bounds <- as.data.frame(f_bounds)

	colnames(chi2_bounds) <- c('I0', 'I1')
    colnames(f_bounds) <- c('I0', 'I1')
    rownames(chi2_bounds) <- c(paste0("a", alpha), 'Mean', 'Variance')
    rownames(f_bounds) <- c(paste0("a", alpha), 'Mean', 'Variance')

	return(list(f_bounds = f_bounds, chisq_bounds = chi2_bounds))
}

#' Critical value bounds stochastic simulation for t-bounds test for no
#' cointegration
#'
#' \code{t_bounds_sim} simulates the critical value bounds for the t-bounds test
#' for no cointegration \cite{Pesaran et al. (2001)}.
#'
#' @param case An integer (1, 3 or 5) specifying whether the 'intercept' and/or
#'   the trend' have to participate in the short-run relationship (see section
#'   'Cases' in \code{\link{bounds_t_test}}).
#' @param k The number of independent variables.
#' @param alpha A numeric vector between 0 and 1 indicating the significance
#'   level of the critical value bounds. Multiple values can be used.
#' @param T An integer indicating the number of observations.
#' @param R An integer indicating how many iterations will be used. Default is
#'   40000.
#'
#' @return \code{t_bounds_sim} returns a data frame with the critical value
#'   bounds for the t-statistic.
#' @seealso \code{\link{f_bounds_sim}} \code{\link{bounds_t_test}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal ts
#'

t_bounds_sim <- function(case, k, alpha, T, R = 40000) {
    t_crit_val0 <- c()
    t_crit_val1 <- c()

    if (case %in% c(2,4)) {
        stop("The t-statistic bounds test is not applicable under case II and case IV", call. = FALSE)
    } else if (case %in% c(1)) {
        restriction_place <- k+1
    } else if (case %in% c(3)) {
        restriction_place <- k+2
    } else if (case %in% c(5)) {
        restriction_place <- k+3
    }
    for(i in 1:R) {
        # set the independent covariates
        y1 <- cumsum(stats::rnorm(T))
        dy1 <- diff(y1, 1)
        y1lag1 <- y1[1:(T-1)]

        if (k != 0) {
            x <- matrix(stats::rnorm(T * k), T, k)
            x1 <- apply(x, 2, cumsum)
            xlag1 <- x[1:(T - 1), ]
            x1lag1 <- x1[1:(T - 1), ]
        }

        # set the deterministic linear trend
        if (case %in% c(1,3)) {
            # dataset already set
        } else if (case %in% c(5)) {
            if (k == 0) {
                xlag1 <- c(t = 1:(T - 1))
                x1lag1 <- c(t = 1:(T - 1))
            } else {
                xlag1 <- cbind(t = c(1:(T - 1)), xlag1)
                x1lag1 <- cbind(t = c(1:(T - 1)), x1lag1)
            }
        }

        if (case == 1) {
            if (k == 0) {
                t_crit_val0[i] <- summary(stats::lm(dy1 ~ y1lag1 -1))$coefficients[restriction_place, 3] # I(0) and I(1) are identical
                t_crit_val1[i] <- t_crit_val0[i]
            } else {
                t_crit_val0[i] <- summary(stats::lm(dy1 ~ xlag1 + y1lag1 -1))$coefficients[restriction_place, 3]
                t_crit_val1[i] <- summary(stats::lm(dy1 ~ x1lag1 + y1lag1 -1))$coefficients[restriction_place, 3]
            }
        } else if (case %in% c(5)) {
            if (k == 0) {
                t_crit_val0[i] <- summary(stats::lm(dy1 ~ xlag1 + y1lag1))$coefficients[restriction_place, 3] # I(0) and I(1) are identical
                t_crit_val1[i] <- t_crit_val0[i]
            } else {
                t_crit_val0[i] <- summary(stats::lm(dy1 ~ xlag1 + y1lag1))$coefficients[restriction_place, 3]
                t_crit_val1[i] <- summary(stats::lm(dy1 ~ x1lag1 + y1lag1))$coefficients[restriction_place, 3]
            }
        } else if (case %in% c(3)) {
            if (k == 0) {
                t_crit_val0[i] <- summary(stats::lm(dy1 ~ y1lag1))$coefficients[restriction_place, 3] # I(0) and I(1) are identical
                t_crit_val1[i] <- t_crit_val0[i]
            } else {
                t_crit_val0[i] <- summary(stats::lm(dy1 ~ xlag1 + y1lag1))$coefficients[restriction_place, 3]
                t_crit_val1[i] <- summary(stats::lm(dy1 ~ x1lag1 + y1lag1))$coefficients[restriction_place, 3]
            }
        }
    }

    t_LB <- stats::quantile(t_crit_val0, alpha)
    t_UB <- stats::quantile(t_crit_val1, alpha)
    t_bounds <- cbind(t_LB, t_UB)

    t_mean0 <- mean(t_crit_val0)
    t_mean1 <- mean(t_crit_val1)
    t_bounds <- rbind(t_bounds, c(t_mean0, t_mean1))
    t_var0 <- stats::var(t_crit_val0)
    t_var1 <- stats::var(t_crit_val1)
    t_bounds <- rbind(t_bounds, c(t_var0, t_var1))

    t_bounds <- as.data.frame(t_bounds)

    colnames(t_bounds) <- c('I0', 'I1')
    rownames(t_bounds) <- c(paste0("a", alpha), 'Mean', 'Variance')

    return(t_bounds)
}

#' Multipliers estimation
#'
#' \code{multipliers} is a generic function used to estimate short-run (impact),
#' delay, interim and long-run (total) multipliers, accompanied by their
#' corresponding standard errors, t-statistics and p-values.
#'
#' The function invokes two different \code{\link[utils]{methods}}, one for
#' objects of \code{\link[base]{class}} 'ardl' and one for objects of
#' \code{class} 'uecm'. This is because of the different (but equivalent)
#' transformation functions that are used for each class/model ('ardl' and
#' 'uecm') to estimate the multipliers.
#'
#' \code{type = 0} is equivalent to \code{type = "sr"}.
#'
#' Note that the interim multipliers are the cumulative sum of the delays, and
#' that the sum of the delay multipliers (for long enough periods) and thus
#' a distant enough interim multiplier match the long-run multipliers.
#'
#' The delay (interim) multiplier can be interpreted as the effect on the
#' dependent variable in period t+s, resulting from an instant (sustained) shock
#' to an independent variable in period t.
#'
#' The delta method is used for approximating the standard errors (and thus the
#' t-statistics and p-values) of the estimated long-run and delay multipliers.
#'
#' @param object An object of \code{\link[base]{class}} 'ardl' or 'uecm'.
#' @param type A character string describing the type of multipliers. Use "lr"
#' for long-run (total) multipliers (default), "sr" or 0 for short-run (impact)
#' multipliers or an integer between 1 and 200 for delay and interim multipliers.
#' @param vcov_matrix The estimated covariance matrix of the parameter estimates.
#'   Used by the transformation function to estimate the standard errors (and
#'   so the t-statistics and p-values) of the multipliers. The default is
#'   \code{vcov(object)} (when \code{vcov_matrix = NULL}), but other estimations
#'   of the covariance matrix of the regression's estimated coefficients can
#'   also be used (e.g., using \code{\link[sandwich]{vcovHC}} or
#'   \code{\link[sandwich]{vcovHAC}}).
#' @param se A logical indicating whether you want standard errors for delay
#' multipliers to be provided. The default is FALSE. Note that this parameter
#' does not refer to the standard errors for the long-run and short-run
#' multipliers, for which are always calculated. IMPORTANT: Calculating standard
#' errors for long periods of delays may cause your computer to run out of
#' memory and terminate your R session, losing important unsaved work. As a rule
#' of thumb, try not to exceed \code{type = 19} when \code{se = TRUE}.
#'
#' @return \code{multipliers} returns (for long and short run multipliers) a
#'   data.frame containing the independent variables (including possibly
#'   existing intercept or trend and excluding the fixed variables) and their
#'   corresponding standard errors, t-statistics and p-values. For delay and
#'   interim multipliers it returns a list with a data.frame for each variable,
#'   containing the delay and interim multipliers for each period.
#'
#' @section Mathematical Formula:
#'
#' \strong{Short-Run Multipliers:}
#' \describe{
#'   \item{As derived from an ARDL:}{}
#' }
#' \deqn{\frac{\partial y_{t}}{\partial x_{j,t}} = b_{j,0} \;\;\;\;\; j \in \{1,\dots,k\}}
#'
#' \describe{
#'   \item{As derived from an Unrestricted ECM:}{}
#' }
#' \deqn{\frac{\partial y_{t}}{\partial x_{j,t}} = \omega_{j} \;\;\;\;\; j \in \{1,\dots,k\}}
#'
#' \describe{
#'   \item{Constant and Linear Trend:}{}
#' }
#' \deqn{c_{0}}
#' \deqn{c_{1}}
#'
#' \strong{Delay & Interim Multipliers:}
#' \describe{
#'   \item{As derived from an ARDL:}{}
#' }
#' \deqn{Delay_{x_{j},s} = \frac{\partial y_{t+s}}{\partial x_{j,t}} = b_{j,s} + \sum_{i=1}^{min\{p,s\}} b_{y,i} \frac{\partial y_{t+(s-i)}}{\partial x_{j,t}} \;\;\;\;\; b_{j,s} = 0 \;\; \forall \;\; s > q}
#' \deqn{Interim_{x_{j},s} = \sum_{i=0}^{s} Delay_{x_{j},s}}
#'
#' \describe{
#'   \item{Constant and Linear Trend:}{}
#' }
#' \deqn{Delay_{intercept,s} = c_{0} + \sum_{i=1}^{min\{p,s\}} b_{y,i} Delay_{intercept,s-i} \;\;\;\;\; c_{0} = 0 \;\; \forall \;\; s \neq 0}
#' \deqn{Interim_{intercept,s} = \sum_{i=0}^{s} Delay_{intercept,s}}
#' \deqn{Delay_{trend,s} = c_{1} + \sum_{i=1}^{min\{p,s\}} b_{y,i} Delay_{trend,s-i} \;\;\;\;\; c_{1} = 0 \;\; \forall \;\; s \neq 0}
#' \deqn{Interim_{trend,s} = \sum_{i=0}^{s} Delay_{trend,s}}
#'
#' \strong{Long-Run Multipliers:}
#' \describe{
#'   \item{As derived from an ARDL:}{}
#' }
#' \deqn{\frac{\partial y_{t+\infty}}{\partial x_{j,t}} = \theta_{j} = \frac{\sum_{l=0}^{q_{j}}b_{j,l}}{1-\sum_{i=1}^{p}b_{y,i}} \;\;\;\;\; j \in \{1,\dots,k\}}
#' \describe{
#'   \item{Constant and Linear Trend:}{}
#' }
#' \deqn{\mu = \frac{c_{0}}{1-\sum_{i=1}^{p}b_{y,i}}}
#' \deqn{\delta = \frac{c_{1}}{1-\sum_{i=1}^{p}b_{y,i}}}
#'
#' \describe{
#'   \item{As derived from an Unrestricted ECM:}{}
#' }
#' \deqn{\frac{\partial y_{t+\infty}}{\partial x_{j,t}} = \theta_{j} = \frac{\pi_{j}}{-\pi_{y}} \;\;\;\;\; j \in \{1,\dots,k\}}
#' \describe{
#'   \item{Constant and Linear Trend:}{}
#' }
#' \deqn{\mu = \frac{c_{0}}{-\pi_{y}}}
#' \deqn{\delta = \frac{c_{1}}{-\pi_{y}}}
#'
#' @seealso \code{\link{ardl}}, \code{\link{uecm}}, \code{\link{plot_delay}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords math
#' @export
#' @examples
#' data(denmark)
#'
#' ## Estimate the long-run multipliers of an ARDL(3,1,3,2) model ---------
#'
#' # From an ARDL model
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' mult_ardl <- multipliers(ardl_3132)
#' mult_ardl
#'
#' # From an UECM
#' uecm_3132 <- uecm(ardl_3132)
#' mult_uecm <- multipliers(uecm_3132)
#' mult_uecm
#'
#' all.equal(mult_ardl, mult_uecm)
#'
#'
#' ## Estimate the short-run multipliers of an ARDL(3,1,3,2) model --------
#'
#' mult_sr <- multipliers(uecm_3132, type = "sr")
#' mult_0 <- multipliers(uecm_3132, type = 0)
#' all.equal(mult_sr, mult_0)
#'
#'
#' ## Estimate the delay & interim multipliers of an ARDL(3,1,3,2) model --
#'
#' mult_lr <- multipliers(uecm_3132, type = "lr")
#' mult_inter80 <- multipliers(uecm_3132, type = 80)
#'
#' mult_lr
#' sum(mult_inter80$`(Intercept)`$Delay)
#' mult_inter80$`(Intercept)`$Interim[nrow(mult_inter80$`(Intercept)`)]
#' sum(mult_inter80$LRY$Delay)
#' mult_inter80$LRY$Interim[nrow(mult_inter80$LRY)]
#' sum(mult_inter80$IBO$Delay)
#' mult_inter80$IBO$Interim[nrow(mult_inter80$IBO)]
#' sum(mult_inter80$IDE$Delay)
#' mult_inter80$IDE$Interim[nrow(mult_inter80$IDE)]
#' plot(mult_inter80$LRY$Delay, type='l')
#' plot(mult_inter80$LRY$Interim, type='l')
#'
#' mult_inter12 <- multipliers(uecm_3132, type = 12, se = TRUE)
#' plot_delay(mult_inter12, interval = 0.95)

multipliers <- function(object, type = "lr", vcov_matrix = NULL, se = FALSE) {
    UseMethod("multipliers")
}

#' @rdname multipliers
#' @export
#'

multipliers.ardl <- function(object, type = "lr", vcov_matrix = NULL, se = FALSE) {

    # no visible binding for global variable NOTE solution
    group_id <- coeff <- sums <- NULL; rm(group_id, coeff, sums)

    if (!(type %in% c("lr", "sr", 0:200))) {
        stop("'type' should be one of 'lr', 'sr' or a number between 0 and 200", call. = FALSE)
    }

    if (is.null(vcov_matrix)) vcov_matrix <- stats::vcov(object)
    kw <- object$parsed_formula$kw
    kx <- object$parsed_formula$kx
    kfixed <- object$parsed_formula$kfixed
    objcoef <- object$coefficients
    orders_x <- object$order[-1]
    from <- kw + 1
    to <- kw + object$order[1]
    if (type == "lr") {
        # create table without the y in levels and fixed
        x_table <- dplyr::tibble(name = names(objcoef), coeff = objcoef) %>%
            dplyr::slice(-(from:to)) %>%
            dplyr::slice(1:(dplyr::n() - object$parsed_formula$kfixed))
        # create table only with y in levels
        y_table <- dplyr::tibble(objcoef) %>% dplyr::slice(from:to)
        # create groups to sum by group
        x_table <- x_table %>%
            dplyr::mutate(group_id = if (kw != 0) {
                c(1:kw, rep(from:(from + object$parsed_formula$kx - 1), orders_x + 1))
            } else {
                c(rep(from:(from + object$parsed_formula$kx - 1), orders_x + 1))
            })
        # create the sums of levels of x and trends
        temp <- x_table %>%
            dplyr::group_by(group_id) %>%
            dplyr::summarise(sums = sum(coeff)) %>%
            dplyr::select(sums)

        # calculate coefficients of multipliers
        multipliers_coef <- (temp / (1 - sum(y_table)))[ ,1]

        if (kw != 0) {
            names(multipliers_coef) <- c(names(objcoef)[1:kw],
                                         object$parsed_formula$x_part$var)
        } else {
            names(multipliers_coef) <- object$parsed_formula$x_part$var
        }

        multipliers_se <- delta_method(object, vcov_matrix = vcov_matrix)
        multipliers <- data.frame(multipliers_coef, multipliers_se, multipliers_coef/multipliers_se,
                                  2 * stats::pt(-abs(multipliers_coef/multipliers_se), df = stats::df.residual(object))) # df = n - # of estimated coefficients
        multipliers <- data.frame(rownames(multipliers), multipliers)
        names(multipliers) <- c("Term", "Estimate", "Std. Error", "t value", "Pr(>|t|)")
        rownames(multipliers) <- 1:nrow(multipliers)

        return(multipliers)
    } else {
        b0 <- 1
        for (i in 1:(length(orders_x) -1)) {
            b0 <- c(b0, b0[length(b0)] + orders_x[i] +1)
        }
        b0 <- if (kw!=0) c(1:kw, b0+kw) else b0
        # create table without the y in levels and fixed
        sr_mult <- as.data.frame(summary(object)$coefficients) %>%
            dplyr::slice(-(from:to)) %>% dplyr::slice(b0)
        sr_mult <- cbind(Term = rownames(sr_mult), sr_mult)
        rownames(sr_mult) <- 1:nrow(sr_mult)
        if (type %in% 1:200) { # anything except 0:200 would have been stopped in the earlier check
            interim = type
            delays_table <- as.data.frame(summary(object)$coefficients) %>%
                dplyr::slice(-(from:to))
            if (kfixed != 0) {
                delays_table <- delays_table %>% dplyr::slice(-((nrow(delays_table)-kfixed+1):nrow(delays_table)))
            }
            delay <- list()
            int_mult <- list()
            if (se) xpressions <- list()
            orders_wx <- c(rep(0, kw), orders_x)
            if (kw != 0) {
                delay_names <- c(rownames(delays_table)[1:kw], object$parsed_formula$x_part$var)
            } else {
                delay_names <- object$parsed_formula$x_part$var
            }
            y_table <- dplyr::tibble(objcoef) %>% dplyr::slice(from:to)
            for (k in 1:(kx+kw)) {
                delay[[k]] <- delays_table$Estimate[1:(orders_wx[k]+1)]
                names(delay)[k] <- delay_names[k]
                int_mult[[k]] <- data.frame()
                names(int_mult)[[k]] <- delay_names[k]
                if (se) {
                    xpressions[[k]] <- data.frame()
                    names(xpressions)[[k]] <- delay_names[k]
                }
                for (ss in 0:interim) {
                    weights_n <- min(object$order[1], ss)
                    direct <- ifelse(ss <= orders_wx[k], delay[[k]][ss+1], 0)
                    if (se) {
                        skip_w_y <- ifelse((kw != 0) & k %in% 1:kw, 0 + k-1, kw + object$order[1])
                        if (direct == 0) {
                            direct_xpression <- NULL
                        } else {
                            direct_xpression <- paste0("x", skip_w_y + ifelse(!k %in% 1:kw, sum(orders_wx[kw:(k-1)]) + k-kw-1, 0) + (ss + 1))
                        }
                    }
                    if (ss == 0) {
                        int_mult[[k]] <- data.frame(Period = ss, Delay = direct)
                        if (se) xpressions[[k]] <- data.frame(Period = ss, xpression = direct_xpression)
                    } else {
                        int_mult[[k]] <- rbind(int_mult[[k]],
                                               data.frame(Period = ss,
                                                          Delay = direct +
                                                              sum(y_table[1:weights_n,] *
                                                                      rev(int_mult[[k]][(ss-(weights_n-1)):ss,"Delay"]))))
                        if (se) {
                            xpressions[[k]] <- rbind(xpressions[[k]],
                                                     data.frame(Period = ss,
                                                                xpression = paste0(c(direct_xpression,
                                                                                     paste0("x", (1:weights_n)+kw, "*(",
                                                                                            rev(xpressions[[k]][(ss-(weights_n-1)):ss,"xpression"]),
                                                                                            ")")), collapse = "+"
                                                                )))
                        }
                    }
                }
                delays_table <- delays_table %>% dplyr::slice(-(1:(orders_wx[k] + 1)))
            }
            if (se) xpressions <- lapply(xpressions, FUN = function(x) {data.frame(Period = x$Period, xpression = paste0("~ ", x$xpression))})
            if (se) xpressions <- lapply(xpressions, FUN = function(x) {msm::deltamethod(lapply(x$xpression, stats::formula), stats::coef(object), vcov_matrix)})
            for (i in 1:length(int_mult)) {
                if (se) int_mult[[i]] <- cbind(int_mult[[i]], "Std. Error Delay" = xpressions[[i]])
                rownames(int_mult[[i]]) <- NULL
            }
            int_mult <- lapply(int_mult, FUN = function(x) {cbind(x, Interim = cumsum(x$Delay))})
            return(int_mult)
        } else {
            return(sr_mult)
        }
    }
}

#' @rdname multipliers
#' @export
#'

multipliers.uecm <- function(object, type = "lr", vcov_matrix = NULL, se = FALSE) {

    if (!(type %in% c("lr", "sr", 0:200))) {
        stop("'type' should be one of 'lr', 'sr' or a number between 0 and 200", call. = FALSE)
    }

    if (is.null(vcov_matrix)) vcov_matrix <- stats::vcov(object)

    if (type %in% 1:200) {
        return(multipliers(object = ardl(object), type = type, vcov_matrix = vcov_matrix, se = se))
    }

    kw <- object$parsed_formula$kw
    objcoef <- object$coefficients
    orders_x <- object$order[-1]
    objnames_ws <- gsub(" ", "", names(objcoef), fixed = TRUE)
    objxvars <- object$parsed_formula$x_part$var
    if (type == "lr") {
        if (kw != 0) {
            multipliers_coef <- c(
                -objcoef[1:kw] / objcoef[kw + 1],
                -objcoef[(kw + 2):(kw + object$parsed_formula$kx + 1)] / objcoef[kw + 1]
            )
        } else {
            multipliers_coef <- -objcoef[2:(object$parsed_formula$kx + 1)] / objcoef[1]
        }

        multipliers_coef_se <- delta_method(object, vcov_matrix = vcov_matrix)
        multipliers <- data.frame(multipliers_coef, multipliers_coef_se, multipliers_coef/multipliers_coef_se,
                                  2 * stats::pt(-abs(multipliers_coef/multipliers_coef_se), stats::df.residual(object))) # df = n - # of estimated coefficients
    } else {
        # use gsub because sometimes the spacing goes weird
        Xt_1 <- c()
        Xt_1[orders_x != 0] <- gsub(" ", "", paste0("d(", objxvars, ")"),
            fixed = TRUE)[orders_x != 0]
        Xt_1[orders_x == 0] <- gsub(" ", "", objxvars[orders_x == 0], fixed = TRUE)
        if (kw != 0) {
            srm <- c(1:kw, which(objnames_ws %in% Xt_1))
            multipliers <- summary(object)$coefficients[srm,]
        } else {
            srm <- objnames_ws %in% Xt_1
            multipliers <- summary(object)$coefficients[srm,]
        }
    }

    pure_names <- gsub(".*\\(([^),]+).*", "\\1", rownames(multipliers)[(kw+1):nrow(multipliers)])
    row_indices <- match(objxvars, pure_names)
    if (kw != 0) {
        multipliers <- data.frame(c(names(objcoef)[1:kw], objxvars),
                                  rbind(multipliers[1:kw, ], multipliers[row_indices+kw, ]))
    } else {
        multipliers <- data.frame(objxvars,  multipliers[row_indices, ])
    }
    names(multipliers) <- c("Term", "Estimate", "Std. Error", "t value", "Pr(>|t|)")
    rownames(multipliers) <- 1:nrow(multipliers)

    return(multipliers)
}

#' Create plot for the long-run (cointegrating) equation
#'
#' Creates a plot for the long-run relationship in comparison with the dependent
#' variable, and the fitted values of the model. This is a basic
#' \code{\link[ggplot2]{ggplot}} with a few customizable parameters.
#'
#' @param object An object of \code{\link[base]{class}} `ardl`.
#' @param coint_eq The objected returned from \code{\link{coint_eq}}.
#' @param facets A logical indicating whether the long-run relationship appears
#' in a separate plot. Default is FALSE.
#' @param show_fitted A logical indicating whether the fitted values are shown.
#' Default is FALSE.
#' @param show.legend A logical indicating whether the legend is shown.
#' Default is FALSE.
#' @param xlab Name displayed at the x axis. Default is "Time".
#' @param ... Currently unused argument.
#'
#' @return \code{plot_lr} returns a \code{\link[ggplot2]{ggplot}} object.
#'
#' @seealso \code{\link{coint_eq}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords iplots
#' @export
#'
#' @examples
#' ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
#' ce2 <- coint_eq(ardl_3132, case = 2)
#'
#' plot_lr(ardl_3132, coint_eq = ce2)
#'
#' ## Compare fitted values and place long-run relationship separately ----
#'
#' ce3 <- coint_eq(ardl_3132, case = 3)
#' plot_lr(ardl_3132, coint_eq = ce3, facets = TRUE, show_fitted = TRUE,
#'         show.legend = TRUE)
#'

plot_lr <- function(object, coint_eq, facets = FALSE, show_fitted = FALSE,
                    show.legend = FALSE, xlab = "Time", ...) {
    y_var <- object$parsed_formula$y_part$var
    plot_data <- zoo::cbind.zoo(object$data[,y_var], coint_eq, object$fitted.values)
    lr_name <- "long-run"
    fit_name <- "fitted"
    names(plot_data) <- c(y_var, lr_name, fit_name)

    plot_function <- function(object, coint_eq, facets, show_fitted, show.legend, xlab, .data) {
        facet1 <- ggplot2::ggplot(plot_data, ggplot2::aes(x = zoo::index(plot_data))) +
            ggplot2::geom_line(ggplot2::aes(y = .data[[y_var]],
                                            color = y_var), show.legend = show.legend) +
            ggplot2::labs(x = xlab, y = y_var, color = "") +
            ggplot2::theme_minimal() +
            ggplot2::theme(legend.position = "bottom")

        if (!facets) {
            facet1 <- facet1 +
                ggplot2::geom_line(ggplot2::aes(y = .data[[lr_name]], color = lr_name), show.legend = show.legend)
            if (show_fitted) {
                facet1 <- facet1 +
                    ggplot2::geom_line(ggplot2::aes(y = .data[[fit_name]], color = fit_name),
                              na.rm = TRUE, show.legend = show.legend) +
                    ggplot2::scale_color_manual(guide = "legend",
                                       values = c("black", "red", "blue"),
                                       breaks = c(y_var, lr_name, fit_name),
                                       labels = c(y_var, lr_name, fit_name))
            } else {
                facet1 <- facet1 +
                    ggplot2::scale_color_manual(guide = "legend",
                                       values = c("black", "red"),
                                       breaks = c(y_var, lr_name),
                                       labels = c(y_var, lr_name))
            }
        } else {
            if (show_fitted) {
                facet1 <- facet1 +
                    ggplot2::geom_line(ggplot2::aes(y = .data[[fit_name]], color = fit_name),
                              na.rm = TRUE, show.legend = show.legend) +
                    ggplot2::scale_color_manual(guide = "legend",
                                       labels = c(y_var, fit_name),
                                       breaks = c(y_var, fit_name),
                                       values = c("black", "blue"))
            } else {
                facet1 <- facet1 +
                    ggplot2::scale_color_manual(guide = "legend",
                                       values = "black",
                                       breaks = y_var,
                                       labels = y_var)
            }
        }

        if (facets) {
            facet2 <- ggplot2::ggplot(plot_data, ggplot2::aes(x = zoo::index(plot_data))) +
                ggplot2::geom_line(ggplot2::aes(y = .data[[lr_name]], color = lr_name), show.legend = show.legend) +
                ggplot2::labs(x = xlab, y = y_var, color = "") +
                ggplot2::scale_color_manual(values = "red", labels = lr_name) +
                ggplot2::theme_minimal() +
                ggplot2::theme(legend.position = "bottom")

            gridExtra::grid.arrange(facet1, facet2, ncol = 1)
        } else {
            facet1
        }
    }

    plot_function(object = object, coint_eq = coint_eq, facets = facets,
                  show_fitted = show_fitted, show.legend = show.legend, xlab = xlab)
}

#' Delta method
#'
#' An internal generic function, customized for approximating the standard
#' errors of the estimated multipliers.
#'
#' The function invokes two different \code{\link[utils]{methods}}, one for
#' objects of \code{\link[base]{class}} 'ardl' and one for objects of
#' \code{class} 'uecm'. This is because of the different (but equivalent)
#' transformation functions that are used for each class/model ('ardl' and
#' 'uecm') to estimate the multipliers.
#'
#' @inheritParams multipliers
#'
#' @return \code{delta_method} returns a numeric vector of the same length as
#'   the number of the independent variables (excluding the fixed ones) in the
#'   model.
#'
#' @seealso \code{\link{multipliers}}, \code{\link{ardl}}, \code{\link{uecm}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords internal
#'

delta_method <- function(object, vcov_matrix = NULL) {
    UseMethod("delta_method")
}

#' @rdname delta_method
#' @export

delta_method.ardl <- function(object, vcov_matrix = NULL) {

    if (is.null(vcov_matrix)) vcov_matrix <- stats::vcov(object)
    estmean <- stats::coef(object)
    estvar <- vcov_matrix
    w_part <- object$parsed_formula$w_part$var
    kx <- object$parsed_formula$kx

    p <- object$order[1]
    q <- object$order[-1]
    m <- 1 : (sum(object$order) + length(object$order) - 1)
    # case 2, 3
    # keep first this case to avoid w_part[1] in case when length(w_part)==0
    if ((length(w_part) == 0) | (length(w_part) == 2) ) { # 2nd statement for the case with only trend
        place_y <- 2:(p+1)
        m <- c(m, m[length(m)] + 1)
        bins_w <- 1
    # case 1
    } else if ((w_part[1] == "- 1") & length(w_part) == 1) {
        place_y <- 1:p
        bins_w <- c()
    # case 4, 5
    } else if (length(w_part) == 1) {
        place_y <- 3:(p+2)
        m <- 1:(m[length(m)] + 2)
        bins_w <- c(1, 1)
    }

    place_x <- m[!(m %in% place_y)]
    bin_breaks <- c(0, cumsum(c(bins_w, q + 1)))
    pi_yx.x <- lapply(1:(length(bin_breaks) - 1), function(i) {
        paste0("~ (",
            paste0("x", place_x[(bin_breaks[i] + 1) : (bin_breaks[i + 1])], collapse = " + "),
            ")")
        })
    pi_yy <- paste0(" / (1 - (",
        paste0("x", place_y, collapse = " + "),
        "))")
    restrictions <- paste0(pi_yx.x, pi_yy)

    lr_se <- sapply(1:length(restrictions), function(i) {
        msm::deltamethod(stats::formula(restrictions[i]), estmean, estvar)
        })
    return(lr_se)
}

#' @rdname delta_method
#' @export

delta_method.uecm <- function(object, vcov_matrix = NULL) {
    if (is.null(vcov_matrix)) vcov_matrix <- stats::vcov(object)
    estmean <- stats::coef(object)
    estvar <- vcov_matrix
    w_part <- object$parsed_formula$w_part$var
    kx <- object$parsed_formula$kx

    # case 2, 3
    # keep first this case to avoid w_part[1] in case when length(w_part)==0
    if ((length(w_part) == 0) | (length(w_part) == 2) ) { # 2nd statement for the case with only trend
        restrictions <- paste0("~ x", c(1, 3:(kx + 2)), " / x2")
    # case 1
    } else if ((w_part[1] == "- 1") & length(w_part) == 1) {
        restrictions <- paste0("~ x", c(2:(kx + 1)), " / x1")
    # case 4, 5
    } else if (length(w_part) == 1) {
        restrictions <- paste0("~ x", c(1, 2, 4:(kx + 3)), " / x3")
    }

    lr_se <- sapply(1:length(restrictions), function(i) {
        msm::deltamethod(stats::formula(restrictions[i]), estmean, estvar)
        })
    return(lr_se)
}

#' Automatic ARDL model selection
#'
#' It searches for the best ARDL order specification, according to the selected
#' criterion, taking into account the constraints provided.
#'
#' @param formula A "formula" describing the linear model. Details for model
#'   specification are given under 'Details' in the help file of the
#'   \code{\link{ardl}} function.
#' @param max_order It sets the maximum order for each variable where the search
#'   is taking place. A numeric vector of the same length as the total number of
#'   variables (excluding the fixed ones, see 'Details' in the help file of the
#'   \code{\link{ardl}} function). It should only contain positive integers. An
#'   integer could be provided if the maximum order for all variables is the
#'   same.
#' @param fixed_order It allows setting a fixed order for some variables. The
#'   algorithm will not search for any other order than this. A numeric vector
#'   of the same length as the total number of variables (excluding the fixed
#'   ones). It should contain positive integers or 0 to set as a constraint. A
#'   -1 should be provided for any variable that should not be constrained.
#'   \code{fixed_order} overrides the corresponding \code{max_order} and
#'   \code{starting_order}.
#' @param starting_order Specifies the order for each variable from which each
#'   search will start. It is a numeric vector of the same length as the total
#'   number of variables (excluding the fixed ones). It should contain positive
#'   integers or 0 or only one integer could be provided if the starting order
#'   for all variables is the same. Default is set to NULL. If unspecified
#'   (\code{NULL}) and \code{grid = FALSE}, then all possible \eqn{ARDL(p)}
#'   models are calculated (constraints are taken into account), where \eqn{p}
#'   is the minimum value in \code{max_order}. Note that where
#'   \code{starting_order} is provided, its first element will be the minimum
#' value of \eqn{p} that the searching algorithm will consider (think of it like
#' a 'minimum p order' restriction) (see 'Searching algorithm' below). If
#' \code{grid = TRUE}, only the first argument (\eqn{p}) will have an effect.
#' @param selection A character string specifying the selection criterion
#'   according to which the candidate models will be ranked. Default is
#'   \code{\link[stats]{AIC}}. Any other selection criterion can be used (a user
#'   specified or a function from another package) as long as it can be applied
#'   as \code{selection(model)}. The preferred model is the one with the smaller
#'   value of the selection criterion. If the selection criterion works the
#'   other way around (the bigger the better), \code{selection_minmax = "max"}
#'   should also be supplied (see 'Examples' below).
#' @param selection_minmax A character string that indicates whether the
#'   criterion in \code{selection} is supposed to be minimized (default) or
#'   maximized.
#' @param grid If \code{FALSE} (default), the stepwise searching regression
#'   algorithm will search for the best model by adding and subtracting terms
#'   corresponding to different ARDL orders. If \code{TRUE}, the whole set of
#'   all possible ARDL models (accounting for constraints) will be evaluated.
#'   Note that this method can be very time-consuming in case that
#'   \code{max_order} is big and there are many independent variables that
#'   create a very big number of possible combinations.
#' @param search_type A character string describing the search type. If
#'   "horizontal" (default), the searching algorithm increases or decreases by 1
#'   the order of each variable in each iteration. When the order of the last
#'   variable has been accessed, it begins again from the first variable until
#'   it converges. If "vertical", the searching algorithm increases or decreases
#'   by 1 the order of a variable until it converges. Then it continues the same
#'   for the next variable. The two options result to very similar top orders.
#'   The default ("horizontal"), sometimes is a little more accurate, but the
#'   "vertical" is almost 2 times faster. Not applicable if \code{grid = TRUE}.
#' @inheritParams ardl
#'
#' @return \code{auto_ardl} returns a list which contains:
#'   \item{\code{best_model}}{An object of class \code{c("ardl", "dynlm", "lm")}}
#'   \item{\code{best_order}}{A numeric vector with the order of the best model selected}
#'   \item{\code{top_orders}}{A data.frame with the orders of the top 20 models}
#'
#' @section Searching algorithm: The algorithm performs the optimization process
#'   starting from multiple starting points concerning the autoregressive order
#'   \eqn{p}. The searching algorithm will perform a complete search, each time
#'   starting from a different starting order. These orders are presented in the
#'   tables below, for \code{grid = FALSE} and different values of
#'   \code{starting_order}.
#'
#'   \code{starting_order = NULL}:
#'   \tabular{ccccccc}{
#'   ARDL(p) \tab -> \tab p \tab q1 \tab q2 \tab ... \tab qk\cr
#'   ARDL(1) \tab -> \tab 1 \tab 1 \tab 1 \tab ... \tab 1\cr
#'   ARDL(2) \tab -> \tab 2 \tab 2 \tab 2 \tab ... \tab 2\cr
#'   : \tab -> \tab : \tab : \tab : \tab : \tab :\cr
#'   ARDL(P) \tab -> \tab P \tab P \tab P \tab ... \tab P
#'   }
#'
#'   \code{starting_order = c(3, 0, 1, 2)}:
#'   \tabular{cccc}{
#'   p \tab q1 \tab q2 \tab q3\cr
#'   3 \tab 0 \tab 1 \tab 2\cr
#'   4 \tab 0 \tab 1 \tab 2\cr
#'   : \tab : \tab : \tab :\cr
#'   P \tab 0 \tab 1 \tab 2
#'   }
#'
#' @seealso \code{\link{ardl}}
#' @author Kleanthis Natsiopoulos, \email{klnatsio@@gmail.com}
#' @keywords optimize models ts
#' @export
#' @examples
#' data(denmark)
#'
#' ## Find the best ARDL order --------------------------------------------
#'
#' # Up to 5 for the autoregressive order (p) and 4 for the rest (q1, q2, q3)
#'
#' # Using the defaults search_type = "horizontal", grid = FALSE and selection = "AIC"
#' # ("Not run" indications only for testing purposes)
#' \dontrun{
#' model1 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                     max_order = c(5,4,4,4))
#' model1$top_orders
#'
#' ## Same, with search_type = "vertical" -------------------------------
#'
#' model1_h <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                       max_order = c(5,4,4,4), search_type = "vertical")
#' model1_h$top_orders
#'
#' ## Find the global optimum ARDL order ----------------------------------
#'
#' # It may take more than 10 seconds
#' model_grid <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                         max_order = c(5,4,4,4), grid = TRUE)
#'
#' ## Different selection criteria ----------------------------------------
#'
#' # Using BIC as selection criterion instead of AIC
#' model1_b <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                       max_order = c(5,4,4,4), selection = "BIC")
#' model1_b$top_orders
#'
#' # Using other criteria like adjusted R squared (the bigger the better)
#' adjr2 <- function(x) { summary(x)$adj.r.squared }
#' model1_adjr2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                            max_order = c(5,4,4,4), selection = "adjr2",
#'                            selection_minmax = "max")
#' model1_adjr2$top_orders
#'
#' # Using functions from other packages as selection criteria
#' if (requireNamespace("AICcmodavg", quietly = TRUE)) {
#'
#' library(AICcmodavg)
#' model1_aicc <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                           max_order = c(5,4,4,4), selection = "AICc")
#' model1_aicc$top_orders
#' adjr2 <- function(x){ summary(x)$adj.r.squared }
#' model1_adjr2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                            max_order = c(5,4,4,4), selection = "adjr2",
#'                            selection_minmax = "max")
#' model1_adjr2$top_orders
#'
#' ## DIfferent starting order --------------------------------------------
#'
#' # The searching algorithm will start from the following starting orders:
#' # p q1 q2 q3
#' # 1 1  3  2
#' # 2 1  3  2
#' # 3 1  3  2
#' # 4 1  3  2
#' # 5 1  3  2
#'
#' model1_so <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                         max_order = c(5,4,4,4), starting_order = c(1,1,3,2))
#'
#' # Starting from p=3 (don't search for p=1 and p=2)
#' # Starting orders:
#' # p q1 q2 q3
#' # 3 1  3  2
#' # 4 1  3  2
#' # 5 1  3  2
#'
#' model1_so_3 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                         max_order = c(5,4,4,4), starting_order = c(3,1,3,2))
#'
#' # If starting_order = NULL, the starting orders for each iteration will be:
#' # p q1 q2 q3
#' # 1 1  1  1
#' # 2 2  2  2
#' # 3 3  3  3
#' # 4 4  4  4
#' # 5 5  5  5
#' }
#'
#' ## Add constraints -----------------------------------------------------
#'
#' # Restrict only the order of IBO to be 2
#' model1_ibo2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                         max_order = c(5,4,4,4), fixed_order = c(-1,-1,2,-1))
#' model1_ibo2$top_orders
#'
#' # Restrict the order of LRM to be 3 and the order of IBO to be 2
#' model1_lrm3_ibo2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                         max_order = c(5,4,4,4), fixed_order = c(3,-1,2,-1))
#' model1_lrm3_ibo2$top_orders
#'
#' ## Set the starting date for the regression (data starts at "1974 Q1") -
#'
#' # Set regression starting date to "1976 Q1"
#' model1_76q1 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
#'                         max_order = c(5,4,4,4), start = "1976 Q1")
#' start(model1_76q1$best_model)
#' }

auto_ardl <- function(formula, data, max_order, fixed_order = -1, starting_order = NULL,
                      selection = "AIC", selection_minmax = c("min", "max"),
                      grid = FALSE, search_type = c("horizontal", "vertical"),
                      start = NULL, end = NULL, ...) {

    if (!any(c("ts", "zoo", "zooreg") %in% class(data))) {
        data <- stats::ts(data, start = 1, end = nrow(data), frequency = 1)
    }
    if (missing(max_order) == TRUE) {stop("'max_order' is a mandatory argument.", call. = FALSE)}
    if (missing(selection) == TRUE) {selection <- "AIC"}
    search_type <- match.arg(search_type)
    selection_minmax <- match.arg(selection_minmax)
    parsed_formula <- parse_formula(formula = formula, colnames_data = colnames(data))
    max_order <- parse_order(orders = max_order, order_name = "max_order",
                             var_names = parsed_formula$z_part$var, kz = parsed_formula$kz)
    fixed_order <- parse_order(orders = fixed_order, order_name = "fixed_order",
                               var_names = parsed_formula$z_part$var, kz = parsed_formula$kz, restriction = -1)
    if (!missing(starting_order)) {
        starting_order_null <- FALSE
        if (starting_order[1] < 1) { stop("In 'starting_order', the starting order of p (first argument) can't be less than 1.", call. = FALSE)}
        starting_order <- parse_order(orders = starting_order, order_name = "starting_order",
                                      var_names = parsed_formula$z_part$var, kz = parsed_formula$kz)
        if (any(starting_order > max_order)) {stop("'starting_order' can't be greater than 'max_order'.", call. = FALSE)}
    } else {
        starting_order_null <- TRUE
    }
    if (any(fixed_order > max_order)) {stop("'fixed_order' can't be greater than 'max_order'.", call. = FALSE)}
    start_sample <- start
    end_sample <- end

    # acceptable orders for each p and q
    order_list <- lapply(seq_len(length(max_order)), function(i) {
        # because order of y can't be 0
        if (i == 1) {
            if (fixed_order[i] == (-1)) {
                if (is.null(starting_order)) {
                    1:max_order[i]
                } else {
                    starting_order[1]:max_order[i]
                }
            } else {
                fixed_order[i]
            }
        } else {
            if (fixed_order[i] == (-1)) {
                0:max_order[i]
            } else {
                fixed_order[i]
            }
        }
    })

    # starting_order considering fixed_order
    if (is.null(starting_order)) {
        if (grid == FALSE) {
            starting_order <- ifelse(fixed_order[-1] == -1, 0, fixed_order[-1])
            starting_order <- c(ifelse(fixed_order[1] == -1, 1, fixed_order[1]), starting_order)
            # build orders for VAR
            order_var <- lapply(1:min(max_order), function(i) rep(i, parsed_formula$kz))
            if (any(fixed_order != -1)) { # build orders for var with rectricted terms
                for (j in 1:length(order_var)) {
                    order_var[[j]][which(fixed_order != -1)] <- fixed_order[which(fixed_order != -1)]
                }
            }
        }
    } else {
        starting_order <- ifelse(fixed_order == -1, starting_order, fixed_order)
    }

    if (grid == TRUE) {
        # create the full search grid
        order_grid <- expand.grid(order_list, KEEP.OUT.ATTRS = FALSE)

        # run all models and estimate "selection" (using eval to avoid if for each "selection")
        best_selection <- lapply(seq_len(nrow(order_grid)), function(i) {
            eval(parse(text = paste(selection, "(
                                    ardl(formula = formula, data = data, order = unlist(order_grid[", i, ", ]), start = start_sample, end = end_sample, ...)
            )")))
        }) %>%
            unlist()
        if (selection_minmax == "max") best_selection <- (-1)*best_selection
        # keep top 20 orders
        top_orders <- dplyr::bind_cols(order = order_grid, selection = best_selection) %>%
            dplyr::arrange(selection) %>% dplyr::slice(1:20)
    } else {
        top_orders <- c() #initialization
        if (starting_order_null) {
            for_each_p <- 1:length(order_var) #$#$ or try order_list[[1]]
        } else {
            for_each_p <- order_list[[1]]
        }
        for (i in for_each_p) { # for each parse
            if (starting_order_null) {
                order1 <- order_var[[i]] # access each var model
            } else {
                order1 <- starting_order
                order1[1] <- i
            }
            m1 <- ardl(formula = formula, data = data, order = order1, start = start_sample, end = end_sample, ...)
            selection1 <- eval(parse(text = paste(selection, "(m1)")))
            if (selection_minmax == "max") selection1 <- (-1)*selection1
            top_orders <- rbind(top_orders, c(order1, selection1))

            if (search_type == "vertical") {
                for (j in 1:parsed_formula$kx) {
                    if (length(order_list[[j + 1]]) != 1) { # if fixed order, don't search
                        ardl_converge <- FALSE
                        while ((ardl_converge == FALSE) & (order1[j + 1] < dplyr::last(order_list[[j + 1]]))) {
                            order2_back <- order1
                            order2_forth <- order1
                            if (order1[j + 1] == 0) { # so that when q=0 it doesn't search for q<0.
                                # order2_back is the order1
                                order2_forth[j + 1] <- order1[j + 1] + 1
                            } else {
                                order2_back[j + 1] <- order1[j + 1] - 1
                                order2_forth[j + 1] <- order1[j + 1] + 1
                            }
                            order_back_ready <- unlist(lapply(1:nrow(top_orders), FUN = function(z) sum(order2_back == top_orders[z,-ncol(top_orders)])))
                            order_forth_ready <- unlist(lapply(1:nrow(top_orders), FUN = function(z) sum(order2_forth == top_orders[z,-ncol(top_orders)])))
                            if (parsed_formula$kz %in% order_back_ready) {
                                selection_m2_back <- top_orders[which(order_back_ready == parsed_formula$kz)[1], ncol(top_orders)]
                            } else {
                                m2_back <- ardl(formula = formula, data = data, order = order2_back, start = start_sample, end = end_sample, ...)
                                selection_m2_back <- eval(parse(text = paste(selection, "(m2_back)")))
                                if (selection_minmax == "max") selection_m2_back <- (-1)*selection_m2_back
                            }
                            if (parsed_formula$kz %in% order_forth_ready) {
                                selection_m2_forth <- top_orders[which(order_forth_ready == parsed_formula$kz)[1], ncol(top_orders)]
                            } else {
                                m2_forth <- ardl(formula = formula, data = data, order = order2_forth, start = start_sample, end = end_sample, ...)
                                selection_m2_forth <- eval(parse(text = paste(selection, "(m2_forth)")))
                                if (selection_minmax == "max") selection_m2_forth <- (-1)*selection_m2_forth
                            }
                            selection2 <- min(selection_m2_back, selection_m2_forth)[1]
                            order2 <- list(order2_back, order2_forth)[which(c(selection_m2_back, selection_m2_forth) == min(selection_m2_back, selection_m2_forth))][[1]]
                            top_orders <- rbind(top_orders, c(order1, selection1), c(order2, selection2))
                            if (selection2 < selection1) {
                                selection1 <- selection2
                                order1 <- order2
                            } else {
                                ardl_converge <- TRUE
                            }
                        }
                    }
                }
            } else { # horizontal
                ardl_converge <- FALSE
                failed_orders <- data.frame(matrix(0, 1, parsed_formula$kz)) # initialization
                while (ardl_converge == FALSE) {
                    for(j in 1:parsed_formula$kx) {
                        if ((length(order_list[[j + 1]]) != 1) & (order1[j + 1] < max_order[j + 1])) { # if fixed order or max_order reached, don't search
                            order2_back <- order1
                            order2_forth <- order1
                            if (order1[j + 1] == 0) {
                                # order2_back is the order1
                                order2_forth[j + 1] <- order1[j + 1] + 1
                            } else {
                                order2_back[j + 1] <- order1[j + 1] - 1
                                order2_forth[j + 1] <- order1[j + 1] + 1
                            }
                            order_back_ready <- unlist(lapply(1:nrow(top_orders), FUN = function(z) sum(order2_back == top_orders[z,-ncol(top_orders)])))
                            order_forth_ready <- unlist(lapply(1:nrow(top_orders), FUN = function(z) sum(order2_forth == top_orders[z,-ncol(top_orders)])))
                            if (parsed_formula$kz %in% order_back_ready) {
                                selection_m2_back <- top_orders[which(order_back_ready == parsed_formula$kz)[1], ncol(top_orders)]
                            } else {
                                m2_back <- ardl(formula = formula, data = data, order = order2_back, start = start_sample, end = end_sample, ...)
                                selection_m2_back <- eval(parse(text = paste(selection, "(m2_back)")))
                                if (selection_minmax == "max") selection_m2_back <- (-1)*selection_m2_back
                            }
                            if (parsed_formula$kz %in% order_forth_ready) {
                                selection_m2_forth <- top_orders[which(order_forth_ready == parsed_formula$kz)[1], ncol(top_orders)]
                            } else {
                                m2_forth <- ardl(formula = formula, data = data, order = order2_forth, start = start_sample, end = end_sample, ...)
                                selection_m2_forth <- eval(parse(text = paste(selection, "(m2_forth)")))
                                if (selection_minmax == "max") selection_m2_forth <- (-1)*selection_m2_forth
                            }
                            selection2 <- min(selection_m2_back, selection_m2_forth)[1]
                            order2 <- list(order2_back, order2_forth)[which(c(selection_m2_back, selection_m2_forth) == min(selection_m2_back, selection_m2_forth))][[1]]
                            top_orders <- rbind(top_orders, c(order1, selection1), c(order2, selection2))
                            if (selection2 < selection1) {
                                selection1 <- selection2
                                order1 <- order2
                            } else {
                                # to converge when it reaches an order that has already failed
                                if (nrow(dplyr::distinct(rbind(failed_orders, order2), .keep_all = TRUE)) == nrow(failed_orders)) {
                                    ardl_converge <- TRUE
                                } else {
                                    failed_orders <- rbind(failed_orders, order2)
                                }
                            }
                        } else if (!(order1[j + 1] < max_order[j + 1])) { # to avoid infinite loop when all orders are maxed (the case where best mode is the bigest)
                            # Either all fixed or maxed (except the 1st) either order > max (in case of "both" where max = max + 1)
                            if ((all(((order1 == fixed_order) | (order1 == max_order))[-1] == TRUE)) | (order1[j + 1] > max_order[j + 1])) {
                                ardl_converge <- TRUE
                            }
                        }
                    }
                }
            }
        }

        # keep top 20 orders
        top_orders <- dplyr::as_tibble(data.frame(matrix(top_orders[,-ncol(top_orders)], ncol = ncol(top_orders)-1), selection = top_orders[,ncol(top_orders)]) %>%
            dplyr::distinct(.keep_all = TRUE) %>%
            dplyr::arrange(selection) %>% dplyr::slice(1:20))
    }

    # choose best order
    best_order <- top_orders[1, 1:(ncol(top_orders) - 1)] %>% as.numeric()
    names(best_order) <- parsed_formula$z_part$var
    # choose best model
    best_model <- ardl(formula = formula, data = data, order = best_order, start = start_sample, end = end_sample, ...)


    # prepare objects before exporting ----------------------------------------

    # rename top_orders columns
    temp <- names(top_orders)
    names(temp) <- c(parsed_formula$z_part$var, selection)
    top_orders <- top_orders %>%
        dplyr::rename(!!temp)
    # correct sign of the selection criterion
    if (selection_minmax == "max") {
        top_orders[,ncol(top_orders)] <- -top_orders[,ncol(top_orders)]
    }

    # return list
    return_list <- list(best_model = best_model, best_order = best_order,
                        top_orders = data.frame(top_orders))

    return(return_list)
}

#' ARDL: ARDL, ECM and Bounds-Test for Cointegration
"_PACKAGE"

#'
#' @keywords internal
#' @docType package
#' @name ARDL-package
#'
#' @importFrom dplyr %>%
#' @importFrom zoo merge.zoo
#'
# no visible binding for global variable '.' (solution line)
globalVariables(c("."))

.onAttach <- function(libname, pkgname) {
    packageStartupMessage(format(utils::citation("ARDL"), bibtex = FALSE)[[1]],
                      "\n\n",
                      format(utils::citation("ARDL"), bibtex = FALSE)[[2]],
                      "\n\n",
                      format(utils::citation("ARDL"), bibtex = FALSE)[[3]])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsers.R
\name{parse_case}
\alias{parse_case}
\title{Case parser}
\usage{
parse_case(parsed_formula, case)
}
\arguments{
\item{parsed_formula}{A list containing the formula parts as returned from
\code{\link{parse_formula}}.}

\item{case}{An integer from 1-5 or a character string specifying whether the
'intercept' and/or the 'trend' have to participate in the short-run or the
long-run relationship (cointegrating equation) (see section 'Cases' below).}
}
\value{
An integer from 1-5 representing the case.
}
\description{
It parses the 'case' and checks the integrity of the 'case' input and the
compatibility with the formula.
}
\details{
Note that the statistical significance of 'ect' in a RECM should not be
tested using the corresponding t-statistic (or the p-value) because it
doesn't follow a standard t-distribution. Instead, the
\code{\link{bounds_t_test}} should be used.
}
\section{References}{
 Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds
  testing approaches to the analysis of level relationships. \emph{Journal of
  Applied Econometrics}, 16(3), 289-326
}

\seealso{
\code{\link{parse_formula}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crit_val_bounds_sim.R
\name{t_bounds_sim}
\alias{t_bounds_sim}
\title{Critical value bounds stochastic simulation for t-bounds test for no
cointegration}
\usage{
t_bounds_sim(case, k, alpha, T, R = 40000)
}
\arguments{
\item{case}{An integer (1, 3 or 5) specifying whether the 'intercept' and/or
the trend' have to participate in the short-run relationship (see section
'Cases' in \code{\link{bounds_t_test}}).}

\item{k}{The number of independent variables.}

\item{alpha}{A numeric vector between 0 and 1 indicating the significance
level of the critical value bounds. Multiple values can be used.}

\item{T}{An integer indicating the number of observations.}

\item{R}{An integer indicating how many iterations will be used. Default is
40000.}
}
\value{
\code{t_bounds_sim} returns a data frame with the critical value
  bounds for the t-statistic.
}
\description{
\code{t_bounds_sim} simulates the critical value bounds for the t-bounds test
for no cointegration \cite{Pesaran et al. (2001)}.
}
\seealso{
\code{\link{f_bounds_sim}} \code{\link{bounds_t_test}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsers.R
\name{parse_order}
\alias{parse_order}
\title{Order parser}
\usage{
parse_order(orders, order_name, var_names, kz, restriction = FALSE)
}
\arguments{
\item{orders}{A numeric vector of the same length as the total number of
variables (excluding the fixed ones). If the input is \code{order} or
\code{max_order} it should only contain positive integers or 0. If the
input is \code{fixed_order} it should also contain the value '-1'
indicating that a specific order should not be fixed. An integer could be
provided if all variables are of the same order (or all '-1' in the case of
\code{fixed_order}).}

\item{order_name}{The name of the function argument that is passed into
\code{order}.}

\item{var_names}{The names of the variables corresponding to the orders.}

\item{kz}{An integer. The number of dependent and independent variables.}

\item{restriction}{When the input in \code{orders} is either \code{order} or
\code{max_order} it should be \code{FALSE} (default). When the input is
\code{fixed_order} it should be '-1' indicating that the input in
\code{orders} is a restriction for the 'order' of the model (either upper
bound or fixed order).}
}
\value{
A numeric vector of the same length as the total number of variables
  (excluding the fixed ones).
}
\description{
It parses the order and checks the integrity of the order input.
}
\seealso{
\code{\link{parse_formula}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_delay.R
\name{plot_delay}
\alias{plot_delay}
\title{Create plots for the delay multipliers}
\usage{
plot_delay(
  multipliers,
  facets_ncol = 2,
  interval = FALSE,
  interval_color = "blue",
  show.legend = FALSE,
  xlab = "Period",
  ylab = "Delay",
  ...
)
}
\arguments{
\item{multipliers}{A list returned from \code{\link{multipliers}},
in which \code{type} is a positive integer to return delay multipliers.}

\item{facets_ncol}{If a positive integer, it indicates the number of the
columns in the facet. If FALSE, each plot is created separately. The default
is 2.}

\item{interval}{If FALSE (default), no uncertainty intervals are drawn. If a
positive integer, the intervals are this number times the standard error. If
a number between 0 and 1 (e.g. 0.95), the equivalent confidence interval is
drawn (e.g. 95\% CI). In case of the confidence intervals, they are based on
the Gaussian distribution.}

\item{interval_color}{The color of the uncertainty intervals. Default is
"blue".}

\item{show.legend}{A logical indicating whether the interval legend is shown.
Default is FALSE.}

\item{xlab, ylab}{Names displayed at the x and y axes respectively. Default is
"Period" and "Delay" respectively.}

\item{...}{Currently unused argument.}
}
\value{
\code{plot_delay} returns a number of \code{\link[ggplot2]{ggplot}}
objects.
}
\description{
Creates plots for the delay multipliers and their uncertainty intervals based
on their estimated standard errors. This is a basic
\code{\link[ggplot2]{ggplot}} with a few customizable parameters.
}
\examples{
ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
delay_mult <- multipliers(ardl_3132, type = 12, se = TRUE)

## Simply plot the delay multipliers -----------------------------------

plot_delay(delay_mult)

## Rearrange them ------------------------------------------------------

plot_delay(delay_mult, facets_ncol = 1)

## Add 1 standard deviation uncertainty intervals ----------------------

plot_delay(delay_mult, interval = 1)

## Add 95\% confidence intervals, change color and add legend -----------

plot_delay(delay_mult, interval = 0.95, interval_color = "darkgrey",
           show.legend = TRUE)
}
\seealso{
\code{\link{multipliers}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{iplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bounds_test.R
\name{bounds_f_test}
\alias{bounds_f_test}
\title{Bounds Wald-test for no cointegration}
\usage{
bounds_f_test(
  object,
  case,
  alpha = NULL,
  pvalue = TRUE,
  exact = FALSE,
  R = 40000,
  test = c("F", "Chisq"),
  vcov_matrix = NULL
)
}
\arguments{
\item{object}{An object of \code{\link[base]{class}} 'ardl' or 'uecm'.}

\item{case}{An integer from 1-5 or a character string specifying whether the
'intercept' and/or the 'trend' have to participate in the short-run or the
long-run relationship (cointegrating equation) (see section 'Cases' below).}

\item{alpha}{A numeric value between 0 and 1 indicating the significance
level of the critical value bounds. If \code{NULL} (default), no critical
value bounds for a specific level of significance are provide, only the
p-value. See section 'alpha, bounds and p-value' below for details.}

\item{pvalue}{A logical indicating whether you want the p-value to be
provided. The default is \code{TRUE}. See section 'alpha, bounds and
p-value' below for details.}

\item{exact}{A logical indicating whether you want asymptotic (T = 1000) or
exact sample size critical value bounds and p-value. The default is
\code{FALSE} for asymptotic. See section 'alpha, bounds and p-value' below
for details.}

\item{R}{An integer indicating how many iterations will be used if
\code{exact = TRUE}. Default is 40000.}

\item{test}{A character vector indicating whether you want the Wald test to
be expressed as 'F' or as 'Chisq' statistic. Default is "F".}

\item{vcov_matrix}{The estimated covariance matrix of the parameter estimates
of the uecm model. Used to estimate the test statistic. The default is 
\code{vcov(object)} (when \code{vcov_matrix = NULL}), but other estimations
of the covariance matrix of the regression's estimated coefficients can 
also be used (e.g., using \code{\link[sandwich]{vcovHC}} or 
\code{\link[sandwich]{vcovHAC}}). Only applicable if the input object is of
class "uecm".}
}
\value{
A list with class "htest" containing the following components:
  \item{\code{method}}{a character string indicating what type of test was
     performed.}
  \item{\code{alternative}}{a character string describing the alternative
     hypothesis.}
  \item{\code{statistic}}{the value of the test statistic.}
  \item{\code{null.value}}{the value of the population parameters \code{k}
     (the number of independent variables) and \code{T} (the number of
     observations) specified by the null hypothesis.}
  \item{\code{data.name}}{a character string giving the name(s) of the data.}
  \item{\code{parameters}}{numeric vector containing the critical value
     bounds.}
  \item{\code{p.value}}{the p-value of the test.}
  \item{\code{PSS2001parameters}}{numeric vector containing the critical
     value bounds as presented by \cite{Pesaran et al. (2001)}. See section
     'alpha, bounds and p-value' below for details.}
  \item{\code{tab}}{data.frame containing the statistic, the critical value
     bounds, the alpha level of significance and the p-value.}
}
\description{
\code{bounds_f_test} performs the Wald bounds-test for no cointegration
\cite{Pesaran et al. (2001)}. It is a Wald test on the parameters of a UECM
(Unrestricted Error Correction Model) expressed either as a Chisq-statistic
or as an F-statistic.
}
\section{Hypothesis testing}{

  \deqn{\Delta y_{t} = c_{0} + c_{1}t +
  \pi_{y}y_{t-1} + \sum_{j=1}^{k}\pi_{j}x_{j,t-1} +
  \sum_{i=1}^{p-1}\psi_{y,i}\Delta y_{t-i} +
  \sum_{j=1}^{k}\sum_{l=1}^{q_{j}-1} \psi_{j,l}\Delta x_{j,t-l} +
  \sum_{j=1}^{k}\omega_{j}\Delta x_{j,t} + \epsilon_{t}}

\describe{
  \item{Cases 1, 3, 5:}{}
}
  \deqn{\mathbf{H_{0}:} \pi_{y} = \pi_{1} = \dots = \pi_{k} = 0}
  \deqn{\mathbf{H_{1}:} \pi_{y} \neq \pi_{1} \neq \dots \neq \pi_{k} \neq 0}

\describe{
  \item{Case 2:}{}
}
  \deqn{\mathbf{H_{0}:} \pi_{y} = \pi_{1} = \dots = \pi_{k} = c_{0} = 0}
  \deqn{\mathbf{H_{1}:} \pi_{y} \neq \pi_{1} \neq \dots \neq \pi_{k} \neq c_{0} \neq 0}

\describe{
  \item{Case 4:}{}
}
  \deqn{\mathbf{H_{0}:} \pi_{y} = \pi_{1} = \dots = \pi_{k} = c_{1} = 0}
  \deqn{\mathbf{H_{1}:} \pi_{y} \neq \pi_{1} \neq \dots \neq \pi_{k} \neq c_{1} \neq 0}
}

\section{alpha, bounds and p-value}{
 In this section it is explained how the
  critical value bounds and p-values are obtained.
  \itemize{
     \item If \code{exact = FALSE}, then the asymptotic (T = 1000) critical
         value bounds and p-value are provided.
     \item Only the asymptotic critical value bounds and p-values, and only
         for k <= 10 are precalculated, everything else has to be computed.
     \item Precalculated critical value bounds and p-values were simulated
         using \code{set.seed(2020)} and \code{R = 70000}.
     \item Precalculated critical value bounds exist only for \code{alpha}
         being one of the 0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.15 or 0.2,
         everything else has to be computed.
     \item If \code{alpha} is one of the 0.1, 0.05, 0.025 or 0.01 (and
         \code{exact = FALSE} and k <= 10),\cr
         \code{PSS2001parameters} shows the critical value bounds presented
         in \cite{Pesaran et al. (2001)}
         (less precise).
  }
}

\section{Cases}{
 According to \cite{Pesaran et al. (2001)}, we distinguish the
long-run relationship (cointegrating equation) (and thus the bounds-test and
the Restricted ECMs) between 5 different cases. These differ in terms of
whether the 'intercept' and/or the 'trend' are restricted to participate in
the long-run relationship or they are unrestricted and so they participate in
the short-run relationship.

\describe{
  \item{Case 1:}{\itemize{
    \item No \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 1 or "n" where "n" stands for none.}}
  \item{Case 2:}{\itemize{
    \item Restricted \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 2 or "rc" where "rc" stands for restricted
     constant.}}
  \item{Case 3:}{\itemize{
    \item Unrestricted \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 3 or "uc" where "uc" stands for unrestricted
    constant.}}
  \item{Case 4:}{\itemize{
    \item Unrestricted \emph{intercept} and restricted \emph{trend}.
    \item \code{case} inputs: 4 or "ucrt" where "ucrt" stands for
    unrestricted constant and restricted trend.}}
  \item{Case 5:}{\itemize{
    \item Unrestricted \emph{intercept} and unrestricted \emph{trend}.
    \item \code{case} inputs: 5 or "ucut" where "ucut" stands for
     unrestricted constant and unrestricted trend.}}
}

Note that you can't restrict (or leave unrestricted) a parameter that doesn't
exist in the input model. For example, you can't compute \code{recm(object,
case=3)} if the object is an ARDL (or UECM) model with no intercept. The same
way, you can't compute \code{bounds_f_test(object, case=5)} if the object is
an ARDL (or UECM) model with no linear trend.
}

\section{References}{
 Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds
  testing approaches to the analysis of level relationships. \emph{Journal of
  Applied Econometrics}, 16(3), 289-326
}

\examples{
data(denmark)

## How to use cases under different models (regarding deterministic terms)

## Construct an ARDL(3,1,3,2) model with different deterministic terms -

# Without constant
ardl_3132_n <- ardl(LRM ~ LRY + IBO + IDE -1, data = denmark, order = c(3,1,3,2))

# With constant
ardl_3132_c <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))

# With constant and trend
ardl_3132_ct <- ardl(LRM ~ LRY + IBO + IDE + trend(LRM), data = denmark, order = c(3,1,3,2))

## F-bounds test for no level relationship (no cointegration) ----------

# For the model without a constant
bounds_f_test(ardl_3132_n, case = 1)
# or
bounds_f_test(ardl_3132_n, case = "n")

# For the model with a constant
# Including the constant term in the long-run relationship (restricted constant)
bounds_f_test(ardl_3132_c, case = 2)
# or
bounds_f_test(ardl_3132_c, case = "rc")

# Including the constant term in the short-run relationship (unrestricted constant)
bounds_f_test(ardl_3132_c, case = "uc")
# or
bounds_f_test(ardl_3132_c, case = 3)

# For the model with constant and trend
# Including the constant term in the short-run and the trend in the long-run relationship
# (unrestricted constant and restricted trend)
bounds_f_test(ardl_3132_ct, case = "ucrt")
# or
bounds_f_test(ardl_3132_ct, case = 4)

# For the model with constant and trend
# Including the constant term and the trend in the short-run relationship
# (unrestricted constant and unrestricted trend)
bounds_f_test(ardl_3132_ct, case = "ucut")
# or
bounds_f_test(ardl_3132_ct, case = 5)

## Note that you can't restrict a deterministic term that doesn't exist

# For example, the following tests will produce an error:
\dontrun{
bounds_f_test(ardl_3132_c, case = 1)
bounds_f_test(ardl_3132_ct, case = 3)
bounds_f_test(ardl_3132_c, case = 4)
}

## Asymptotic p-value and critical value bounds (assuming T = 1000) ----

# Include critical value bounds for a certain level of significance

# F-statistic is larger than the I(1) bound (for a=0.05) as expected (p-value < 0.05)
bft <- bounds_f_test(ardl_3132_c, case = 2, alpha = 0.05)
bft
bft$tab

# Traditional but less precise critical value bounds, as presented in Pesaran et al. (2001)
bft$PSS2001parameters

# F-statistic is slightly larger than the I(1) bound (for a=0.005)
# as p-value is slightly smaller than 0.005
bounds_f_test(ardl_3132_c, case = 2, alpha = 0.005)

## Exact sample size p-value and critical value bounds -----------------

# Setting a seed is suggested to allow the replication of results
# 'R' can be increased for more accurate resutls

# F-statistic is smaller than the I(1) bound (for a=0.01) as expected (p-value > 0.01)
# Note that the exact sample p-value (0.01285) is very different than the
# asymptotic (0.004418)
# It can take more than 30 seconds
\dontrun{
set.seed(2020)
bounds_f_test(ardl_3132_c, case = 2, alpha = 0.01, exact = TRUE)
}

## "F" and "Chisq" statistics ------------------------------------------

# The p-value is the same, the test-statistic and critical value bounds are different
# but analogous
bounds_f_test(ardl_3132_c, case = 2, alpha = 0.01)
bounds_f_test(ardl_3132_c, case = 2, alpha = 0.01, test = "Chisq")
}
\seealso{
\code{\link{bounds_t_test}} \code{\link{ardl}} \code{\link{uecm}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{htest}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{NT2022}
\alias{NT2022}
\title{The UK earnings equation data from Natsiopoulos and Tzeremes (2022)}
\format{
A time-series object with 196 rows and 9 variables. Time period from 1971:Q1
until 2019:Q4.
\describe{
  \item{time}{time variable}
  \item{w}{real wage}
  \item{Prod}{labor productivity}
  \item{UR}{unemployment rate}
  \item{Wedge}{wedge effect}
  \item{Union}{union power}
  \item{D7475}{income policies 1974:Q1-1975:Q4}
  \item{D7579}{income policies 1975:Q1-1979:Q4}
  \item{UnionR}{union membership}
}
}
\source{
\url{http://qed.econ.queensu.ca/jae/datasets/natsiopoulos001/}
}
\usage{
NT2022
}
\description{
This data set contains the series used by Natsiopoulos and Tzeremes (2022) for re-estimating the UK earnings equation.
The clean format of the data retrieved from the Data Archive of Natsiopoulos and Tzeremes (2022).
}
\details{
An object of class "zooreg" "zoo".
}
\references{
Kleanthis Natsiopoulos and Nickolaos G. Tzeremes, (2022), "ARDL bounds test for
Cointegration: Replicating the Pesaran et al. (2001) Results for the UK Earnings Equation Using R",
\emph{Journal of Applied Econometrics}, \bold{37, 5}, 1079--1090.
\doi{10.1002/jae.2919}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crit_val_bounds_sim.R
\name{f_test_custom}
\alias{f_test_custom}
\title{F-test of regression's overall significance}
\usage{
f_test_custom(dep_var, indep_vars, model_res, const = TRUE)
}
\arguments{
\item{dep_var}{A numeric vector or a matrix with one column representing the
dependent variable.}

\item{indep_vars}{A matrix representing the independent variables.}

\item{model_res}{A numeric vector representing the regression's residuals.}

\item{const}{A logical indicating whether the constant term should be
restricted too.}
}
\value{
\code{f_test_custom} returns a list containing the F-statistic and
  the numerator's degrees of freedom.
}
\description{
\code{f_test_custom} performs an overall significance F-test on a regression.
It is used along with \code{\link[stats]{.lm.fit}} to get the F-statistic as
this is about 10 times faster than extracting it from a regression using
\code{\link[stats]{lm}}.
}
\seealso{
\code{\link{vcov_custom}} \code{\link{f_bounds_sim}}
  \code{\link{t_bounds_sim}}independent
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsers.R
\name{parse_formula}
\alias{parse_formula}
\title{Formula parser}
\usage{
parse_formula(formula, colnames_data)
}
\arguments{
\item{formula}{A "formula" describing the linear model. Details for model
specification are given under 'Details'.}

\item{colnames_data}{A character vector containing the colnames of the data
used in the formula (usually via \code{colnames(data)}).}
}
\value{
A list containing other lists with the names of the dependent,
  independent and fixed variables, the constant and linear trends and the
  number of variables in each category.
}
\description{
It parses the formula and separates the dependent, independent and fixed
variables and also the constant and linear trends (if present).
}
\details{
The notation we follow (e.g., using y, x, z, w etc.) is according to
\cite{Pesaran et al. (2001)}.

The \code{formula} should contain only variables that exist in the data
provided through \code{data} plus some additional functions supported by
\code{\link[dynlm]{dynlm}} (i.e., \code{trend()}).

You can also specify fixed variables that are not supposed to be lagged (e.g.
dummies etc.) simply by placing them after \code{|}. For example, \code{y ~
x1 + x2 | z1 + z2} where \code{z1} and \code{z2} are the fixed variables and
should not be considered in \code{order}. Note that the \code{|} notion
should not be confused with the same notion in \code{dynlm} where it
introduces instrumental variables.
}
\section{References}{
 Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds
  testing approaches to the analysis of level relationships. \emph{Journal of
  Applied Econometrics}, 16(3), 289-326
}

\seealso{
\code{\link{parse_order}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recm.R
\name{recm}
\alias{recm}
\title{Restricted ECM regression}
\usage{
recm(object, case)
}
\arguments{
\item{object}{An object of \code{\link[base]{class}} 'ardl' or 'uecm'.}

\item{case}{An integer from 1-5 or a character string specifying whether the
'intercept' and/or the 'trend' have to participate in the short-run or the
long-run relationship (cointegrating equation) (see section 'Cases' below).}
}
\value{
\code{recm} returns an object of \code{\link[base]{class}}
  \code{c("recm", "dynlm", "lm")}. In addition, attributes 'order', 'data',
  'parsed_formula' and 'full_formula' are provided.
}
\description{
Creates the Restricted Error Correction Model (RECM). This is the conditional
RECM, which is the RECM of the underlying ARDL.
}
\details{
Note that the statistical significance of 'ect' in a RECM should not be
tested using the corresponding t-statistic (or the p-value) because it
doesn't follow a standard t-distribution. Instead, the
\code{\link{bounds_t_test}} should be used.
}
\section{Mathematical Formula}{
 The formula of a Restricted ECM conditional to
  an \eqn{ARDL(p,q_{1},\dots,q_{k})}{ARDL(p,q1,...,qk)} is: \deqn{\Delta
  y_{t} = c_{0} + c_{1}t + \sum_{i=1}^{p-1}\psi_{y,i}\Delta y_{t-i} +
  \sum_{j=1}^{k}\sum_{l=1}^{q_{j}-1} \psi_{j,l}\Delta x_{j,t-l} +
  \sum_{j=1}^{k}\omega_{j}\Delta x_{j,t} + \pi_{y}ECT_{t} + \epsilon_{t}}
  \deqn{\psi_{j,l} = 0 \;\; \forall \;\; q_{j} = 1, \psi_{j,l} = \omega_{j} =
   0 \;\; \forall \;\; q_{j} = 0}
\describe{
  \item{Under Case 1:}{\itemize{
     \item \eqn{c_{0}=c_{1}=0}
     \item \eqn{ECT = y_{t-1} - (\sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
  \item{Under Case 2:}{\itemize{
     \item \eqn{c_{0}=c_{1}=0}
     \item \eqn{ECT = y_{t-1} - (\mu + \sum_{j=1}^{k}\theta_{j} x_{j,t-1})}}}
  \item{Under Case 3:}{\itemize{
     \item \eqn{c_{1}=0}
     \item \eqn{ECT = y_{t-1} - (\sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
  \item{Under Case 4:}{\itemize{
     \item \eqn{c_{1}=0}
     \item \eqn{ECT = y_{t-1} - (\delta(t-1)+ \sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
  \item{Under Case 5:}{\itemize{
     \item \eqn{ECT = y_{t-1} - (\sum_{j=1}^{k} \theta_{j} x_{j,t-1})}}}
}
In all cases, \eqn{x_{j,t-1}} in \eqn{ECT} is replaced by \eqn{x_{j,t} \;\;\;\;\; \forall \;\; q_{j} = 0}
}

\section{Cases}{
 According to \cite{Pesaran et al. (2001)}, we distinguish the
long-run relationship (cointegrating equation) (and thus the bounds-test and
the Restricted ECMs) between 5 different cases. These differ in terms of
whether the 'intercept' and/or the 'trend' are restricted to participate in
the long-run relationship or they are unrestricted and so they participate in
the short-run relationship.

\describe{
  \item{Case 1:}{\itemize{
    \item No \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 1 or "n" where "n" stands for none.}}
  \item{Case 2:}{\itemize{
    \item Restricted \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 2 or "rc" where "rc" stands for restricted
     constant.}}
  \item{Case 3:}{\itemize{
    \item Unrestricted \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 3 or "uc" where "uc" stands for unrestricted
    constant.}}
  \item{Case 4:}{\itemize{
    \item Unrestricted \emph{intercept} and restricted \emph{trend}.
    \item \code{case} inputs: 4 or "ucrt" where "ucrt" stands for
    unrestricted constant and restricted trend.}}
  \item{Case 5:}{\itemize{
    \item Unrestricted \emph{intercept} and unrestricted \emph{trend}.
    \item \code{case} inputs: 5 or "ucut" where "ucut" stands for
     unrestricted constant and unrestricted trend.}}
}

Note that you can't restrict (or leave unrestricted) a parameter that doesn't
exist in the input model. For example, you can't compute \code{recm(object,
case=3)} if the object is an ARDL (or UECM) model with no intercept. The same
way, you can't compute \code{bounds_f_test(object, case=5)} if the object is
an ARDL (or UECM) model with no linear trend.
}

\section{References}{
 Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds
  testing approaches to the analysis of level relationships. \emph{Journal of
  Applied Econometrics}, 16(3), 289-326
}

\examples{
data(denmark)

## Estimate the RECM, conditional to it's underlying ARDL(3,1,3,2) -----

# Indirectly from an ARDL
ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
recm_3132 <- recm(ardl_3132, case = 2)

# Indirectly from an UECM
uecm_3132 <- uecm(ardl_3132)
recm_3132_ <- recm(uecm_3132, case = 2)
identical(recm_3132, recm_3132_)
summary(recm_3132)

## Error Correction Term (ect) & Speed of Adjustment -------------------

# The coefficient of the ect,
# shows the Speed of Adjustment towards equilibrium.
# Note that this can be also be obtained from an UECM,
# through the coefficient of the term L(y, 1) (where y is the dependent variable).
tail(recm_3132$coefficients, 1)
uecm_3132$coefficients[2]

## Post-estimation testing ---------------------------------------------

# See examples in the help file of the uecm() function

}
\seealso{
\code{\link{ardl}} \code{\link{uecm}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{models}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/to_lm.R
\name{to_lm}
\alias{to_lm}
\title{Convert dynlm model (ardl, uecm, recm) to lm model}
\usage{
to_lm(object, fix_names = FALSE, data_class = NULL, ...)
}
\arguments{
\item{object}{An object of \code{\link[base]{class}} 'ardl', 'uecm' or 'recm'.}

\item{fix_names}{A logical, indicating whether the variable names should be
rewritten without special functions and character in the names such as "d()"
or "L()". When \code{fix_names = TRUE}, the characters "(", and "," are
replaces with ".", and ")" and spaces are deleted. The name of the dependent
variable is always transformed, regardless of the value of this parameter.
Default is FALSE.}

\item{data_class}{If "ts", it converts the data class to
\code{\link[stats]{ts}} (see examples for its usage). The default is
\code{\link[base]{NULL}}, which uses the same data provided in the original
object.}

\item{...}{Currently unused argument.}
}
\value{
\code{to_lm} returns an object of \code{\link[base]{class}}
\code{"lm"}.
}
\description{
Takes a \code{\link[dynlm]{dynlm}} model of \code{\link[base]{class}} 'ardl',
'uecm' or 'recm' and converts it into an \code{\link[stats]{lm}} model. This
can help using the model as a regular \code{\link[stats]{lm}} model with
functions that are not compatible with \code{\link[dynlm]{dynlm}} models such
as the \code{\link[stats]{predict}} function to forecast.
}
\examples{
## Convert ARDL into lm ------------------------------------------------

ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
ardl_3132_lm <- to_lm(ardl_3132)
summary(ardl_3132)$coefficients
summary(ardl_3132_lm)$coefficients

## Convert UECM into lm ------------------------------------------------

uecm_3132 <- uecm(ardl_3132)
uecm_3132_lm <- to_lm(uecm_3132)
summary(uecm_3132)$coefficients
summary(uecm_3132_lm)$coefficients

## Convert RECM into lm ------------------------------------------------

recm_3132 <- recm(ardl_3132, case = 2)
recm_3132_lm <- to_lm(recm_3132)
summary(recm_3132)$coefficients
summary(recm_3132_lm)$coefficients

## Use the lm model to forecast ----------------------------------------

# Forecast using the in-sample data
insample_data <- ardl_3132$model
head(insample_data)
predicted_values <- predict(ardl_3132_lm, newdata = insample_data)

# The predicted values are expected to be the same as the fitted values
ardl_3132$fitted.values
predicted_values

# Convert to ts class for the plot
predicted_values <- ts(predicted_values, start = c(1974,4), frequency=4)
plot(denmark$LRM, lwd=4) #The input dependent variable
lines(ardl_3132$fitted.values, lwd=4, col="blue") #The fitted values
lines(predicted_values, lty=2, lwd=2, col="red") #The predicted values

## Convert to lm for post-estimation testing ---------------------------

# Ramsey's RESET test for functional form
library(lmtest) # for resettest()
library(strucchange) # for efp(), and sctest()

\dontrun{
    # This produces an error.
    # resettest() cannot use data of class 'zoo' such as the 'denmark' data
    # used to build the original model
    resettest(uecm_3132, type = c("regressor"))
}

uecm_3132_lm <- to_lm(uecm_3132, data_class = "ts")
resettest(uecm_3132_lm, power = 2)

# CUSUM test for structural change detection
\dontrun{
    # This produces an error.
    # efp() does not understand special functions such as "d()" and "L()"
    efp(uecm_3132$full_formula, data = uecm_3132$model)
}

uecm_3132_lm_names <- to_lm(uecm_3132, fix_names = TRUE)
fluctuation <- efp(uecm_3132_lm_names$full_formula,
                   data = uecm_3132_lm_names$model)
sctest(fluctuation)
plot(fluctuation)

}
\seealso{
\code{\link{ardl}}, \code{\link{uecm}}, \code{\link{recm}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{models}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formula_builders.R
\name{build_uecm_formula}
\alias{build_uecm_formula}
\title{UECM formula specification builder}
\usage{
build_uecm_formula(parsed_formula, order)
}
\arguments{
\item{parsed_formula}{A list containing the formula parts as returned from
\code{\link{parse_formula}}.}

\item{order}{A numeric vector with the underlying ARDL order as returned from
\code{\link[=parse_order]{parse_order(restriction = FALSE)}}.}
}
\value{
\code{build_uecm_formula} returns a list containing the full formula
  and the independent and dependent parts of the formula separated. The full
  formula is ready to be used as input in the \code{dynlm} function.
}
\description{
It creates the UECM (Unrestricted Error Correction Model) specification
according to the given "formula" and the corresponding "order" of the
underlying ARDL.
}
\seealso{
\code{\link{build_ardl_formula}}, \code{\link{build_recm_formula}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multipliers.R
\name{multipliers}
\alias{multipliers}
\alias{multipliers.ardl}
\alias{multipliers.uecm}
\title{Multipliers estimation}
\usage{
multipliers(object, type = "lr", vcov_matrix = NULL, se = FALSE)

\method{multipliers}{ardl}(object, type = "lr", vcov_matrix = NULL, se = FALSE)

\method{multipliers}{uecm}(object, type = "lr", vcov_matrix = NULL, se = FALSE)
}
\arguments{
\item{object}{An object of \code{\link[base]{class}} 'ardl' or 'uecm'.}

\item{type}{A character string describing the type of multipliers. Use "lr"
for long-run (total) multipliers (default), "sr" or 0 for short-run (impact)
multipliers or an integer between 1 and 200 for delay and interim multipliers.}

\item{vcov_matrix}{The estimated covariance matrix of the parameter estimates.
Used by the transformation function to estimate the standard errors (and
so the t-statistics and p-values) of the multipliers. The default is
\code{vcov(object)} (when \code{vcov_matrix = NULL}), but other estimations
of the covariance matrix of the regression's estimated coefficients can
also be used (e.g., using \code{\link[sandwich]{vcovHC}} or
\code{\link[sandwich]{vcovHAC}}).}

\item{se}{A logical indicating whether you want standard errors for delay
multipliers to be provided. The default is FALSE. Note that this parameter
does not refer to the standard errors for the long-run and short-run
multipliers, for which are always calculated. IMPORTANT: Calculating standard
errors for long periods of delays may cause your computer to run out of
memory and terminate your R session, losing important unsaved work. As a rule
of thumb, try not to exceed \code{type = 19} when \code{se = TRUE}.}
}
\value{
\code{multipliers} returns (for long and short run multipliers) a
  data.frame containing the independent variables (including possibly
  existing intercept or trend and excluding the fixed variables) and their
  corresponding standard errors, t-statistics and p-values. For delay and
  interim multipliers it returns a list with a data.frame for each variable,
  containing the delay and interim multipliers for each period.
}
\description{
\code{multipliers} is a generic function used to estimate short-run (impact),
delay, interim and long-run (total) multipliers, accompanied by their
corresponding standard errors, t-statistics and p-values.
}
\details{
The function invokes two different \code{\link[utils]{methods}}, one for
objects of \code{\link[base]{class}} 'ardl' and one for objects of
\code{class} 'uecm'. This is because of the different (but equivalent)
transformation functions that are used for each class/model ('ardl' and
'uecm') to estimate the multipliers.

\code{type = 0} is equivalent to \code{type = "sr"}.

Note that the interim multipliers are the cumulative sum of the delays, and
that the sum of the delay multipliers (for long enough periods) and thus
a distant enough interim multiplier match the long-run multipliers.

The delay (interim) multiplier can be interpreted as the effect on the
dependent variable in period t+s, resulting from an instant (sustained) shock
to an independent variable in period t.

The delta method is used for approximating the standard errors (and thus the
t-statistics and p-values) of the estimated long-run and delay multipliers.
}
\section{Mathematical Formula}{


\strong{Short-Run Multipliers:}
\describe{
  \item{As derived from an ARDL:}{}
}
\deqn{\frac{\partial y_{t}}{\partial x_{j,t}} = b_{j,0} \;\;\;\;\; j \in \{1,\dots,k\}}

\describe{
  \item{As derived from an Unrestricted ECM:}{}
}
\deqn{\frac{\partial y_{t}}{\partial x_{j,t}} = \omega_{j} \;\;\;\;\; j \in \{1,\dots,k\}}

\describe{
  \item{Constant and Linear Trend:}{}
}
\deqn{c_{0}}
\deqn{c_{1}}

\strong{Delay & Interim Multipliers:}
\describe{
  \item{As derived from an ARDL:}{}
}
\deqn{Delay_{x_{j},s} = \frac{\partial y_{t+s}}{\partial x_{j,t}} = b_{j,s} + \sum_{i=1}^{min\{p,s\}} b_{y,i} \frac{\partial y_{t+(s-i)}}{\partial x_{j,t}} \;\;\;\;\; b_{j,s} = 0 \;\; \forall \;\; s > q}
\deqn{Interim_{x_{j},s} = \sum_{i=0}^{s} Delay_{x_{j},s}}

\describe{
  \item{Constant and Linear Trend:}{}
}
\deqn{Delay_{intercept,s} = c_{0} + \sum_{i=1}^{min\{p,s\}} b_{y,i} Delay_{intercept,s-i} \;\;\;\;\; c_{0} = 0 \;\; \forall \;\; s \neq 0}
\deqn{Interim_{intercept,s} = \sum_{i=0}^{s} Delay_{intercept,s}}
\deqn{Delay_{trend,s} = c_{1} + \sum_{i=1}^{min\{p,s\}} b_{y,i} Delay_{trend,s-i} \;\;\;\;\; c_{1} = 0 \;\; \forall \;\; s \neq 0}
\deqn{Interim_{trend,s} = \sum_{i=0}^{s} Delay_{trend,s}}

\strong{Long-Run Multipliers:}
\describe{
  \item{As derived from an ARDL:}{}
}
\deqn{\frac{\partial y_{t+\infty}}{\partial x_{j,t}} = \theta_{j} = \frac{\sum_{l=0}^{q_{j}}b_{j,l}}{1-\sum_{i=1}^{p}b_{y,i}} \;\;\;\;\; j \in \{1,\dots,k\}}
\describe{
  \item{Constant and Linear Trend:}{}
}
\deqn{\mu = \frac{c_{0}}{1-\sum_{i=1}^{p}b_{y,i}}}
\deqn{\delta = \frac{c_{1}}{1-\sum_{i=1}^{p}b_{y,i}}}

\describe{
  \item{As derived from an Unrestricted ECM:}{}
}
\deqn{\frac{\partial y_{t+\infty}}{\partial x_{j,t}} = \theta_{j} = \frac{\pi_{j}}{-\pi_{y}} \;\;\;\;\; j \in \{1,\dots,k\}}
\describe{
  \item{Constant and Linear Trend:}{}
}
\deqn{\mu = \frac{c_{0}}{-\pi_{y}}}
\deqn{\delta = \frac{c_{1}}{-\pi_{y}}}
}

\examples{
data(denmark)

## Estimate the long-run multipliers of an ARDL(3,1,3,2) model ---------

# From an ARDL model
ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
mult_ardl <- multipliers(ardl_3132)
mult_ardl

# From an UECM
uecm_3132 <- uecm(ardl_3132)
mult_uecm <- multipliers(uecm_3132)
mult_uecm

all.equal(mult_ardl, mult_uecm)


## Estimate the short-run multipliers of an ARDL(3,1,3,2) model --------

mult_sr <- multipliers(uecm_3132, type = "sr")
mult_0 <- multipliers(uecm_3132, type = 0)
all.equal(mult_sr, mult_0)


## Estimate the delay & interim multipliers of an ARDL(3,1,3,2) model --

mult_lr <- multipliers(uecm_3132, type = "lr")
mult_inter80 <- multipliers(uecm_3132, type = 80)

mult_lr
sum(mult_inter80$`(Intercept)`$Delay)
mult_inter80$`(Intercept)`$Interim[nrow(mult_inter80$`(Intercept)`)]
sum(mult_inter80$LRY$Delay)
mult_inter80$LRY$Interim[nrow(mult_inter80$LRY)]
sum(mult_inter80$IBO$Delay)
mult_inter80$IBO$Interim[nrow(mult_inter80$IBO)]
sum(mult_inter80$IDE$Delay)
mult_inter80$IDE$Interim[nrow(mult_inter80$IDE)]
plot(mult_inter80$LRY$Delay, type='l')
plot(mult_inter80$LRY$Interim, type='l')

mult_inter12 <- multipliers(uecm_3132, type = 12, se = TRUE)
plot_delay(mult_inter12, interval = 0.95)
}
\seealso{
\code{\link{ardl}}, \code{\link{uecm}}, \code{\link{plot_delay}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{math}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.ardl}
\alias{predict.ardl}
\title{Predict method for ARDL model fits}
\usage{
\method{predict}{ardl}(object, newdata, ...)
}
\arguments{
\item{object}{A fitted model of \code{\link[base]{class}} \code{\link{ardl}}.}

\item{newdata}{A time series object (\code{\link[stats]{ts}},
\code{\link[zoo]{zoo}}, \code{\link[zoo]{zooreg}}) or a
\code{\link[base]{data.frame}} containing the new values of the independent
variables \eqn{x} required to make predictions.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
\code{predict.ardl} returns values for \eqn{\hat{y}_{T+1}, ...,
\hat{y}_{T+h}} equal in length to the number of rows in \code{newdata},
as a \code{\link[base]{vector}}.
}
\description{
Makes predictions based on fitted ARDL models, extending the
generic function \code{\link[stats]{predict}} for use with
\code{\link{ardl}}. It is designed to take advantage of lags of the dependent
variable (AR components) from the fitted model and new values of the
independent variable(s) to calculate predictions.
}
\details{
\code{predict.ardl} works recursively to calculate predictions based on
\eqn{y_{t-1}, ..., y_{t-p}} values contained within a fitted model, and new
values of \eqn{x_{T+1}, ..., x_{T+h}}, where \eqn{T} denotes the last observation
of \eqn{t} from the original data and \eqn{h} is the number of new
observations that will be used to calculate predictions. These are supplied
through the \code{newdata} argument. It is important to note that for a given
prediction \eqn{\hat{y}_{T+1}}, the independent variables provided through
\code{newdata}, are expected to follow consecutively from the last observation
\eqn{x_{T}} used to create the model, i.e. \eqn{x_{T+1}, ..., x_{T+h}},
with no gaps. If a time series object is passed to the function, the frequency
and whether the observations follow consecutively will be checked.
}
\section{Calculation example}{


We want to predict \eqn{\hat{y}_{T+1}} given an \eqn{ARDL(2,2)} model:

\deqn{
     \hat{y}_{T+1} = Z \boldsymbol{\beta}
}
where \eqn{Z} is our design matrix and \eqn{\beta} is a vector of our model
coefficients. We need 2 lags for our dependent and independent variables:
\eqn{y_T}, \eqn{y_{T-1}} and \eqn{X_T}, \eqn{X_{T-1}}.
Plus, \eqn{X_{T+1}} (from \code{newdata}), which in our design matrix is
equivalent to time \eqn{t} in our original model:
\deqn{
Z =
    \left[
        \begin{array}{c|cccccc}
        & \mathbf{c} & \mathbf{y_{t-1}} & \mathbf{y_{t-2}} & \mathbf{x_t} &
        \mathbf{x_{t-1}} & \mathbf{x_{t-2}} \\
        \hline
        T\!+\!1 & 1 & y_{T} & y_{T-1} & x_{T+1} & x_{T} & x_{T-1}
        \end{array}
        \right]
}
We proceed to the next iteration, predicting \eqn{\hat{y}_{T+2}}, using the
previously calculated value of \eqn{\hat{y}_{T+1}}:
\deqn{
Z =
\left[
    \begin{array}{c|cccccc}
    & \mathbf{c} & \mathbf{y_{t-1}} & \mathbf{y_{t-2}} & \mathbf{x_t} & \mathbf{x_{t-1}} & \mathbf{x_{t-2}} \\
    \hline
    T\!+\!2 & 1 & y_{T+1} & y_{T} & x_{T+2} & x_{T+1} & x_{T}
    \end{array}
    \right]
}

Iteration continues until we've calculated all the necessary values of
\eqn{\hat{y}_{T+h}} for \code{newdata}.
}

\examples{
data(denmark)
## Estimate an ARDL(2,1,2,1,1) model, using a subset of the denmark data
ardl_21211 <- ardl(LRM ~ LRY + LPY + IBO + IDE, data = denmark,
                   order = c(2,1,2,1,1),
                   start = "1974 Q1", end = "1986 Q2")
## Make predictions based on the remainder of the denmark data not used in
## training the model
predict_data <- window(denmark, start = "1986 Q3", end = "1987 Q3")
## Drop the dependent variable, since we're predicting it
predict_data <- predict_data[, c("LRY", "LPY", "IBO", "IDE")]
predictions <- predict(ardl_21211, predict_data)

## Estimate a ARDL(4,4,4,4) model with a linear trend and dummies
## Create dummies
d_74Q1_75Q3 <- ifelse(time(denmark) >= "1974 Q1" & time(denmark) <= "1975 Q3", 1, 0)
# Add them to the data
denmark_dums <- cbind(denmark, d_74Q1_75Q3)
## Estimate the model
ardl_4444 <- ardl(LRM ~ LRY + LPY + IBO + trend(LRM) | d_74Q1_75Q3,
                  data = denmark_dums,
                  order = c(4,4,4,4),
                  start = "1974 Q1", end = "1986 Q2")
## Take the remaining data not used in training the model
predict_data <- window(denmark_dums, start = "1986 Q3", end = "1987 Q3")
## Drop the dependent variable
predict_data <- predict_data[, c("LRY", "LPY", "IBO", "d_74Q1_75Q3")]
## Compute the predictions
predictions <- predict(ardl_4444, predict_data)

}
\seealso{
\code{\link{ardl}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}

Daniel Finnan, \email{dan@custom-made.org.uk}
}
\keyword{models}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delta_method.R
\name{delta_method}
\alias{delta_method}
\alias{delta_method.ardl}
\alias{delta_method.uecm}
\title{Delta method}
\usage{
delta_method(object, vcov_matrix = NULL)

\method{delta_method}{ardl}(object, vcov_matrix = NULL)

\method{delta_method}{uecm}(object, vcov_matrix = NULL)
}
\arguments{
\item{object}{An object of \code{\link[base]{class}} 'ardl' or 'uecm'.}

\item{vcov_matrix}{The estimated covariance matrix of the parameter estimates.
Used by the transformation function to estimate the standard errors (and
so the t-statistics and p-values) of the multipliers. The default is
\code{vcov(object)} (when \code{vcov_matrix = NULL}), but other estimations
of the covariance matrix of the regression's estimated coefficients can
also be used (e.g., using \code{\link[sandwich]{vcovHC}} or
\code{\link[sandwich]{vcovHAC}}).}
}
\value{
\code{delta_method} returns a numeric vector of the same length as
  the number of the independent variables (excluding the fixed ones) in the
  model.
}
\description{
An internal generic function, customized for approximating the standard
errors of the estimated multipliers.
}
\details{
The function invokes two different \code{\link[utils]{methods}}, one for
objects of \code{\link[base]{class}} 'ardl' and one for objects of
\code{class} 'uecm'. This is because of the different (but equivalent)
transformation functions that are used for each class/model ('ardl' and
'uecm') to estimate the multipliers.
}
\seealso{
\code{\link{multipliers}}, \code{\link{ardl}}, \code{\link{uecm}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ardl-package.R
\docType{package}
\name{ARDL-package}
\alias{ARDL-package}
\title{ARDL: ARDL, ECM and Bounds-Test for Cointegration}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Creates complex autoregressive distributed lag (ARDL) models and constructs the underlying unrestricted and restricted error correction model (ECM) automatically, just by providing the order. It also performs the bounds-test for cointegration as described in Pesaran et al. (2001) \doi{10.1002/jae.616} and provides the multipliers and the cointegrating equation. The validity and the accuracy of this package have been verified by successfully replicating the results of Pesaran et al. (2001) in Natsiopoulos and Tzeremes (2022) \doi{10.1002/jae.2919}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/Natsiopoulos/ARDL}
  \item Report bugs at \url{https://github.com/Natsiopoulos/ARDL/issues}
}

}
\author{
\strong{Maintainer}: Kleanthis Natsiopoulos \email{klnatsio@gmail.com} (\href{https://orcid.org/0000-0003-1180-2984}{ORCID})

Authors:
\itemize{
  \item Kleanthis Natsiopoulos \email{klnatsio@gmail.com} (\href{https://orcid.org/0000-0003-1180-2984}{ORCID})
  \item Nickolaos Tzeremes \email{bus9nt@econ.uth.gr} (\href{https://orcid.org/0000-0002-6938-3404}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{PSS2001}
\alias{PSS2001}
\title{The UK earnings equation data from Pesaran et al. (2001)}
\format{
A time-series object with 112 rows and 7 variables. Time period from 1970:Q1
until 1997:Q4.
\describe{
  \item{w}{real wage}
  \item{Prod}{labor productivity}
  \item{UR}{unemployment rate}
  \item{Wedge}{wedge effect}
  \item{Union}{union power}
  \item{D7475}{income policies 1974:Q1-1975:Q4}
  \item{D7579}{income policies 1975:Q1-1979:Q4}
}
}
\source{
\url{http://qed.econ.queensu.ca/jae/datasets/pesaran001/}\cr
\url{http://qed.econ.queensu.ca/jae/datasets/natsiopoulos001/}
}
\usage{
PSS2001
}
\description{
This data set contains the series used by Pesaran et al. (2001) for estimating the UK earnings equation.
The clean format of the data retrieved from the Data Archive of Natsiopoulos and Tzeremes (2022).
}
\details{
An object of class "zooreg" "zoo".
}
\references{
M. Hashem Pesaran, Richard J. Smith, and Yongcheol Shin, (2001), "Bounds Testing
Approaches to the Analysis of Level Relationships", \emph{Journal of Applied
Econometrics}, \bold{16, 3}, 289--326.

Kleanthis Natsiopoulos and Nickolaos G. Tzeremes, (2022), "ARDL bounds test for
Cointegration: Replicating the Pesaran et al. (2001) Results for the
UK Earnings Equation Using R",
\emph{Journal of Applied Econometrics}, \bold{37, 5}, 1079--1090.
\doi{10.1002/jae.2919}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ardl.R
\name{ardl}
\alias{ardl}
\alias{ardl.uecm}
\alias{ardl.default}
\title{ARDL model regression}
\usage{
ardl(...)

\method{ardl}{uecm}(object, ...)

\method{ardl}{default}(formula, data, order, start = NULL, end = NULL, ...)
}
\arguments{
\item{...}{Additional arguments to be passed to the low level regression
fitting functions.}

\item{object}{An object of \code{\link[base]{class}} 'uecm'.}

\item{formula}{A "formula" describing the linear model. Details for model
specification are given under 'Details'.}

\item{data}{A time series object (e.g., "ts", "zoo" or "zooreg") or a data
frame containing the variables in the model. In the case of a data frame,
it is coerced into a \code{\link[stats]{ts}} object with \code{start = 1},
\code{end = nrow(data)} and \code{frequency = 1}. If not found in data, the
variables are NOT taken from any environment.}

\item{order}{A specification of the order of the ARDL model. A numeric vector
of the same length as the total number of variables (excluding the fixed
ones, see 'Details'). It should only contain positive integers or 0. An
integer could be provided if all variables are of the same order.}

\item{start}{Start of the time period which should be used for fitting the
model.}

\item{end}{End of the time period which should be used for fitting the model.}
}
\value{
\code{ardl} returns an object of \code{\link[base]{class}}
  \code{c("ardl", "dynlm", "lm")}. In addition, attributes 'order', 'data',
  'parsed_formula' and 'full_formula' are provided.
}
\description{
A simple way to construct complex ARDL specifications providing just the
model order additional to the model formula. It uses
\code{\link[dynlm]{dynlm}} under the hood. \code{ardl} is a generic function
and the default method constructs an 'ardl' model while the other method
takes a model of \code{\link[base]{class}} 'uecm' and converts in into an
'ardl'.
}
\details{
The \code{formula} should contain only variables that exist in the data
provided through \code{data} plus some additional functions supported by
\code{\link[dynlm]{dynlm}} (i.e., \code{trend()}).

You can also specify fixed variables that are not supposed to be lagged (e.g.
dummies etc.) simply by placing them after \code{|}. For example, \code{y ~
x1 + x2 | z1 + z2} where \code{z1} and \code{z2} are the fixed variables and
should not be considered in \code{order}. Note that the \code{|} notion
should not be confused with the same notion in \code{dynlm} where it
introduces instrumental variables.
}
\section{Mathematical Formula}{

The general form of an \eqn{ARDL(p,q_{1},\dots,q_{k})}{ARDL(p,q1,...,qk)} is:
\deqn{y_{t} = c_{0} + c_{1}t + \sum_{i=1}^{p}b_{y,i}y_{t-i} +
\sum_{j=1}^{k}\sum_{l=0}^{q_{j}}b_{j,l}x_{j,t-l} + \epsilon_{t}}
}

\examples{
data(denmark)

## Estimate an ARDL(3,1,3,2) model -------------------------------------

ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
summary(ardl_3132)

## Add dummies or other variables that should stay fixed ---------------

d_74Q1_75Q3 <- ifelse(time(denmark) >= 1974 & time(denmark) <= 1975.5, 1, 0)

# the date can also be setted as below
d_74Q1_75Q3_ <- ifelse(time(denmark) >= "1974 Q1" & time(denmark) <= "1975 Q3", 1, 0)
identical(d_74Q1_75Q3, d_74Q1_75Q3_)
den <- cbind(denmark, d_74Q1_75Q3)
ardl_3132_d <- ardl(LRM ~ LRY + IBO + IDE | d_74Q1_75Q3,
                    data = den, order = c(3,1,3,2))
summary(ardl_3132_d)
compare <- data.frame(AIC = c(AIC(ardl_3132), AIC(ardl_3132_d)),
                      BIC = c(BIC(ardl_3132), BIC(ardl_3132_d)))
rownames(compare) <- c("no dummy", "with dummy")
compare

## Estimate an ARDL(3,1,3,2) model with a linear trend -----------------

ardl_3132_tr <- ardl(LRM ~ LRY + IBO + IDE + trend(LRM),
                     data = denmark, order = c(3,1,3,2))

# Alternative time trend specifications:
# time(LRM)                 1974 + (0, 1, ..., 55)/4 time(data)
# trend(LRM)                (1, 2, ..., 55)/4        (1:n)/freq
# trend(LRM, scale = FALSE) (1, 2, ..., 55)          1:n

## Subsample ARDL regression (start after 1975 Q4) ---------------------

ardl_3132_sub <- ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                      order = c(3,1,3,2), start = "1975 Q4")

# the date can also be setted as below
ardl_3132_sub2 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                       order = c(3,1,3,2), start = c(1975,4))
identical(ardl_3132_sub, ardl_3132_sub2)
summary(ardl_3132_sub)

## Ease of use ---------------------------------------------------------

# The model specification of the ardl_3132 model can be created as easy as order=c(3,1,3,2)
# or else, it could be done using the dynlm package as:
library(dynlm)
m <- dynlm(LRM ~ L(LRM, 1) + L(LRM, 2) + L(LRM, 3) + LRY + L(LRY, 1) + IBO + L(IBO, 1) +
           L(IBO, 2) + L(IBO, 3) + IDE + L(IDE, 1) + L(IDE, 2), data = denmark)
identical(m$coefficients, ardl_3132$coefficients)

# The full formula can be extracted from the ARDL model, and this is equal to
ardl_3132$full_formula
m2 <- dynlm(ardl_3132$full_formula, data = ardl_3132$data)
identical(m$coefficients, m2$coefficients)

## Post-estimation testing ---------------------------------------------

# See examples in the help file of the uecm() function

}
\seealso{
\code{\link{uecm}}, \code{\link{recm}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{models}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_lr.R
\name{plot_lr}
\alias{plot_lr}
\title{Create plot for the long-run (cointegrating) equation}
\usage{
plot_lr(
  object,
  coint_eq,
  facets = FALSE,
  show_fitted = FALSE,
  show.legend = FALSE,
  xlab = "Time",
  ...
)
}
\arguments{
\item{object}{An object of \code{\link[base]{class}} `ardl`.}

\item{coint_eq}{The objected returned from \code{\link{coint_eq}}.}

\item{facets}{A logical indicating whether the long-run relationship appears
in a separate plot. Default is FALSE.}

\item{show_fitted}{A logical indicating whether the fitted values are shown.
Default is FALSE.}

\item{show.legend}{A logical indicating whether the legend is shown.
Default is FALSE.}

\item{xlab}{Name displayed at the x axis. Default is "Time".}

\item{...}{Currently unused argument.}
}
\value{
\code{plot_lr} returns a \code{\link[ggplot2]{ggplot}} object.
}
\description{
Creates a plot for the long-run relationship in comparison with the dependent
variable, and the fitted values of the model. This is a basic
\code{\link[ggplot2]{ggplot}} with a few customizable parameters.
}
\examples{
ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
ce2 <- coint_eq(ardl_3132, case = 2)

plot_lr(ardl_3132, coint_eq = ce2)

## Compare fitted values and place long-run relationship separately ----

ce3 <- coint_eq(ardl_3132, case = 3)
plot_lr(ardl_3132, coint_eq = ce3, facets = TRUE, show_fitted = TRUE,
        show.legend = TRUE)

}
\seealso{
\code{\link{coint_eq}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{iplots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uecm.R
\name{uecm}
\alias{uecm}
\alias{uecm.ardl}
\alias{uecm.default}
\title{Unrestricted ECM regression}
\usage{
uecm(...)

\method{uecm}{ardl}(object, ...)

\method{uecm}{default}(formula, data, order, start = NULL, end = NULL, ...)
}
\arguments{
\item{...}{Additional arguments to be passed to the low level regression
fitting functions.}

\item{object}{An object of \code{\link[base]{class}} 'ardl'.}

\item{formula}{A "formula" describing the linear model. Details for model
specification are given under 'Details'.}

\item{data}{A time series object (e.g., "ts", "zoo" or "zooreg") or a data
frame containing the variables in the model. In the case of a data frame,
it is coerced into a \code{\link[stats]{ts}} object with \code{start = 1},
\code{end = nrow(data)} and \code{frequency = 1}. If not found in data, the
variables are NOT taken from any environment.}

\item{order}{A specification of the order of the underlying ARDL model (e.g.,
for the UECM of an ARDL(1,0,2) model it should be \code{order = c(1,0,2)}).
A numeric vector of the same length as the total number of variables
(excluding the fixed ones, see 'Details'). It should only contain positive
integers or 0. An integer could be provided if all variables are of the
same order.}

\item{start}{Start of the time period which should be used for fitting the
model.}

\item{end}{End of the time period which should be used for fitting the model.}
}
\value{
\code{uecm} returns an object of \code{\link[base]{class}}
  \code{c("uecm", "dynlm", "lm")}. In addition, attributes 'order', 'data',
  'parsed_formula' and 'full_formula' are provided.
}
\description{
\code{uecm} is a generic function used to construct Unrestricted Error
Correction Models (UECM). The function invokes two different
\code{\link[utils]{methods}}. The default method works exactly like
\code{\link{ardl}}. The other method requires an object of
\code{\link[base]{class}} 'ardl'. Both methods create the conditional UECM,
which is the UECM of the underlying ARDL.
}
\details{
The \code{formula} should contain only variables that exist in the data
provided through \code{data} plus some additional functions supported by
\code{\link[dynlm]{dynlm}} (i.e., \code{trend()}).

You can also specify fixed variables that are not supposed to be lagged (e.g.
dummies etc.) simply by placing them after \code{|}. For example, \code{y ~
x1 + x2 | z1 + z2} where \code{z1} and \code{z2} are the fixed variables and
should not be considered in \code{order}. Note that the \code{|} notion
should not be confused with the same notion in \code{dynlm} where it
introduces instrumental variables.
}
\section{Mathematical Formula}{
 The formula of an Unrestricted ECM conditional
  to an \eqn{ARDL(p,q_{1},\dots,q_{k})}{ARDL(p,q1,...,qk)} is: \deqn{\Delta
  y_{t} = c_{0} + c_{1}t + \pi_{y}y_{t-1} + \sum_{j=1}^{k}\pi_{j}x_{j,t-1} +
  \sum_{i=1}^{p-1}\psi_{y,i}\Delta y_{t-i} +
  \sum_{j=1}^{k}\sum_{l=1}^{q_{j}-1} \psi_{j,l}\Delta x_{j,t-l} +
  \sum_{j=1}^{k}\omega_{j}\Delta x_{j,t} + \epsilon_{t}}
  \deqn{\psi_{j,l} = 0 \;\; \forall \;\; q_{j} \leq 1, \;\;\;\;\; \psi_{y,i}
  = 0 \;\; if \;\; p = 1}
  In addition, \eqn{x_{j,t-1}} and \eqn{\Delta x_{j,t}}
  cancel out becoming \eqn{x_{j,t} \;\; \forall \;\; q_{j} = 0}
}

\examples{
data(denmark)

## Estimate the UECM, conditional to it's underlying ARDL(3,1,3,2) -----

# Indirectly
ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
uecm_3132 <- uecm(ardl_3132)

# Directly
uecm_3132_ <- uecm(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
identical(uecm_3132, uecm_3132_)
summary(uecm_3132)

## Post-estimation testing ---------------------------------------------

library(lmtest) # for bgtest(), bptest(), and resettest()
library(tseries) # for jarque.bera.test()
library(strucchange) # for efp(), and sctest()

# Breusch-Godfrey test for higher-order serial correlation
bgtest(uecm_3132, order = 4)

# Breusch-Pagan test against heteroskedasticity
bptest(uecm_3132)

# Ramsey's RESET test for functional form
\dontrun{
    # This produces an error.
    # resettest() cannot use data of class 'zoo' such as the 'denmark' data
    # used to build the original model
    resettest(uecm_3132, type = c("regressor"))
}

uecm_3132_lm <- to_lm(uecm_3132, data_class = "ts")
resettest(uecm_3132_lm, power = 2)

# Jarque-Bera test for normality
jarque.bera.test(residuals(uecm_3132))

# CUSUM test for structural change detection
\dontrun{
    # This produces an error.
    # efp() does not understand special functions such as "d()" and "L()"
    efp(uecm_3132$full_formula, data = uecm_3132$model)
}

uecm_3132_lm_names <- to_lm(uecm_3132, fix_names = TRUE)
fluctuation <- efp(uecm_3132_lm_names$full_formula,
                   data = uecm_3132_lm_names$model)
sctest(fluctuation)
plot(fluctuation)

}
\seealso{
\code{\link{ardl}} \code{\link{recm}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{models}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coint_eq.R
\name{coint_eq}
\alias{coint_eq}
\alias{coint_eq.recm}
\alias{coint_eq.default}
\title{Cointegrating equation (long-run level relationship)}
\usage{
coint_eq(object, case)

\method{coint_eq}{recm}(object, ...)

\method{coint_eq}{default}(object, case)
}
\arguments{
\item{object}{An object of \code{\link[base]{class}} 'ardl', 'uecm' or
'recm'.}

\item{case}{An integer from 1-5 or a character string specifying whether the
'intercept' and/or the 'trend' have to participate in the long-run level
relationship (cointegrating equation) (see section 'Cases' below). If the
input object is of class 'recm', \code{case} is not needed as the model is
already under a certain case.}

\item{...}{Currently unused argument.}
}
\value{
\code{coint_eq} returns an numeric vector containing the
  cointegrating equation.
}
\description{
Creates the cointegrating equation (long-run level relationship) providing an
'ardl', 'uecm' or 'recm' model.
}
\section{Cases}{
 According to \cite{Pesaran et al. (2001)}, we distinguish the
long-run relationship (cointegrating equation) (and thus the bounds-test and
the Restricted ECMs) between 5 different cases. These differ in terms of
whether the 'intercept' and/or the 'trend' are restricted to participate in
the long-run relationship or they are unrestricted and so they participate in
the short-run relationship.

\describe{
  \item{Case 1:}{\itemize{
    \item No \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 1 or "n" where "n" stands for none.}}
  \item{Case 2:}{\itemize{
    \item Restricted \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 2 or "rc" where "rc" stands for restricted
     constant.}}
  \item{Case 3:}{\itemize{
    \item Unrestricted \emph{intercept} and no \emph{trend}.
    \item \code{case} inputs: 3 or "uc" where "uc" stands for unrestricted
    constant.}}
  \item{Case 4:}{\itemize{
    \item Unrestricted \emph{intercept} and restricted \emph{trend}.
    \item \code{case} inputs: 4 or "ucrt" where "ucrt" stands for
    unrestricted constant and restricted trend.}}
  \item{Case 5:}{\itemize{
    \item Unrestricted \emph{intercept} and unrestricted \emph{trend}.
    \item \code{case} inputs: 5 or "ucut" where "ucut" stands for
     unrestricted constant and unrestricted trend.}}
}

Note that you can't restrict (or leave unrestricted) a parameter that doesn't
exist in the input model. For example, you can't compute \code{recm(object,
case=3)} if the object is an ARDL (or UECM) model with no intercept. The same
way, you can't compute \code{bounds_f_test(object, case=5)} if the object is
an ARDL (or UECM) model with no linear trend.
}

\section{References}{
 Pesaran, M. H., Shin, Y., & Smith, R. J. (2001). Bounds
  testing approaches to the analysis of level relationships. \emph{Journal of
  Applied Econometrics}, 16(3), 289-326
}

\examples{
data(denmark)
library(zoo) # for cbind.zoo()

## Estimate the Cointegrating Equation of an ARDL(3,1,3,2) model -------

# From an ARDL model (under case 2, restricted constant)
ardl_3132 <- ardl(LRM ~ LRY + IBO + IDE, data = denmark, order = c(3,1,3,2))
ce2_ardl <- coint_eq(ardl_3132, case = 2)

# From an UECM (under case 2, restricted constant)
uecm_3132 <- uecm(ardl_3132)
ce2_uecm <- coint_eq(uecm_3132, case = 2)

# From a RECM (under case 2, restricted constant)
# Notice that if a RECM has already been estimated under a certain case,
# the 'coint_eq()' can't be under different case, so no 'case' argument needed.
recm_3132 <- recm(uecm_3132, case = 2)
# The RECM is already under case 2, so the 'case' argument is no needed
ce2_recm <- coint_eq(recm_3132)

identical(ce2_ardl, ce2_uecm, ce2_recm)

## Check for a degenerate level relationship ---------------------------

# The bounds F-test under both cases reject the Null Hypothesis of no level relationship.
bounds_f_test(ardl_3132, case = 2)
bounds_f_test(ardl_3132, case = 3)

# The bounds t-test also rejects the NUll Hypothesis of no level relationship.
bounds_t_test(ardl_3132, case = 3)

# But when the constant enters the long-run equation (case 3)
# this becomes a degenerate relationship.
ce3_ardl <- coint_eq(ardl_3132, case = 3)

plot_lr(ardl_3132, coint_eq = ce2_ardl, show.legend = TRUE)

plot_lr(ardl_3132, coint_eq = ce3_ardl, show.legend = TRUE)
plot_lr(ardl_3132, coint_eq = ce3_ardl, facets = TRUE, show.legend = TRUE)
}
\seealso{
\code{\link{plot_lr}} \code{\link{ardl}} \code{\link{uecm}} \code{\link{recm}}
  \code{\link{bounds_f_test}} \code{\link{bounds_t_test}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crit_val_bounds_sim.R
\name{vcov_custom}
\alias{vcov_custom}
\title{Variance-Covariance matrix of a regression}
\usage{
vcov_custom(indep_vars, model_res)
}
\arguments{
\item{indep_vars}{A matrix representing the independent variables.}

\item{model_res}{A numeric vector representing the regression's residuals.}
}
\value{
\code{vcov_custom} returns a Variance-Covariance matrix.
}
\description{
\code{vcov_custom} creates the Variance-Covariance matrix of a regression. It
is used instead of the \code{\link[stats]{vcov}} because the latter doesn't
work with \code{\link[stats]{.lm.fit}}.
}
\seealso{
\code{\link{f_test_custom}} \code{\link{f_bounds_sim}}
  \code{\link{t_bounds_sim}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto_ardl.R
\name{auto_ardl}
\alias{auto_ardl}
\title{Automatic ARDL model selection}
\usage{
auto_ardl(
  formula,
  data,
  max_order,
  fixed_order = -1,
  starting_order = NULL,
  selection = "AIC",
  selection_minmax = c("min", "max"),
  grid = FALSE,
  search_type = c("horizontal", "vertical"),
  start = NULL,
  end = NULL,
  ...
)
}
\arguments{
\item{formula}{A "formula" describing the linear model. Details for model
specification are given under 'Details' in the help file of the
\code{\link{ardl}} function.}

\item{data}{A time series object (e.g., "ts", "zoo" or "zooreg") or a data
frame containing the variables in the model. In the case of a data frame,
it is coerced into a \code{\link[stats]{ts}} object with \code{start = 1},
\code{end = nrow(data)} and \code{frequency = 1}. If not found in data, the
variables are NOT taken from any environment.}

\item{max_order}{It sets the maximum order for each variable where the search
is taking place. A numeric vector of the same length as the total number of
variables (excluding the fixed ones, see 'Details' in the help file of the
\code{\link{ardl}} function). It should only contain positive integers. An
integer could be provided if the maximum order for all variables is the
same.}

\item{fixed_order}{It allows setting a fixed order for some variables. The
algorithm will not search for any other order than this. A numeric vector
of the same length as the total number of variables (excluding the fixed
ones). It should contain positive integers or 0 to set as a constraint. A
-1 should be provided for any variable that should not be constrained.
\code{fixed_order} overrides the corresponding \code{max_order} and
\code{starting_order}.}

\item{starting_order}{Specifies the order for each variable from which each
  search will start. It is a numeric vector of the same length as the total
  number of variables (excluding the fixed ones). It should contain positive
  integers or 0 or only one integer could be provided if the starting order
  for all variables is the same. Default is set to NULL. If unspecified
  (\code{NULL}) and \code{grid = FALSE}, then all possible \eqn{ARDL(p)}
  models are calculated (constraints are taken into account), where \eqn{p}
  is the minimum value in \code{max_order}. Note that where
  \code{starting_order} is provided, its first element will be the minimum
value of \eqn{p} that the searching algorithm will consider (think of it like
a 'minimum p order' restriction) (see 'Searching algorithm' below). If
\code{grid = TRUE}, only the first argument (\eqn{p}) will have an effect.}

\item{selection}{A character string specifying the selection criterion
according to which the candidate models will be ranked. Default is
\code{\link[stats]{AIC}}. Any other selection criterion can be used (a user
specified or a function from another package) as long as it can be applied
as \code{selection(model)}. The preferred model is the one with the smaller
value of the selection criterion. If the selection criterion works the
other way around (the bigger the better), \code{selection_minmax = "max"}
should also be supplied (see 'Examples' below).}

\item{selection_minmax}{A character string that indicates whether the
criterion in \code{selection} is supposed to be minimized (default) or
maximized.}

\item{grid}{If \code{FALSE} (default), the stepwise searching regression
algorithm will search for the best model by adding and subtracting terms
corresponding to different ARDL orders. If \code{TRUE}, the whole set of
all possible ARDL models (accounting for constraints) will be evaluated.
Note that this method can be very time-consuming in case that
\code{max_order} is big and there are many independent variables that
create a very big number of possible combinations.}

\item{search_type}{A character string describing the search type. If
"horizontal" (default), the searching algorithm increases or decreases by 1
the order of each variable in each iteration. When the order of the last
variable has been accessed, it begins again from the first variable until
it converges. If "vertical", the searching algorithm increases or decreases
by 1 the order of a variable until it converges. Then it continues the same
for the next variable. The two options result to very similar top orders.
The default ("horizontal"), sometimes is a little more accurate, but the
"vertical" is almost 2 times faster. Not applicable if \code{grid = TRUE}.}

\item{start}{Start of the time period which should be used for fitting the
model.}

\item{end}{End of the time period which should be used for fitting the model.}

\item{...}{Additional arguments to be passed to the low level regression
fitting functions.}
}
\value{
\code{auto_ardl} returns a list which contains:
  \item{\code{best_model}}{An object of class \code{c("ardl", "dynlm", "lm")}}
  \item{\code{best_order}}{A numeric vector with the order of the best model selected}
  \item{\code{top_orders}}{A data.frame with the orders of the top 20 models}
}
\description{
It searches for the best ARDL order specification, according to the selected
criterion, taking into account the constraints provided.
}
\section{Searching algorithm}{
 The algorithm performs the optimization process
  starting from multiple starting points concerning the autoregressive order
  \eqn{p}. The searching algorithm will perform a complete search, each time
  starting from a different starting order. These orders are presented in the
  tables below, for \code{grid = FALSE} and different values of
  \code{starting_order}.

  \code{starting_order = NULL}:
  \tabular{ccccccc}{
  ARDL(p) \tab -> \tab p \tab q1 \tab q2 \tab ... \tab qk\cr
  ARDL(1) \tab -> \tab 1 \tab 1 \tab 1 \tab ... \tab 1\cr
  ARDL(2) \tab -> \tab 2 \tab 2 \tab 2 \tab ... \tab 2\cr
  : \tab -> \tab : \tab : \tab : \tab : \tab :\cr
  ARDL(P) \tab -> \tab P \tab P \tab P \tab ... \tab P
  }

  \code{starting_order = c(3, 0, 1, 2)}:
  \tabular{cccc}{
  p \tab q1 \tab q2 \tab q3\cr
  3 \tab 0 \tab 1 \tab 2\cr
  4 \tab 0 \tab 1 \tab 2\cr
  : \tab : \tab : \tab :\cr
  P \tab 0 \tab 1 \tab 2
  }
}

\examples{
data(denmark)

## Find the best ARDL order --------------------------------------------

# Up to 5 for the autoregressive order (p) and 4 for the rest (q1, q2, q3)

# Using the defaults search_type = "horizontal", grid = FALSE and selection = "AIC"
# ("Not run" indications only for testing purposes)
\dontrun{
model1 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                    max_order = c(5,4,4,4))
model1$top_orders

## Same, with search_type = "vertical" -------------------------------

model1_h <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                      max_order = c(5,4,4,4), search_type = "vertical")
model1_h$top_orders

## Find the global optimum ARDL order ----------------------------------

# It may take more than 10 seconds
model_grid <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                        max_order = c(5,4,4,4), grid = TRUE)

## Different selection criteria ----------------------------------------

# Using BIC as selection criterion instead of AIC
model1_b <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                      max_order = c(5,4,4,4), selection = "BIC")
model1_b$top_orders

# Using other criteria like adjusted R squared (the bigger the better)
adjr2 <- function(x) { summary(x)$adj.r.squared }
model1_adjr2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                           max_order = c(5,4,4,4), selection = "adjr2",
                           selection_minmax = "max")
model1_adjr2$top_orders

# Using functions from other packages as selection criteria
if (requireNamespace("AICcmodavg", quietly = TRUE)) {

library(AICcmodavg)
model1_aicc <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                          max_order = c(5,4,4,4), selection = "AICc")
model1_aicc$top_orders
adjr2 <- function(x){ summary(x)$adj.r.squared }
model1_adjr2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                           max_order = c(5,4,4,4), selection = "adjr2",
                           selection_minmax = "max")
model1_adjr2$top_orders

## DIfferent starting order --------------------------------------------

# The searching algorithm will start from the following starting orders:
# p q1 q2 q3
# 1 1  3  2
# 2 1  3  2
# 3 1  3  2
# 4 1  3  2
# 5 1  3  2

model1_so <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                        max_order = c(5,4,4,4), starting_order = c(1,1,3,2))

# Starting from p=3 (don't search for p=1 and p=2)
# Starting orders:
# p q1 q2 q3
# 3 1  3  2
# 4 1  3  2
# 5 1  3  2

model1_so_3 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                        max_order = c(5,4,4,4), starting_order = c(3,1,3,2))

# If starting_order = NULL, the starting orders for each iteration will be:
# p q1 q2 q3
# 1 1  1  1
# 2 2  2  2
# 3 3  3  3
# 4 4  4  4
# 5 5  5  5
}

## Add constraints -----------------------------------------------------

# Restrict only the order of IBO to be 2
model1_ibo2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                        max_order = c(5,4,4,4), fixed_order = c(-1,-1,2,-1))
model1_ibo2$top_orders

# Restrict the order of LRM to be 3 and the order of IBO to be 2
model1_lrm3_ibo2 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                        max_order = c(5,4,4,4), fixed_order = c(3,-1,2,-1))
model1_lrm3_ibo2$top_orders

## Set the starting date for the regression (data starts at "1974 Q1") -

# Set regression starting date to "1976 Q1"
model1_76q1 <- auto_ardl(LRM ~ LRY + IBO + IDE, data = denmark,
                        max_order = c(5,4,4,4), start = "1976 Q1")
start(model1_76q1$best_model)
}
}
\seealso{
\code{\link{ardl}}
}
\author{
Kleanthis Natsiopoulos, \email{klnatsio@gmail.com}
}
\keyword{models}
\keyword{optimize}
\keyword{ts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{denmark}
\alias{denmark}
\title{The Danish data on money income prices and interest rates}
\format{
A time-series object with 55 rows and 5 variables. Time period from 1974:Q1
until 1987:Q3.
\describe{
  \item{LRM}{logarithm of real money, M2}
  \item{LRY}{logarithm of real income}
  \item{LPY}{logarithm of price deflator}
  \item{IBO}{bond rate}
  \item{IDE}{bank deposit rate}
}
}
\source{
\url{https://onlinelibrary.wiley.com/doi/10.1111/j.1468-0084.1990.mp52002003.x}
}
\usage{
denmark
}
\description{
This data set contains the series used by S. Johansen and K. Juselius for
estimating a money demand function of Denmark.
}
\details{
An object of class "zooreg" "zoo".
}
\references{
Johansen, S. and Juselius, K. (1990), Maximum Likelihood
Estimation and Inference on Cointegration -- with Applications to the Demand
for Money, \emph{Oxford Bulletin of Economics and Statistics}, \bold{52, 2},
169--210.
}
\keyword{datasets}
